#include <NMath.h>
#include <CFilter.h>
#include "CDiffusion.h"

// constructor
CDiffusion::CDiffusion() {
  mTau1D  = 1.0;
  mTau2D  = 1.0;
  mTau3D  = 1.0;
  mLambda = 4;
  mEpsilon = 0.001;
  mP = -0.5;
  mSmoothing = 0;
  mDiffFunc = cTV;
  mLimitedCondNo = false;
  mEnhancedStencil = false;
}

// Isotropic -------------------------------------------------------------------

void CDiffusion::isotropicDiffusion(CVector<float>& aVector, float aTime) {
  if (aTime == 0.0) return;
  float aTau = mTau1D;
  if (aTime < aTau) mTau1D = aTime;
  int aIterations = (int)ceil(aTime/mTau1D);
  CVector<float> aDiffusivity(aVector.size());
  for (int t = 0; t < aIterations; t++) {
    isotropicDiffVector(aVector,aDiffusivity);
    isotropicDiffuse(aVector,aDiffusivity);
  }
  mTau1D = aTau;
}

// isotropicDiffusion (2-D)
void CDiffusion::isotropicDiffusion(CMatrix<float>& aMatrix, float aTime) {
  if (aTime == 0.0) return;
  float aTau = mTau2D;
  if (aTime < aTau) mTau2D = aTime;
  int aIterations = (int)ceil(aTime/mTau2D);
  // One-sided differences scheme
  if (mSmoothing == 0) {
    CMatrix<float> GX(aMatrix.xSize(),aMatrix.ySize());
    CMatrix<float> GY(aMatrix.xSize(),aMatrix.ySize());
    for (int k = 0; k < aIterations; k++) {
      // Compute diffusivity -> GX,GY
      isotropicDiffMatrix(aMatrix,GX,GY);
      // Compute diffusion
      isotropicDiffuse(aMatrix,GX,GY);
    }
  }
  // Central differences scheme
  else {
    CMatrix<float> G(aMatrix.xSize(),aMatrix.ySize());
    for (int k = 0; k < aIterations; k++) {
      // Compute diffusivity -> G
      isotropicDiffMatrix(aMatrix,G);
      // Compute diffusion
      isotropicDiffuse(aMatrix,G);
    }
  }
  mTau2D = aTau;
}

// isotropicDiffusionMultiChannel (2-D)
void CDiffusion::isotropicDiffusionMultiChannel(CTensor<float>& aTensor, float aTime) {
  if (aTime == 0.0) return;
  float aTau = mTau2D;
  if (aTime < aTau) mTau2D = aTime;
  int aIterations = (int)ceil(aTime/mTau2D);
  CMatrix<float> aTemp(aTensor.xSize(),aTensor.ySize());
  // One-sided differences scheme
  if (mSmoothing == 0) {
    CMatrix<float> GX(aTensor.xSize(),aTensor.ySize());
    CMatrix<float> GY(aTensor.xSize(),aTensor.ySize());
    for (int k = 0; k < aIterations; k++) {
      // Compute diffusivity -> GX,GY
      isotropicDiffMatrixMultiChannel(aTensor,GX,GY);
      // Compute diffusion
      for (int k = 0; k < aTensor.zSize(); k++) {
        aTensor.getMatrix(aTemp,k);
        isotropicDiffuse(aTemp,GX,GY);
        aTensor.putMatrix(aTemp,k);
      }
    }
  }
  // Central differences scheme
  else {
    CMatrix<float> G(aTensor.xSize(),aTensor.ySize());
    for (int k = 0; k < aIterations; k++) {
      // Compute diffusivity -> G
      isotropicDiffMatrixMultiChannel(aTensor,G);
      // Diffuse
      for (int k = 0; k < aTensor.zSize(); k++) {
        aTensor.getMatrix(aTemp,k);
        isotropicDiffuse(aTemp,G);
        aTensor.putMatrix(aTemp,k);
      }
    }
  }
  mTau2D = aTau;
}

// isotropicDiffusion (3-D)
void CDiffusion::isotropicDiffusion(CTensor<float>& aTensor, float aTime) {
  if (aTime == 0.0) return;
  float aTau = mTau3D;
  if (aTime < aTau) mTau3D = aTime;
  int aIterations = (int)ceil(aTime/mTau3D);
  // Central differences scheme
  CTensor<float> G(aTensor.xSize(),aTensor.ySize(),aTensor.zSize());
  for (int k = 0; k < aIterations; k++) {
    // Compute diffusivity -> G
    isotropicDiffTensor(aTensor,G);
    // Diffuse
    isotropicDiffuse(aTensor,G);
  }
  mTau3D = aTau;
}


// isotropicDiffusionMultiChannel (3-D)
void CDiffusion::isotropicDiffusionMultiChannel(CTensor4D<float>& aTensor, float aTime) {
  if (aTime == 0.0) return;
  float aTau = mTau3D;
  if (aTime < aTau) mTau3D = aTime;
  int aIterations = (int)ceil(aTime/mTau3D);
  CTensor<float> aTemp(aTensor.xSize(),aTensor.ySize(),aTensor.zSize());
  // Central differences scheme
  CTensor<float> G(aTensor.xSize(),aTensor.ySize(),aTensor.zSize());
  for (int k = 0; k < aIterations; k++) {
    // Compute diffusivity -> G
    isotropicDiffTensorMultiChannel(aTensor,G);
    // Diffuse
    for (int k = 0; k < aTensor.aSize(); k++) {
      aTensor.getTensor3D(aTemp,k);
      isotropicDiffuse(aTemp,G);
      aTensor.putTensor3D(aTemp,k);
    }
  }
  mTau3D = aTau;
}

// Anisotropic -----------------------------------------------------------------

// anisotropicDiffusion
void CDiffusion::anisotropicDiffusion(CMatrix<float>& aMatrix, float aTime) {
  if (aTime == 0.0) return;
  float aTau = mTau2D;
  if (aTime < aTau) mTau2D = aTime;
  int aIterations = (int)ceil(aTime/mTau2D);
  CMatrix<float> A(aMatrix.xSize(),aMatrix.ySize());
  CMatrix<float> B(aMatrix.xSize(),aMatrix.ySize());
  CMatrix<float> C(aMatrix.xSize(),aMatrix.ySize());
  for (int t = 0; t < aIterations; t++) {
    anisotropicDiffMatrix(aMatrix,A,B,C);
    // Diffuse
    if (mEnhancedStencil) anisotropicEnhancedDiffuse(aMatrix,A,B,C);
    else anisotropicDiffuse(aMatrix,A,B,C);
  }
  mTau2D = aTau;
}

// anisotropicDiffusionMultiChannel
void CDiffusion::anisotropicDiffusionMultiChannel(CTensor<float>& aTensor, float aTime, float aRho) {
  if (aTime == 0.0) return;
  float aTau = mTau2D;
  if (aTime < aTau) mTau2D = aTime;
  int aIterations = (int)ceil(aTime/mTau2D);
  CMatrix<float> aDXX(aTensor.xSize(),aTensor.ySize());
  CMatrix<float> aDXY(aTensor.xSize(),aTensor.ySize());
  CMatrix<float> aDYY(aTensor.xSize(),aTensor.ySize());
  CMatrix<float> aTemp(aTensor.xSize(),aTensor.ySize());
  for (int t = 0; t < aIterations; t++) {
    // Compute diffusion tensor
    anisotropicDiffMatrixMultiChannel(aTensor,aDXX,aDXY,aDYY,aRho);
    // Diffuse
    for (int k = 0; k < aTensor.zSize(); k++) {
      aTensor.getMatrix(aTemp,k);
      if (mEnhancedStencil) anisotropicEnhancedDiffuse(aTemp,aDXX,aDXY,aDYY);
      else anisotropicDiffuse(aTemp,aDXX,aDXY,aDYY);
      aTensor.putMatrix(aTemp,k);
    }
  }
  mTau2D = aTau;
}

// anisotropicDiffusion (3-D)
void CDiffusion::anisotropicDiffusion(CTensor<float>& aTensor, float aTime) {
  if (aTime == 0.0) return;
  float aTau = mTau3D;
  if (aTime < aTau) mTau3D = aTime;
  int aIterations = (int)ceil(aTime/mTau3D);
  CTensor<float> aDXX(aTensor.xSize(),aTensor.ySize(),aTensor.zSize());
  CTensor<float> aDYY(aTensor.xSize(),aTensor.ySize(),aTensor.zSize());
  CTensor<float> aDZZ(aTensor.xSize(),aTensor.ySize(),aTensor.zSize());
  CTensor<float> aDXY(aTensor.xSize(),aTensor.ySize(),aTensor.zSize());
  CTensor<float> aDXZ(aTensor.xSize(),aTensor.ySize(),aTensor.zSize());
  CTensor<float> aDYZ(aTensor.xSize(),aTensor.ySize(),aTensor.zSize());
  for (int t = 0; t < aIterations; t++) {
    // Compute diffusivity
    anisotropicDiffTensor(aTensor,aDXX,aDYY,aDZZ,aDXY,aDXZ,aDYZ);
    // Diffuse
    if (mEnhancedStencil) anisotropicEnhancedDiffuse(aTensor,aDXX,aDYY,aDZZ,aDXY,aDXZ,aDYZ);
    else anisotropicDiffuse(aTensor,aDXX,aDYY,aDZZ,aDXY,aDXZ,aDYZ);
  }
  mTau3D = aTau;
}

// anisotropicDiffusionMultiChannel (3-D)
void CDiffusion::anisotropicDiffusionMultiChannel(CTensor4D<float>& aTensor, float aTime) {
  if (aTime == 0.0) return;
  float aTau = mTau3D;
  if (aTime < aTau) mTau3D = aTime;
  int aIterations = (int)ceil(aTime/mTau3D);
  CTensor<float> aDXX(aTensor.xSize(),aTensor.ySize(),aTensor.zSize());
  CTensor<float> aDYY(aTensor.xSize(),aTensor.ySize(),aTensor.zSize());
  CTensor<float> aDZZ(aTensor.xSize(),aTensor.ySize(),aTensor.zSize());
  CTensor<float> aDXY(aTensor.xSize(),aTensor.ySize(),aTensor.zSize());
  CTensor<float> aDXZ(aTensor.xSize(),aTensor.ySize(),aTensor.zSize());
  CTensor<float> aDYZ(aTensor.xSize(),aTensor.ySize(),aTensor.zSize());
  CTensor<float> aTemp(aTensor.xSize(),aTensor.ySize(),aTensor.zSize());
  for (int t = 0; t < aIterations; t++) {
    // Compute diffusion tensor
    anisotropicDiffTensorMultiChannel(aTensor,aDXX,aDYY,aDZZ,aDXY,aDXZ,aDYZ);
    // Diffuse
    for (int k = 0; k < aTensor.aSize(); k++) {
      aTensor.getTensor3D(aTemp,k);
      if (mEnhancedStencil) anisotropicEnhancedDiffuse(aTemp,aDXX,aDYY,aDZZ,aDXY,aDXZ,aDYZ);
      else anisotropicDiffuse(aTemp,aDXX,aDYY,aDZZ,aDXY,aDXZ,aDYZ);
      aTensor.putTensor3D(aTemp,k);
    }
  }
  mTau3D = aTau;
}

// =============================================================================
// Semi-implicifloat AOS
// =============================================================================

// isotropicDiffusionSemiImplicifloat (1-D)
void CDiffusion::isotropicDiffusionSemiImplicit(CVector<float>& aVector, float aTime) {
  if (aTime == 0.0) return;
  CVector<float> G(aVector.size());
  float aTau = mTau1D;
  if (aTime < aTau) mTau1D = aTime;
  int aIterations = (int)ceil(aTime/mTau1D);
  for (float k = 0; k < aIterations; k++) {
    // Compute diffusivity -> G
    isotropicDiffVector(aVector,G);
    // Compute diffusion
    isotropicDiffuseSemiImplicit(aVector,G);
  }
  mTau1D = aTau;
}

// isotropicDiffusionSemiImplicifloat (2D)
void CDiffusion::isotropicDiffusionSemiImplicit(CMatrix<float>& aMatrix, float aTime) {
  if (aTime == 0.0) return;
  float aTau = mTau2D;
  if (aTime < aTau) mTau2D = aTime;
  int aIterations = (int)ceil(aTime/mTau2D);
  // One-sided differences scheme
  if (mSmoothing == 0) {
    CMatrix<float> GX(aMatrix.xSize(),aMatrix.ySize());
    CMatrix<float> GY(aMatrix.xSize(),aMatrix.ySize());
    for (int k = 0; k < aIterations; k++) {
      // Compute diffusivity -> GX,GY
      isotropicDiffMatrix(aMatrix,GX,GY);
      // Compute diffusion
      isotropicDiffuseSemiImplicit(aMatrix,GX,GY);
    }
  }
  // Central differences scheme
  else {
    CMatrix<float> G(aMatrix.xSize(),aMatrix.ySize());
    for (int k = 0; k < aIterations; k++) {
      // Compute diffusivity -> G
      isotropicDiffMatrix(aMatrix,G);
      // Compute diffusion
      isotropicDiffuseSemiImplicit(aMatrix,G,G);
    }
  }
  mTau2D = aTau;
}

// isotropicDiffusionMultiChannelSemiImplicifloat (2D)
void CDiffusion::isotropicDiffusionMultiChannelSemiImplicit(CTensor<float>& aTensor, float aTime) {
  if (aTime == 0.0) return;
  float aTau = mTau2D;
  if (aTime < aTau) mTau2D = aTime;
  int aIterations = (int)ceil(aTime/mTau2D);
  CMatrix<float> aTemp(aTensor.xSize(),aTensor.ySize());
  // One-sided differences scheme
  if (mSmoothing == 0) {
    CMatrix<float> GX(aTensor.xSize(),aTensor.ySize());
    CMatrix<float> GY(aTensor.xSize(),aTensor.ySize());
    for (int k = 0; k < aIterations; k++) {
      // Compute diffusivity -> GX,GY
      isotropicDiffMatrixMultiChannel(aTensor,GX,GY);
      // Compute diffusion
      for (int z = 0; z < aTensor.zSize(); z++) {
        aTensor.getMatrix(aTemp,z);
        isotropicDiffuseSemiImplicit(aTemp,GX,GY);
        aTensor.putMatrix(aTemp,z);
      }
    }
  }
  // Central differences scheme
  else {
    CMatrix<float> G(aTensor.xSize(),aTensor.ySize());
    for (int k = 0; k < aIterations; k++) {
      // Compute diffusivity -> G
      isotropicDiffMatrixMultiChannel(aTensor,G);
      // Compute diffusion
      for (int z = 0; z < aTensor.zSize(); z++) {
        aTensor.getMatrix(aTemp,z);
        isotropicDiffuseSemiImplicit(aTemp,G,G);
        aTensor.putMatrix(aTemp,z);
      }
    }
  }
  mTau2D = aTau;
}

// isotropicDiffusionSemiImplicifloat (3-D)
void CDiffusion::isotropicDiffusionSemiImplicit(CTensor<float>& aTensor, float aTime) {
  if (aTime == 0.0) return;
  float aTau = mTau3D;
  if (aTime < aTau) mTau3D = aTime;
  int aIterations = (int)ceil(aTime/mTau3D);
  // One-sided differences scheme
  if (mSmoothing == 0) {
    CTensor<float> GX(aTensor.xSize(),aTensor.ySize(),aTensor.zSize());
    CTensor<float> GY(aTensor.xSize(),aTensor.ySize(),aTensor.zSize());
    CTensor<float> GZ(aTensor.xSize(),aTensor.ySize(),aTensor.zSize());
    for (int k = 0; k < aIterations; k++) {
      // Compute diffusivity -> GX,GY,GZ
      isotropicDiffTensor(aTensor,GX,GY,GZ);
      // Compute diffusion
      isotropicDiffuseSemiImplicit(aTensor,GX,GY,GZ);
    }
  }
  // Central differences scheme
  else {
    CTensor<float> G(aTensor.xSize(),aTensor.ySize(),aTensor.zSize());
    for (int k = 0; k < aIterations; k++) {
      // Compute diffusivity -> G
      isotropicDiffTensor(aTensor,G);
      // Compute diffusion
      isotropicDiffuseSemiImplicit(aTensor,G,G,G);
    }
  }
  mTau3D = aTau;
}

// isotropicDiffusionMultiChannelSemiImplicifloat (3D)
void CDiffusion::isotropicDiffusionMultiChannelSemiImplicit(CTensor4D<float>& aTensor, float aTime) {
  if (aTime == 0.0) return;
  float aTau = mTau3D;
  if (aTime < aTau) mTau3D = aTime;
  int aIterations = (int)ceil(aTime/mTau3D);
  CTensor<float> aTemp(aTensor.xSize(),aTensor.ySize(),aTensor.zSize());
  // One-sided differences scheme
  if (mSmoothing == 0) {
    CTensor<float> GX(aTensor.xSize(),aTensor.ySize(),aTensor.zSize());
    CTensor<float> GY(aTensor.xSize(),aTensor.ySize(),aTensor.zSize());
    CTensor<float> GZ(aTensor.xSize(),aTensor.ySize(),aTensor.zSize());
    for (int k = 0; k < aIterations; k++) {
      // Compute diffusivity -> GX,GY,GZ
      isotropicDiffTensorMultiChannel(aTensor,GX,GY,GZ);
      // Compute diffusion
      for (int a = 0; a < aTensor.aSize(); a++) {
        aTensor.getTensor3D(aTemp,a);
        isotropicDiffuseSemiImplicit(aTemp,GX,GY,GZ);
        aTensor.putTensor3D(aTemp,a);
      }
    }
  }
  // Central differences scheme
  else {
    CTensor<float> G(aTensor.xSize(),aTensor.ySize(),aTensor.zSize());
    for (int k = 0; k < aIterations; k++) {
      // Compute diffusivity -> G
      isotropicDiffTensorMultiChannel(aTensor,G);
      // Compute diffusion
      for (int a = 0; a < aTensor.aSize(); a++) {
        aTensor.getTensor3D(aTemp,a);
        isotropicDiffuseSemiImplicit(aTemp,G,G,G);
        aTensor.putTensor3D(aTemp,a);
      }
    }
  }
  mTau3D = aTau;
}

// Anisotropic AOS -------------------------------------------------------------

// anisotropicDiffusionSemiImplicifloat (2-D)
void CDiffusion::anisotropicDiffusionSemiImplicit(CMatrix<float>& aMatrix, float aTime) {
  if (aTime == 0.0) return;
  float aTau = mTau2D;
  if (aTime < aTau) mTau2D = aTime;
  int aIterations = (int)ceil(aTime/mTau2D);
  CMatrix<float> A(aMatrix.xSize(),aMatrix.ySize());
  CMatrix<float> B(aMatrix.xSize(),aMatrix.ySize());
  CMatrix<float> C(aMatrix.xSize(),aMatrix.ySize());
  for (int k = 0; k < aIterations; k++) {
    // Compute diffusivity -> G
    anisotropicDiffMatrix(aMatrix,A,B,C);
    // Compute diffusion
    anisotropicDiffuseSemiImplicit(aMatrix,A,B,C);
  }
  mTau2D = aTau;
}

// anisotropicDiffusionMultiChannelSemiImplicifloat (2-D)
void CDiffusion::anisotropicDiffusionMultiChannelSemiImplicit(CTensor<float>& aTensor, float aTime) {
  if (aTime == 0.0) return;
  float aTau = mTau2D;
  if (aTime < aTau) mTau2D = aTime;
  int aIterations = (int)ceil(aTime/mTau2D);
  CMatrix<float> aTemp(aTensor.xSize(),aTensor.ySize());
  CMatrix<float> A(aTensor.xSize(),aTensor.ySize());
  CMatrix<float> B(aTensor.xSize(),aTensor.ySize());
  CMatrix<float> C(aTensor.xSize(),aTensor.ySize());
  for (int k = 0; k < aIterations; k++) {
    // Compute diffusivity -> G
    anisotropicDiffMatrixMultiChannel(aTensor,A,B,C);
    // Compute diffusion
    for (int z = 0; z < aTensor.zSize(); z++) {
      aTensor.getMatrix(aTemp,z);
      anisotropicDiffuseSemiImplicit(aTemp,A,B,C);
      aTensor.putMatrix(aTemp,z);
    }
  }
  mTau2D = aTau;
}

// =============================================================================
// Two-pixel scheme
// =============================================================================

// twoPixelFlow (1D)
void CDiffusion::twoPixelFlow(CVector<float>& aData, float aTime) {
  CVector<float> aGrad(aData.size());
  CVector<float> aOld(aData.size());
  int aStoppingTime = (int)(aTime/mTau1D);
  for (int t = 0; t < aStoppingTime; t++) {
    aOld = aData;
    // Left pixel
    twoPixelGrad(aData,aGrad,0);
    twoPixelDiffuseLeft(aData,aGrad);
    // Right pixel
    twoPixelGrad(aOld,aGrad,1);
    twoPixelDiffuseRight(aOld,aGrad);
    for (int i = 0; i < aData.size(); i++)
      aData(i) = 0.5*(aData(i)+aOld(i));
  }
}

// twoPixelFlow (2-D)
void CDiffusion::twoPixelFlow(CMatrix<float>& aData, float aTime) {
  CMatrix<float> aGrad(aData.xSize(),aData.ySize());
  int aStoppingTime = (int)(aTime/mTau2D);
  for (int t = 0; t < aStoppingTime; t++) {
    // Left pixel
    twoPixelGradX(aData,aGrad,0);
    twoPixelDiffuseLeft(aData,aGrad);
    // Right pixel
    twoPixelGradX(aData,aGrad,1);
    twoPixelDiffuseRight(aData,aGrad);
    // Upper pixel
    twoPixelGradY(aData,aGrad,0);
    twoPixelDiffuseUp(aData,aGrad);
    // Lower pixel
    twoPixelGradY(aData,aGrad,1);
    twoPixelDiffuseDown(aData,aGrad);
  }
}

// twoPixelFlowMultiChannel (2-D)
void CDiffusion::twoPixelFlowMultiChannel(CTensor<float>& aData, float aTime) {
  CMatrix<float> aGrad(aData.xSize(),aData.ySize());
  CMatrix<float> aTemp(aData.xSize(),aData.ySize());
  int aStoppingTime = (int)(aTime/mTau2D);
  for (int t = 0; t < aStoppingTime; t++) {
    // Left pixel
    twoPixelGradXMultiChannel(aData,aGrad,0);
    for (int k = 0; k < aData.zSize(); k++) {
      aData.getMatrix(aTemp,k);
      twoPixelDiffuseLeft(aTemp,aGrad);
      aData.putMatrix(aTemp,k);
    }
    // Right pixel
    twoPixelGradXMultiChannel(aData,aGrad,1);
    for (int k = 0; k < aData.zSize(); k++) {
      aData.getMatrix(aTemp,k);
      twoPixelDiffuseRight(aTemp,aGrad);
      aData.putMatrix(aTemp,k);
    }
    // Upper pixel
    twoPixelGradYMultiChannel(aData,aGrad,0);
    for (int k = 0; k < aData.zSize(); k++) {
      aData.getMatrix(aTemp,k);
      twoPixelDiffuseUp(aTemp,aGrad);
      aData.putMatrix(aTemp,k);
    }
    // Lower pixel
    twoPixelGradYMultiChannel(aData,aGrad,1);
    for (int k = 0; k < aData.zSize(); k++) {
      aData.getMatrix(aTemp,k);
      twoPixelDiffuseDown(aTemp,aGrad);
      aData.putMatrix(aTemp,k);
    }
  }
}

// twoPixelFlow (3-D)
void CDiffusion::twoPixelFlow(CTensor<float>& aData, float aTime) {
  CTensor<float> aGrad(aData.xSize(),aData.ySize(),aData.zSize());
  int aStoppingTime = (int)(aTime/mTau3D);
  for (int t = 0; t < aStoppingTime; t++) {
    // Left pixel
    twoPixelGradX(aData,aGrad,0);
    twoPixelDiffuseLeft(aData,aGrad);
    // Upper pixel
    twoPixelGradY(aData,aGrad,0);
    twoPixelDiffuseUp(aData,aGrad);
    // Inner pixel
    twoPixelGradZ(aData,aGrad,0);
    twoPixelDiffuseInner(aData,aGrad);
    // Right pixel
    twoPixelGradX(aData,aGrad,1);
    twoPixelDiffuseRight(aData,aGrad);
    // Lower pixel
    twoPixelGradY(aData,aGrad,1);
    twoPixelDiffuseDown(aData,aGrad);
    // Outer pixel
    twoPixelGradZ(aData,aGrad,1);
    twoPixelDiffuseOuter(aData,aGrad);
  }
}

// twoPixelFlowMultiChannel (3-D)
void CDiffusion::twoPixelFlowMultiChannel(CTensor4D<float>& aData, float aTime) {
  CTensor<float> aGrad(aData.xSize(),aData.ySize(),aData.zSize());
  CTensor<float> aTemp(aData.xSize(),aData.ySize(),aData.zSize());
  int aStoppingTime = (int)(aTime/mTau3D);
  for (int t = 0; t < aStoppingTime; t++) {
    // Left pixel
    twoPixelGradXMultiChannel(aData,aGrad,0);
    for (int k = 0; k < aData.aSize(); k++) {
      aData.getTensor3D(aTemp,k);
      twoPixelDiffuseLeft(aTemp,aGrad);
      aData.putTensor3D(aTemp,k);
    }
    // Upper pixel
    twoPixelGradYMultiChannel(aData,aGrad,0);
    for (int k = 0; k < aData.aSize(); k++) {
      aData.getTensor3D(aTemp,k);
      twoPixelDiffuseUp(aTemp,aGrad);
      aData.putTensor3D(aTemp,k);
    }
    // Inner Pixel
    twoPixelGradZMultiChannel(aData,aGrad,0);
    for (int k = 0; k < aData.aSize(); k++) {
      aData.getTensor3D(aTemp,k);
      twoPixelDiffuseInner(aTemp,aGrad);
      aData.putTensor3D(aTemp,k);
    }
    // Right pixel
    twoPixelGradXMultiChannel(aData,aGrad,1);
    for (int k = 0; k < aData.aSize(); k++) {
      aData.getTensor3D(aTemp,k);
      twoPixelDiffuseRight(aTemp,aGrad);
      aData.putTensor3D(aTemp,k);
    }
    // Lower pixel
    twoPixelGradYMultiChannel(aData,aGrad,1);
    for (int k = 0; k < aData.aSize(); k++) {
      aData.getTensor3D(aTemp,k);
      twoPixelDiffuseDown(aTemp,aGrad);
      aData.putTensor3D(aTemp,k);
    }
    // Outer Pixel
    twoPixelGradZMultiChannel(aData,aGrad,1);
    for (int k = 0; k < aData.aSize(); k++) {
      aData.getTensor3D(aTemp,k);
      twoPixelDiffuseOuter(aTemp,aGrad);
      aData.putTensor3D(aTemp,k);
    }

  }
}

// =============================================================================
// Regularization
// =============================================================================

// regularization (1-D)
void CDiffusion::regularization(const CVector<float>& aData, CVector<float>& aResult, float aAlpha, int aIterations) {
  CVector<float> G(aData.size());
  CVector<float> M(aData.size());
  CVector<float> Y(aData.size());
  CVector<float> I(aData.size());
  CVector<float> J(aData.size());
  CVector<float> K(aData.size());
  int aSize = aData.size();
  aResult = aData;
  for (int t = 0; t < aIterations; t++) {
    // Compute diffusivity -> G
    isotropicDiffVector(aResult,G);
    // Build matrix
    // (a) One-sided differences
    if (mSmoothing == 0) {
      // diagonal -> I   off-diagonals -> J,K
      float aTemp1 = aAlpha*G(0);
      float aTemp2;
      I(0) = 1.0+aTemp1;
      J(0) = -aTemp1;
      for (int x = 1; x+1 < aSize; x++) {
        aTemp1 = aAlpha*G(x);
        aTemp2 = aAlpha*G(x-1);
        I(x) = 1+aTemp1+aTemp2;
        J(x) = -aTemp1;
        K(x-1) = -aTemp2;
      }
      aTemp2 = aAlpha*G(aSize-2);
      I(aSize-1) = 1.0+aTemp2;
      J(aSize-1) = 0;
      K(aSize-2) = -aTemp2;
    }
    // (b) Central differences
    else {
      float aTau = 0.5*aAlpha;
      // diagonal -> I   off-diagonals -> J,K
      float aTemp1 = aTau*(G(0)+G(1));
      float aTemp2;
      I(0) = 1.0+aTemp1;
      J(0) = -aTemp1;
      for (int x = 1; x+1 < aSize; x++) {
        aTemp1 = aTau*(G(x)+G(x+1));
        aTemp2 = aTau*(G(x)+G(x-1));
        I(x) = 1+aTemp1+aTemp2;
        J(x) = -aTemp1;
        K(x-1) = -aTemp2;
      }
      aTemp2 = aTau*(G(aSize-1)+G(aSize-2));
      I(aSize-1) = 1.0+aTemp2;
      J(aSize-1) = 0;
      K(aSize-2) = -aTemp2;
    }
    // Thomas algorithm
    // Forward substitution
    M(0) = 1.0/I(0);
    Y(0) = aData(0);
    int i = 0;
    do {
      float l = K(i)*M(i);
      i++;
      M(i) = 1.0/(I(i)-l*J(i-1));
      Y(i) = aData(i)-l*Y(i-1);
    }
    while (i < aSize-1);
    // Backward substitution
    aResult(aSize-1) = Y(aSize-1)*M(aSize-1);
    for (int j = aSize-2; j >= 0; j--)
      aResult(j) = (Y(j)-J(j)*aResult(j+1))*M(j);
  }
}

// regularization (2D)
void CDiffusion::regularization(const CMatrix<float>& aData, CMatrix<float>& aResult, float aAlpha, int aIterations, float aOmega) {
  aResult = aData;
  CMatrix<float> GX(aData.xSize(),aData.ySize());
  CMatrix<float> GY(aData.xSize(),aData.ySize());
  CMatrix<float> aHelp(aData);
  float invAlpha = 1.0/aAlpha;
  aHelp *= invAlpha;
  for (int t = 0; t < aIterations; t++) {
    // Compute diffusivity -> GX,GY
    if (t % 5 == 0) isotropicDiffMatrix(aResult,GX,GY);
    for (int y = 0; y < aData.ySize(); y++)
      for (int x = 0; x < aData.xSize(); x++) {
        float aSum = 0;
        float aNum = 0;
        float aTemp;
        if (x > 0) {
          aTemp = GX(x-1,y);
          aSum += aTemp*aResult(x-1,y);
          aNum += aTemp;
        }
        if (y > 0) {
          aTemp = GY(x,y-1);
          aSum += aTemp*aResult(x,y-1);
          aNum += aTemp;
        }
        if (x+1 < aData.xSize()) {
          aTemp = GX(x,y);
          aSum += aTemp*aResult(x+1,y);
          aNum += aTemp;
        }
        if (y+1 < aData.ySize()) {
          aTemp = GY(x,y);
          aSum += aTemp*aResult(x,y+1);
          aNum += aTemp;
        }
        aTemp = aOmega/(aNum + invAlpha);
        aResult(x,y) = (1-aOmega)*aResult(x,y)+(aSum+aHelp(x,y))*aTemp;
      }
    }
}

// -----------------------------------------------------------------------------
// P R I V A float E
// -----------------------------------------------------------------------------

// presmoothedDerivative (1-D)
void CDiffusion::presmoothedDerivative(const CVector<float>& aVector, CVector<float>& aGradient) {
  int aStillToSmooth = 2*mSmoothing;
  // Too wide binomial stencils induce numeric problems
  // -> Smooth the data iteratively until the stencil gets an appropriate width
  if (aStillToSmooth >= 20) {
    CVector<float> aTemp(aVector);
    CGauss<float> aGauss(9,0);
    while (aStillToSmooth > 9) {
      NFilter::filter(aTemp,aGauss);
      aStillToSmooth -= 8;
    }
    CGauss<float> aDerivative((aStillToSmooth & 0xFFFFFFFE)+1,1);
    NFilter::filter(aTemp,aGradient,aDerivative);
  }
  // Fasfloat computation of derivatives for small stencils (no copying necessary)
  else {
    CGauss<float> aDerivative((aStillToSmooth & 0xFFFFFFFE)+1,1);
    NFilter::filter(aVector,aGradient,aDerivative);
  }
}

// presmoothedDerivative (2-D)
void CDiffusion::presmoothedDerivative(const CMatrix<float>& aMatrix, CMatrix<float>& aGradientX, CMatrix<float>& aGradientY) {
  int aStillToSmooth = 2*mSmoothing;
  // Too wide binomial stencils induce numeric problems
  // -> Smooth the data iteratively until the stencil gets a appropriate width
  if (aStillToSmooth >= 20) {
    CMatrix<float> aTemp(aMatrix);
    CGauss<float> aGauss(9,0);
    while (aStillToSmooth > 9) {
      NFilter::filter(aTemp,aGauss,aGauss);
      aStillToSmooth -= 8;
    }
    CGauss<float> aDerivative((aStillToSmooth & 0xFFFFFFFE)+1,1);
    NFilter::filter(aTemp,aGradientX,aDerivative,1);
    NFilter::filter(aTemp,aGradientY,1,aDerivative);
  }
  // Fasfloat computation of Derivatives for small stencils (no copying necessary)
  else {
    CGauss<float> aDerivative((aStillToSmooth & 0xFFFFFFFE)+1,1);
    NFilter::filter(aMatrix,aGradientX,aDerivative,1);
    NFilter::filter(aMatrix,aGradientY,1,aDerivative);
  }
}

// presmoothedDerivative (3-D)
void CDiffusion::presmoothedDerivative(const CTensor<float>& aTensor, CTensor<float>& aGradientX, CTensor<float>& aGradientY, CTensor<float>& aGradientZ) {
  int aStillToSmooth = 2*mSmoothing;
  // Too wide binomial stencils induce numeric problems
  // -> Smooth the data iteratively until the stencil gets a appropriate width
  if (aStillToSmooth >= 20) {
    CTensor<float> aTemp(aTensor);
    CGauss<float> aGauss(9,0);
    while (aStillToSmooth > 9) {
      NFilter::filter(aTemp,aGauss,aGauss,aGauss);
      aStillToSmooth -= 8;
    }
    CGauss<float> aDerivative((aStillToSmooth & 0xFFFFFFFE)+1,1);
    NFilter::filter(aTemp,aGradientX,aDerivative,1,1);
    NFilter::filter(aTemp,aGradientY,1,aDerivative,1);
    NFilter::filter(aTemp,aGradientZ,1,1,aDerivative);
  }
  // Fasfloat computation of Derivatives for small stencils (no copying necessary)
  else {
    CGauss<float> aDerivative((aStillToSmooth & 0xFFFFFFFE)+1,1);
    NFilter::filter(aTensor,aGradientX,aDerivative,1,1);
    NFilter::filter(aTensor,aGradientY,1,aDerivative,1);
    NFilter::filter(aTensor,aGradientZ,1,1,aDerivative);
  }
}

// Diffusivities ---------------------------------------------------------------

// isotropicDiffVector
void CDiffusion::isotropicDiffVector(const CVector<float>& aData, CVector<float>& aResult) {
  float invLambdaSqr = 1.0/(mLambda*mLambda);
  // (a) One-sided differences
  if (mSmoothing == 0)
    for (int i = 0; i < aData.size()-1; i++) {
      float aGrad = aData(i+1)-aData(i);
      aResult(i) = diffusivity(mDiffFunc,aGrad*aGrad,invLambdaSqr);
    }
  // (b) Central differences
  else {
    CVector<float> aGradient(aData.size());
    presmoothedDerivative(aData,aGradient);
    for (int i = 0; i < aData.size(); i++) {
      float aAbsolute = aGradient(i)*aGradient(i);
      aResult(i) = diffusivity(mDiffFunc,aAbsolute,invLambdaSqr);
    }
  }
}

// isotropicDiffMatrix (central differences)
void CDiffusion::isotropicDiffMatrix(const CMatrix<float>& aData, CMatrix<float>& aResult) {
  CMatrix<float> aGradientX(aData.xSize(),aData.ySize());
  CMatrix<float> aGradientY(aData.xSize(),aData.ySize());
  float invLambdaSqr = 1.0/(mLambda*mLambda);
  presmoothedDerivative(aData,aGradientX,aGradientY);
  for (int i = 0; i < aData.xSize()*aData.ySize(); i++) {
    float aAbsolute = aGradientX.data()[i]*aGradientX.data()[i]+aGradientY.data()[i]*aGradientY.data()[i];
    aResult.data()[i] = diffusivity(mDiffFunc,aAbsolute,invLambdaSqr);
  }
}

// isotropicDiffMatrix (one-sided differences)
void CDiffusion::isotropicDiffMatrix(const CMatrix<float>& aData, CMatrix<float>& aResultX, CMatrix<float>& aResultY) {
  float invLambdaSqr = 1.0/(mLambda*mLambda);
  CMatrix<float> aGrad(aData.xSize(),aData.ySize());
  CDerivative<float> aDerivative(3);
  // Diffusivities in x direction
  NFilter::filter(aData,aGrad,1,aDerivative);
  int i = 0;
  for (int y = 0; y < aData.ySize(); y++) {
    for (int x = 0; x < aData.xSize()-1; x++) {
      float aGradX = aData.data()[i+1]-aData.data()[i];
      float aGradY = 0.5*(aGrad.data()[i]+aGrad.data()[i+1]);
      float aAbsolute = aGradX*aGradX+aGradY*aGradY;
      aResultX.data()[i] = diffusivity(mDiffFunc,aAbsolute,invLambdaSqr);
      i++;
    }
    i++;
  }
  // Diffusivities in y direction
  NFilter::filter(aData,aGrad,aDerivative,1);
  i = 0;
  for (int y = 0; y < aData.ySize()-1; y++)
    for (int x = 0; x < aData.xSize(); x++) {
      float aGradX = 0.5*(aGrad.data()[i]+aGrad.data()[i+aData.xSize()]);
      float aGradY = aData.data()[i+aData.xSize()]-aData.data()[i];
      float aAbsolute = aGradX*aGradX+aGradY*aGradY;
      aResultY.data()[i] = diffusivity(mDiffFunc,aAbsolute,invLambdaSqr);
      i++;
    }
}

// isotropicDiffMatrixMultiChannel (central differences)
void CDiffusion::isotropicDiffMatrixMultiChannel(const CTensor<float>& aTensor, CMatrix<float>& aResult) {
  CMatrix<float> aGradientX(aTensor.xSize(),aTensor.ySize());
  CMatrix<float> aGradientY(aTensor.xSize(),aTensor.ySize());
  CMatrix<float> aTemp(aTensor.xSize(),aTensor.ySize());
  int aSize = aTensor.xSize()*aTensor.ySize();
  aResult = 0;
  // Compute sum of all channel gradients
  for (int k = 0; k < aTensor.zSize(); k++) {
    aTensor.getMatrix(aTemp,k);
    presmoothedDerivative(aTemp,aGradientX,aGradientY);
    for (int i = 0; i < aSize; i++)
      aResult.data()[i] += aGradientX.data()[i]*aGradientX.data()[i]+aGradientY.data()[i]*aGradientY.data()[i];
  }
  // Compute diffusivity
  float invLambdaSqr = 1.0/(mLambda*mLambda);
  for (int i = 0; i < aSize; i++)
    aResult.data()[i] = diffusivity(mDiffFunc,aResult.data()[i],invLambdaSqr);
}

// isotropicDiffMatrixMultiChannel (one-sided differences)
void CDiffusion::isotropicDiffMatrixMultiChannel(const CTensor<float>& aTensor, CMatrix<float>& aResultX, CMatrix<float>& aResultY) {
  CMatrix<float> aGrad(aTensor.xSize(),aTensor.ySize());
  CMatrix<float> aTemp(aTensor.xSize(),aTensor.ySize());
  CDerivative<float> aDerivative(3);
  aResultX = 0;
  aResultY = 0;
  // Compute sum of all channel gradients
  for (int k = 0; k < aTensor.zSize(); k++) {
    aTensor.getMatrix(aTemp,k);
    // Gradients in x direction
    NFilter::filter(aTemp,aGrad,1,aDerivative);
    int i = 0;
    for (int y = 0; y < aTemp.ySize(); y++) {
      for (int x = 0; x < aTemp.xSize()-1; x++) {
        float aGradX = aTemp.data()[i+1]-aTemp.data()[i];
        float aGradY = 0.5*(aGrad.data()[i]+aGrad.data()[i+1]);
        aResultX.data()[i] += aGradX*aGradX+aGradY*aGradY;
        i++;
      }
      i++;
    }
    // Gradients in y direction
    NFilter::filter(aTemp,aGrad,aDerivative,1);
    i = 0;
    for (int y = 0; y < aTemp.ySize()-1; y++)
      for (int x = 0; x < aTemp.xSize(); x++) {
        float aGradX = 0.5*(aGrad.data()[i]+aGrad.data()[i+aTemp.xSize()]);
        float aGradY = aTemp.data()[i+aTemp.xSize()]-aTemp.data()[i];
        aResultY.data()[i] += aGradX*aGradX+aGradY*aGradY;
        i++;
      }
  }
  // Compute diffusivity
  float invLambdaSqr = 1.0/(mLambda*mLambda);
  int aSize = aTensor.xSize()*aTensor.ySize();
  for (int i = 0; i < aSize; i++) {
    aResultX.data()[i] = diffusivity(mDiffFunc,aResultX.data()[i],invLambdaSqr);
    aResultY.data()[i] = diffusivity(mDiffFunc,aResultY.data()[i],invLambdaSqr);
  }
}

// anisotropicDiffMatrix
void CDiffusion::anisotropicDiffMatrix(const CMatrix<float>& aMatrix, CMatrix<float>& A, CMatrix<float>& B, CMatrix<float>& C) {
  CMatrix<float> aGradientX(aMatrix.xSize(),aMatrix.ySize());
  CMatrix<float> aGradientY(aMatrix.xSize(),aMatrix.ySize());
  presmoothedDerivative(aMatrix,aGradientX,aGradientY);
  float invLambdaSqr = 1.0/(mLambda*mLambda);
  for (int i = 0; i < aMatrix.xSize()*aMatrix.ySize(); i++) {
    float d2x = aGradientX.data()[i]*aGradientX.data()[i];
    float dxy = aGradientX.data()[i]*aGradientY.data()[i];
    float d2y = aGradientY.data()[i]*aGradientY.data()[i];
    float Absolute = d2x+d2y;
    float lambda;
    if (Absolute < 1.0e-35) {
      d2x = 1; d2y = 0; dxy = 0;
      lambda = 1;
    }
    else {
      float InvAbsolute = 1.0/Absolute;
      d2x = d2x*InvAbsolute;
      d2y = d2y*InvAbsolute;
      dxy = dxy*InvAbsolute;
      lambda = diffusivity(mDiffFunc,Absolute,invLambdaSqr);
      if (mLimitedCondNo) {
        if (lambda > 5.8284) lambda = 5.8284;
        else if (5.8284*lambda < 1.0) lambda = 1.001/5.8284;
      }
    }
    A.data()[i] = d2x*lambda+d2y;
    B.data()[i] = dxy*(lambda-1);
    C.data()[i] = d2y*lambda+d2x;
  }
}

// anisotropicDiffMatrixMultiChannel

void CDiffusion::anisotropicDiffMatrixMultiChannel(const CTensor<float>& aTensor, CMatrix<float>& A, CMatrix<float>& B, CMatrix<float>& C, float aRho) {
  CMatrix<float> aGradientX(aTensor.xSize(),aTensor.ySize());
  CMatrix<float> aGradientY(aTensor.xSize(),aTensor.ySize());
  CMatrix<float> aTemp(aTensor.xSize(),aTensor.ySize());
  CMatrix<float> ST(2,2);
  CVector<float> aEigenvalues(2);
  CMatrix<float> aEigenvectors(2,2);
  int aSize = aTensor.xSize()*aTensor.ySize();
  // Compute sum of all channel's structure tensors
  A = 0; B = 0; C = 0;
  for (int k = 0; k < aTensor.zSize(); k++) {
    aTensor.getMatrix(aTemp,k);
    presmoothedDerivative(aTemp,aGradientX,aGradientY);
    for (int i = 0; i < aSize; i++) {
      A.data()[i] += aGradientX.data()[i]*aGradientX.data()[i];
      B.data()[i] += aGradientX.data()[i]*aGradientY.data()[i];
      C.data()[i] += aGradientY.data()[i]*aGradientY.data()[i];
    }
  }
  // Smoothing for coherence enhancing diffusion
  if (aRho > 0) {
    if (aRho <= 2.0) {
      CSmooth<float> aGauss(aRho,3);
      NFilter::filter(A,aGauss,1);
      NFilter::filter(A,1,aGauss);
      NFilter::filter(B,aGauss,1);
      NFilter::filter(B,1,aGauss);
      NFilter::filter(C,aGauss,1);
      NFilter::filter(C,1,aGauss);
    }
    else {
      CMatrix<float> aTemp(A.xSize(),A.ySize());
      int aWidth = (int)(2.0*aRho);
      NFilter::boxFilterX(A,aTemp,aWidth);
      NFilter::boxFilterY(aTemp,A,aWidth);
      NFilter::boxFilterX(A,aTemp,aWidth);
      NFilter::boxFilterY(aTemp,A,aWidth);
      NFilter::boxFilterX(A,aTemp,aWidth);
      NFilter::boxFilterY(aTemp,A,aWidth);
      NFilter::boxFilterX(B,aTemp,aWidth);
      NFilter::boxFilterY(aTemp,B,aWidth);
      NFilter::boxFilterX(B,aTemp,aWidth);
      NFilter::boxFilterY(aTemp,B,aWidth);
      NFilter::boxFilterX(B,aTemp,aWidth);
      NFilter::boxFilterY(aTemp,B,aWidth);
      NFilter::boxFilterX(C,aTemp,aWidth);
      NFilter::boxFilterY(aTemp,C,aWidth);
      NFilter::boxFilterX(C,aTemp,aWidth);
      NFilter::boxFilterY(aTemp,C,aWidth);
      NFilter::boxFilterX(C,aTemp,aWidth);
      NFilter::boxFilterY(aTemp,C,aWidth);
    }
  }
  // Compute diffusion tensor
  float invLambdaSqr = 1.0/(mLambda*mLambda);
  for (int i = 0; i < aSize; i++) {
    ST(0,0) = A.data()[i];
    ST(1,1) = C.data()[i];
    ST(0,1) = ST(1,0) = B.data()[i];
    NMath::PATransformation(ST,aEigenvalues,aEigenvectors);
    aEigenvalues(0) = diffusivity(mDiffFunc,aEigenvalues(0),invLambdaSqr);
    if (aRho > 0) aEigenvalues(1) = 1.0;
    else aEigenvalues(1) = diffusivity(mDiffFunc,aEigenvalues(1),invLambdaSqr);
    if (mLimitedCondNo) {
      if (aEigenvalues(0) < 0.0001) aEigenvalues(0) = 0.0001;
      if (aEigenvalues(1) < 0.0001) aEigenvalues(1) = 0.0001;
      float aFactor = sqrt(aEigenvalues(1)/(aEigenvalues(0)*5.8284));
      aEigenvalues(1) /= aFactor;
      aEigenvalues(0) *= aFactor;
    }
    NMath::PABacktransformation(aEigenvectors,aEigenvalues,ST);
    A.data()[i] = ST(0,0);
    C.data()[i] = ST(1,1);
    B.data()[i] = ST(0,1);
  }
}

// isotropicDiffTensor (3-D) (central differences)
void CDiffusion::isotropicDiffTensor(const CTensor<float>& aTensor, CTensor<float>& aResult) {
  CTensor<float> aGradientX(aTensor.xSize(),aTensor.ySize(),aTensor.zSize());
  CTensor<float> aGradientY(aTensor.xSize(),aTensor.ySize(),aTensor.zSize());
  CTensor<float> aGradientZ(aTensor.xSize(),aTensor.ySize(),aTensor.zSize());
  float invLambdaSqr = 1.0/(mLambda*mLambda);
  presmoothedDerivative(aTensor,aGradientX,aGradientY,aGradientZ);
  int aSize = aTensor.xSize()*aTensor.ySize()*aTensor.zSize();
  for (int i = 0; i < aSize; i++) {
    float aAbsolute = aGradientX.data()[i]*aGradientX.data()[i]+
                  aGradientY.data()[i]*aGradientY.data()[i]+
                  aGradientZ.data()[i]*aGradientZ.data()[i];
    aResult.data()[i] = diffusivity(mDiffFunc,aAbsolute,invLambdaSqr);
  }
}

// isotropicDiffTensor (3-D) (one-sided differences)
void CDiffusion::isotropicDiffTensor(const CTensor<float>& aData, CTensor<float>& aResultX, CTensor<float>& aResultY, CTensor<float>& aResultZ) {
  float invLambdaSqr = 1.0/(mLambda*mLambda);
  CTensor<float> aGrad1(aData.xSize(),aData.ySize(),aData.zSize());
  CTensor<float> aGrad2(aData.xSize(),aData.ySize(),aData.zSize());
  CDerivative<float> aDerivative(3);
  // Diffusivities in x direction
  NFilter::filter(aData,aGrad1,1,aDerivative,1);
  NFilter::filter(aData,aGrad2,1,1,aDerivative);
  int i = 0;
  for (int z = 0; z < aData.zSize(); z++)
    for (int y = 0; y < aData.ySize(); y++) {
      for (int x = 0; x < aData.xSize()-1; x++) {
        float aGradX = aData.data()[i+1]-aData.data()[i];
        float aGradY = 0.5*(aGrad1.data()[i]+aGrad1.data()[i+1]);
        float aGradZ = 0.5*(aGrad2.data()[i]+aGrad2.data()[i+1]);
        float aAbsolute = aGradX*aGradX+aGradY*aGradY+aGradZ*aGradZ;
        aResultX.data()[i] = diffusivity(mDiffFunc,aAbsolute,invLambdaSqr);
        i++;
      }
      i++;
    }
  // Diffusivities in y direction
  NFilter::filter(aData,aGrad1,aDerivative,1,1);
  NFilter::filter(aData,aGrad2,1,1,aDerivative);
  i = 0;
  for (int z = 0; z < aData.zSize(); z++) {
    for (int y = 0; y < aData.ySize()-1; y++)
      for (int x = 0; x < aData.xSize(); x++) {
        float aGradX = 0.5*(aGrad1.data()[i]+aGrad1.data()[i+aData.xSize()]);
        float aGradY = aData.data()[i+aData.xSize()]-aData.data()[i];
        float aGradZ = 0.5*(aGrad2.data()[i]+aGrad2.data()[i+aData.xSize()]);
        float aAbsolute = aGradX*aGradX+aGradY*aGradY+aGradZ*aGradZ;
        aResultY.data()[i] = diffusivity(mDiffFunc,aAbsolute,invLambdaSqr);
        i++;
      }
    i += aData.xSize();
  }
  // Diffusivities in z direction
  NFilter::filter(aData,aGrad1,aDerivative,1,1);
  NFilter::filter(aData,aGrad2,1,aDerivative,1);
  i = 0;
  int aDisplacement = aData.xSize()*aData.ySize();
  for (int z = 0; z < aData.zSize()-1; z++)
    for (int y = 0; y < aData.ySize(); y++)
      for (int x = 0; x < aData.xSize(); x++) {
        float aGradX = 0.5*(aGrad1.data()[i]+aGrad1.data()[i+aDisplacement]);
        float aGradY = 0.5*(aGrad2.data()[i]+aGrad2.data()[i+aDisplacement]);
        float aGradZ = aData.data()[i+aDisplacement]-aData.data()[i];
        float aAbsolute = aGradX*aGradX+aGradY*aGradY+aGradZ*aGradZ;
        aResultZ.data()[i] = diffusivity(mDiffFunc,aAbsolute,invLambdaSqr);
        i++;
      }
}

// isotropicDiffTensorMultiChannel (3-D) (central differences)
void CDiffusion::isotropicDiffTensorMultiChannel(const CTensor4D<float>& aTensor, CTensor<float>& aResult) {
  CTensor<float> aGradientX(aTensor.xSize(),aTensor.ySize(),aTensor.zSize());
  CTensor<float> aGradientY(aTensor.xSize(),aTensor.ySize(),aTensor.zSize());
  CTensor<float> aGradientZ(aTensor.xSize(),aTensor.ySize(),aTensor.zSize());
  CTensor<float> aTemp(aTensor.xSize(),aTensor.ySize(),aTensor.zSize());
  int aSize = aTensor.xSize()*aTensor.ySize()*aTensor.zSize();
  aResult = 0;
  // Compute sum of all channel gradients
  for (int k = 0; k < aTensor.aSize(); k++) {
    aTensor.getTensor3D(aTemp,k);
    presmoothedDerivative(aTemp,aGradientX,aGradientY,aGradientZ);
    for (int i = 0; i < aSize; i++)
      aResult.data()[i] += aGradientX.data()[i]*aGradientX.data()[i]+aGradientY.data()[i]*aGradientY.data()[i]+aGradientZ.data()[i]*aGradientZ.data()[i];
  }
  // Compute diffusivity
  float invLambdaSqr = 1.0/(mLambda*mLambda);
  for (int i = 0; i < aSize; i++)
    aResult.data()[i] = diffusivity(mDiffFunc,aResult.data()[i],invLambdaSqr);
}

// isotropicDiffTensorMultiChannel (3-D) (one-sided differences)
void CDiffusion::isotropicDiffTensorMultiChannel(const CTensor4D<float>& aTensor, CTensor<float>& aResultX, CTensor<float>& aResultY, CTensor<float>& aResultZ) {
  CTensor<float> aGrad1(aTensor.xSize(),aTensor.ySize(),aTensor.zSize());
  CTensor<float> aGrad2(aTensor.xSize(),aTensor.ySize(),aTensor.zSize());
  CTensor<float> aGrad3(aTensor.xSize(),aTensor.ySize(),aTensor.zSize());
  CTensor<float> aTemp(aTensor.xSize(),aTensor.ySize(),aTensor.zSize());
  CDerivative<float> aDerivative(3);
  int aSize = aTensor.xSize()*aTensor.ySize()*aTensor.zSize();
  int a2DSize = aTensor.xSize()*aTensor.ySize();
  aResultX = 0;
  aResultY = 0;
  aResultZ = 0;
  // Compute sum of all channel gradients
  for (int k = 0; k < aTensor.aSize(); k++) {
    aTensor.getTensor3D(aTemp,k);
    // Gradients in x direction
    NFilter::filter(aTemp,aGrad1,aDerivative,1,1);
    NFilter::filter(aTemp,aGrad2,1,aDerivative,1);
    NFilter::filter(aTemp,aGrad3,1,1,aDerivative);
    int i = 0;
    for (int z = 0; z < aTensor.zSize(); z++)
      for (int y = 0; y < aTensor.ySize(); y++) {
        for (int x = 0; x < aTensor.xSize()-1; x++) {
          float aGradX = aTemp.data()[i+1]-aTemp.data()[i];
          float aGradY = 0.5*(aGrad2.data()[i]+aGrad2.data()[i+1]);
          float aGradZ = 0.5*(aGrad3.data()[i]+aGrad3.data()[i+1]);
          aResultX.data()[i] += aGradX*aGradX+aGradY*aGradY+aGradZ*aGradZ;
          i++;
        }
        i++;
      }
    // Gradients in y direction
    i = 0;
    for (int z = 0; z < aTensor.zSize(); z++) {
      for (int y = 0; y < aTensor.ySize()-1; y++)
        for (int x = 0; x < aTensor.xSize(); x++) {
          float aGradX = 0.5*(aGrad1.data()[i]+aGrad1.data()[i+aTensor.xSize()]);
          float aGradY = aTemp.data()[i+aTensor.xSize()]-aTemp.data()[i];
          float aGradZ = 0.5*(aGrad3.data()[i]+aGrad3.data()[i+aTensor.xSize()]);
          aResultY.data()[i] += aGradX*aGradX+aGradY*aGradY+aGradZ*aGradZ;
          i++;
        }
      i += aTensor.xSize();
    }
    // Gradients in z direction
    i = 0;
    for (int z = 0; z < aTensor.zSize()-1; z++)
      for (int j = 0; j < a2DSize; j++) {
        float aGradX = 0.5*(aGrad1.data()[i]+aGrad1.data()[i+a2DSize]);
        float aGradY = 0.5*(aGrad2.data()[i]+aGrad2.data()[i+a2DSize]);
        float aGradZ = aTemp.data()[i+a2DSize]-aTemp.data()[i];
        aResultZ.data()[i] += aGradX*aGradX+aGradY*aGradY+aGradZ*aGradZ;
        i++;
      }
  }
  // Compute diffusivity
  float invLambdaSqr = 1.0/(mLambda*mLambda);
  for (int i = 0; i < aSize; i++) {
    aResultX.data()[i] = diffusivity(mDiffFunc,aResultX.data()[i],invLambdaSqr);
    aResultY.data()[i] = diffusivity(mDiffFunc,aResultY.data()[i],invLambdaSqr);
    aResultZ.data()[i] = diffusivity(mDiffFunc,aResultZ.data()[i],invLambdaSqr);
  }
}

// anisotropicDiffTensor (3-D)
void CDiffusion::anisotropicDiffTensor(const CTensor<float>& aTensor, CTensor<float>& A, CTensor<float>& B, CTensor<float>& C, CTensor<float>& D, CTensor<float>& E, CTensor<float>& F) {
  float rxx = mTau3D*0.5;
  float ryy = mTau3D*0.5;
  float rzz = mTau3D*0.5;
  float rxy = mTau3D*0.25;
  float rxz = mTau3D*0.25;
  float ryz = mTau3D*0.25;
  CTensor<float> aGradientX(aTensor.xSize(),aTensor.ySize(),aTensor.zSize());
  CTensor<float> aGradientY(aTensor.xSize(),aTensor.ySize(),aTensor.zSize());
  CTensor<float> aGradientZ(aTensor.xSize(),aTensor.ySize(),aTensor.zSize());
  CMatrix<float> ST(3,3);
  CVector<float> aEigenvalues(3);
  CMatrix<float> aEigenvectors(3,3);
  presmoothedDerivative(aTensor,aGradientX,aGradientY,aGradientZ);
  float invLambdaSqr = 1.0/(mLambda*mLambda);
  int wholeSize = aTensor.xSize()*aTensor.ySize()*aTensor.zSize();
  for (int i = 0; i < wholeSize; i++) {
    ST(0,0) = aGradientX.data()[i]*aGradientX.data()[i];
    ST(1,1) = aGradientY.data()[i]*aGradientY.data()[i];
    ST(2,2) = aGradientZ.data()[i]*aGradientZ.data()[i];
    ST(0,1) = ST(1,0) = aGradientX.data()[i]*aGradientY.data()[i];
    ST(0,2) = ST(2,0) = aGradientX.data()[i]*aGradientZ.data()[i];
    ST(1,2) = ST(2,1) = aGradientY.data()[i]*aGradientZ.data()[i];
    NMath::PATransformation(ST,aEigenvalues,aEigenvectors);
    aEigenvalues(0) = diffusivity(mDiffFunc,aEigenvalues(0),invLambdaSqr);
    aEigenvalues(1) = diffusivity(mDiffFunc,aEigenvalues(1),invLambdaSqr);
    aEigenvalues(2) = diffusivity(mDiffFunc,aEigenvalues(2),invLambdaSqr);
    NMath::PABacktransformation(aEigenvectors,aEigenvalues,ST);
    A.data()[i] = rxx*ST(0,0);
    B.data()[i] = ryy*ST(1,1);
    C.data()[i] = rzz*ST(2,2);
    D.data()[i] = rxy*ST(0,1);
    E.data()[i] = rxz*ST(0,2);
    F.data()[i] = ryz*ST(1,2);
  }
}

// anisotropicDiffTensorMultiChannel (3-D)
void CDiffusion::anisotropicDiffTensorMultiChannel(const CTensor4D<float>& aTensor, CTensor<float>& A, CTensor<float>& B, CTensor<float>& C, CTensor<float>& D, CTensor<float>& E, CTensor<float>& F) {
  CTensor<float> aGradientX(aTensor.xSize(),aTensor.ySize(),aTensor.zSize());
  CTensor<float> aGradientY(aTensor.xSize(),aTensor.ySize(),aTensor.zSize());
  CTensor<float> aGradientZ(aTensor.xSize(),aTensor.ySize(),aTensor.zSize());
  CTensor<float> aTemp(aTensor.xSize(),aTensor.ySize(),aTensor.zSize());
  CMatrix<float> ST(3,3);
  CVector<float> aEigenvalues(3);
  CMatrix<float> aEigenvectors(3,3);
  int aSize = aTensor.xSize()*aTensor.ySize()*aTensor.zSize();
  // Compute sum of structure tensors of all channels
  A = 0; B = 0; C = 0; D = 0; E = 0; F = 0;
  for (int k = 0; k < aTensor.aSize(); k++) {
    aTensor.getTensor3D(aTemp,k);
    presmoothedDerivative(aTemp,aGradientX,aGradientY,aGradientZ);
    for (int i = 0; i < aSize; i++) {
      A.data()[i] += aGradientX.data()[i]*aGradientX.data()[i];
      B.data()[i] += aGradientY.data()[i]*aGradientY.data()[i];
      C.data()[i] += aGradientZ.data()[i]*aGradientZ.data()[i];
      D.data()[i] += aGradientX.data()[i]*aGradientY.data()[i];
      E.data()[i] += aGradientX.data()[i]*aGradientZ.data()[i];
      F.data()[i] += aGradientY.data()[i]*aGradientZ.data()[i];
    }
  }
  // Compute diffusion tensor
  float rxx = mTau3D*0.5;
  float ryy = mTau3D*0.5;
  float rzz = mTau3D*0.5;
  float rxy = mTau3D*0.25;
  float rxz = mTau3D*0.25;
  float ryz = mTau3D*0.25;
  float invLambdaSqr = 1.0/(mLambda*mLambda);
  for (int i = 0; i < aSize; i++) {
    ST(0,0) = A.data()[i];
    ST(1,1) = B.data()[i];
    ST(2,2) = C.data()[i];
    ST(0,1) = ST(1,0) = D.data()[i];
    ST(0,2) = ST(2,0) = E.data()[i];
    ST(1,2) = ST(2,1) = F.data()[i];
    NMath::PATransformation(ST,aEigenvalues,aEigenvectors);
    aEigenvalues(0) = diffusivity(mDiffFunc,aEigenvalues(0),invLambdaSqr);
    aEigenvalues(1) = diffusivity(mDiffFunc,aEigenvalues(1),invLambdaSqr);
    aEigenvalues(2) = diffusivity(mDiffFunc,aEigenvalues(2),invLambdaSqr);
    NMath::PABacktransformation(aEigenvectors,aEigenvalues,ST);
    A.data()[i] = rxx*ST(0,0);
    B.data()[i] = ryy*ST(1,1);
    C.data()[i] = rzz*ST(2,2);
    D.data()[i] = rxy*ST(0,1);
    E.data()[i] = rxz*ST(0,2);
    F.data()[i] = ryz*ST(1,2);
  }
}

// Diffusion, explicifloat scheme --------------------------------------------------

// isotropicDiffuse
void CDiffusion::isotropicDiffuse(CVector<float>& aVector, CVector<float>& aDiffusivity) {
  CVector<float> aDiff(aVector.size());
  if (mSmoothing == 0) {
    aDiff(0) = aDiffusivity(0)*(aVector(1)-aVector(0));
    // center
    for (int x = 1; x < aVector.size()-1; x++) {
      aDiff(x) =  aDiffusivity(x-1)*(aVector(x-1)-aVector(x));
      aDiff(x) += aDiffusivity(x)*(aVector(x+1)-aVector(x));
    }
    // righfloat rim
    int aIndex = aVector.size()-1;
    aDiff(aIndex) = aDiffusivity(aIndex-1)*(aVector(aIndex-1)-aVector(aIndex));
  }
  else {
    // leffloat rim
    aDiff(0) = 0.5*(aDiffusivity(0)+aDiffusivity(1))*(aVector(1)-aVector(0));
    // center
    for (int x = 1; x < aVector.size()-1; x++) {
      aDiff(x) =  (aDiffusivity(x)+aDiffusivity(x-1))*(aVector(x-1)-aVector(x));
      aDiff(x) += (aDiffusivity(x+1)+aDiffusivity(x))*(aVector(x+1)-aVector(x));
      aDiff(x) *= 0.5;
    }
    // righfloat rim
    int aIndex = aVector.size()-1;
    aDiff(aIndex) = 0.5*(aDiffusivity(aIndex-1)+aDiffusivity(aIndex))*(aVector(aIndex-1)-aVector(aIndex));
  }
  for (int x = 0; x < aVector.size(); x++)
    aVector(x) += mTau1D*aDiff(x);
}

// isotropicDiffuse (2-D)
void CDiffusion::isotropicDiffuse(CMatrix<float>& aMatrix, CMatrix<float>& aDiffusivity) {
  CMatrix<float> aDiff(aMatrix.xSize(),aMatrix.ySize());
  float aUpperDiff,aLowerDiff,aLeftDiff,aRightDiff;
  // Upper and lower rim
  for (int x = 0; x < aMatrix.xSize(); x++) {
    // upper rim
    if (x <= 0) aLeftDiff = 0;
    else aLeftDiff = (aDiffusivity(x,0)+aDiffusivity(x-1,0))*(aMatrix(x,0)-aMatrix(x-1,0));
    if (x+1 >= aMatrix.xSize()) aRightDiff = 0;
    else aRightDiff = (aDiffusivity(x+1,0)+aDiffusivity(x,0))*(aMatrix(x+1,0)-aMatrix(x,0));
    aLowerDiff = (aDiffusivity(x,1)+aDiffusivity(x,0))*(aMatrix(x,1)-aMatrix(x,0));
    aDiff(x,0) = 0.5*mTau2D*(aRightDiff-aLeftDiff+aLowerDiff);
    // lower rim
    int y = aMatrix.ySize()-1;
    if (x <= 0) aLeftDiff = 0;
    else aLeftDiff = (aDiffusivity(x,y)+aDiffusivity(x-1,y))*(aMatrix(x,y)-aMatrix(x-1,y));
    if (x+1 >= aMatrix.xSize()) aRightDiff = 0;
    else aRightDiff = (aDiffusivity(x+1,y)+aDiffusivity(x,y))*(aMatrix(x+1,y)-aMatrix(x,y));
    aUpperDiff = (aDiffusivity(x,y)+aDiffusivity(x,y-1))*(aMatrix(x,y)-aMatrix(x,y-1));
    aDiff(x,y) = 0.5*mTau2D*(aRightDiff-aLeftDiff-aUpperDiff);
  }
  // Left and righfloat rim
  for (int y = 1; y < aMatrix.ySize()-1; y++) {
    // leffloat rim
    aRightDiff = (aDiffusivity(1,y)+aDiffusivity(0,y))*(aMatrix(1,y)-aMatrix(0,y));
    aUpperDiff = (aDiffusivity(0,y)+aDiffusivity(0,y-1))*(aMatrix(0,y)-aMatrix(0,y-1));
    aLowerDiff = (aDiffusivity(0,y+1)+aDiffusivity(0,y))*(aMatrix(0,y+1)-aMatrix(0,y));
    aDiff(0,y) = 0.5*mTau2D*(aRightDiff+aLowerDiff-aUpperDiff);
    // righfloat rim
    int x = aMatrix.xSize()-1;
    aLeftDiff = (aDiffusivity(x,y)+aDiffusivity(x-1,y))*(aMatrix(x,y)-aMatrix(x-1,y));
    aUpperDiff = (aDiffusivity(x,y)+aDiffusivity(x,y-1))*(aMatrix(x,y)-aMatrix(x,y-1));
    aLowerDiff = (aDiffusivity(x,y+1)+aDiffusivity(x,y))*(aMatrix(x,y+1)-aMatrix(x,y));
    aDiff(x,y) = 0.5*mTau2D*(aLowerDiff-aLeftDiff-aUpperDiff);
  }
  // Center
  for (int y = 1; y < aMatrix.ySize()-1; y++)
    for (int x = 1; x < aMatrix.xSize()-1; x++) {
      aLeftDiff = (aDiffusivity(x,y)+aDiffusivity(x-1,y))*(aMatrix(x,y)-aMatrix(x-1,y));
      aRightDiff = (aDiffusivity(x+1,y)+aDiffusivity(x,y))*(aMatrix(x+1,y)-aMatrix(x,y));
      aUpperDiff = (aDiffusivity(x,y)+aDiffusivity(x,y-1))*(aMatrix(x,y)-aMatrix(x,y-1));
      aLowerDiff = (aDiffusivity(x,y+1)+aDiffusivity(x,y))*(aMatrix(x,y+1)-aMatrix(x,y));
      aDiff(x,y) = 0.5*mTau2D*(aRightDiff-aLeftDiff+aLowerDiff-aUpperDiff);
    }
  // Update image
  for (int i = 0; i < aMatrix.xSize()*aMatrix.ySize(); i++)
    aMatrix.data()[i] += aDiff.data()[i];
}

void CDiffusion::isotropicDiffuse(CMatrix<float>& aMatrix, CMatrix<float>& GX, CMatrix<float>& GY) {
  CMatrix<float> aDiff(aMatrix.xSize(),aMatrix.ySize(),0);
  for (int y = 0; y < aMatrix.ySize(); y++)
    for (int x = 0; x < aMatrix.xSize(); x++) {
      if (x > 0) aDiff(x,y) += GX(x-1,y)*(aMatrix(x-1,y)-aMatrix(x,y));
      if (y > 0) aDiff(x,y) += GY(x,y-1)*(aMatrix(x,y-1)-aMatrix(x,y));
      if (x < aMatrix.xSize()-1) aDiff(x,y) += GX(x,y)*(aMatrix(x+1,y)-aMatrix(x,y));
      if (y < aMatrix.ySize()-1) aDiff(x,y) += GY(x,y)*(aMatrix(x,y+1)-aMatrix(x,y));
    }
  // Update image
  for (int i = 0; i < aMatrix.xSize()*aMatrix.ySize(); i++)
    aMatrix.data()[i] += mTau2D*aDiff.data()[i];
}

// isotropicDiffuse (3-D)
void CDiffusion::isotropicDiffuse(CTensor<float>& aTensor, CTensor<float>& aDiffusivity) {
  CTensor<float> aDiff(aTensor.xSize(),aTensor.ySize(),aTensor.zSize());
  float aUpperDiff,aLowerDiff,aLeftDiff,aRightDiff,aFrontDiff,aBackDiff;
  // Upper and lower rim
  for (int z = 0; z < aTensor.zSize(); z++)
    for (int x = 0; x < aTensor.xSize(); x++) {
      // upper rim
      if (x <= 0) aLeftDiff = 0;
      else aLeftDiff = (aDiffusivity(x,0,z)+aDiffusivity(x-1,0,z))*(aTensor(x,0,z)-aTensor(x-1,0,z));
      if (x+1 >= aTensor.xSize()) aRightDiff = 0;
      else aRightDiff = (aDiffusivity(x+1,0,z)+aDiffusivity(x,0,z))*(aTensor(x+1,0,z)-aTensor(x,0,z));
      aLowerDiff = (aDiffusivity(x,1,z)+aDiffusivity(x,0,z))*(aTensor(x,1,z)-aTensor(x,0,z));
      if (z <= 0) aFrontDiff = 0;
      else aFrontDiff = (aDiffusivity(x,0,z)+aDiffusivity(x,0,z-1))*(aTensor(x,0,z)-aTensor(x,0,z-1));
      if (z+1 >= aTensor.zSize()) aBackDiff = 0;
      else aBackDiff = (aDiffusivity(x,0,z+1)+aDiffusivity(x,0,z))*(aTensor(x,0,z+1)-aTensor(x,0,z));
      aDiff(x,0,z) = 0.5*mTau3D*(aRightDiff-aLeftDiff+aLowerDiff+aBackDiff-aFrontDiff);
      // lower rim
      int y = aTensor.ySize()-1;
      if (x <= 0) aLeftDiff = 0;
      else aLeftDiff = (aDiffusivity(x,y,z)+aDiffusivity(x-1,y,z))*(aTensor(x,y,z)-aTensor(x-1,y,z));
      if (x+1 >= aTensor.xSize()) aRightDiff = 0;
      else aRightDiff = (aDiffusivity(x+1,y,z)+aDiffusivity(x,y,z))*(aTensor(x+1,y,z)-aTensor(x,y,z));
      aUpperDiff = (aDiffusivity(x,y,z)+aDiffusivity(x,y-1,z))*(aTensor(x,y,z)-aTensor(x,y-1,z));
      if (z <= 0) aFrontDiff = 0;
      else aFrontDiff = (aDiffusivity(x,y,z)+aDiffusivity(x,y,z-1))*(aTensor(x,y,z)-aTensor(x,y,z-1));
      if (z+1 >= aTensor.zSize()) aBackDiff = 0;
      else aBackDiff = (aDiffusivity(x,y,z+1)+aDiffusivity(x,y,z))*(aTensor(x,y,z+1)-aTensor(x,y,z));
      aDiff(x,y,z) = 0.5*mTau3D*(aRightDiff-aLeftDiff-aUpperDiff+aBackDiff-aFrontDiff);
    }
  // Left and righfloat rim
  for (int z = 0; z < aTensor.zSize(); z++)
    for (int y = 1; y < aTensor.ySize()-1; y++) {
      // leffloat rim
      aRightDiff = (aDiffusivity(1,y,z)+aDiffusivity(0,y,z))*(aTensor(1,y,z)-aTensor(0,y,z));
      aUpperDiff = (aDiffusivity(0,y,z)+aDiffusivity(0,y-1,z))*(aTensor(0,y,z)-aTensor(0,y-1,z));
      aLowerDiff = (aDiffusivity(0,y+1,z)+aDiffusivity(0,y,z))*(aTensor(0,y+1,z)-aTensor(0,y,z));
      if (z <= 0) aFrontDiff = 0;
      else aFrontDiff = (aDiffusivity(0,y,z)+aDiffusivity(0,y,z-1))*(aTensor(0,y,z)-aTensor(0,y,z-1));
      if (z+1 >= aTensor.zSize()) aBackDiff = 0;
      else aBackDiff = (aDiffusivity(0,y,z+1)+aDiffusivity(0,y,z))*(aTensor(0,y,z+1)-aTensor(0,y,z));
      aDiff(0,y,z) = 0.5*mTau3D*(aRightDiff+aLowerDiff-aUpperDiff+aBackDiff-aFrontDiff);
      // righfloat rim
      int x = aTensor.xSize()-1;
      aLeftDiff = (aDiffusivity(x,y,z)+aDiffusivity(x-1,y,z))*(aTensor(x,y,z)-aTensor(x-1,y,z));
      aUpperDiff = (aDiffusivity(x,y,z)+aDiffusivity(x,y-1,z))*(aTensor(x,y,z)-aTensor(x,y-1,z));
      aLowerDiff = (aDiffusivity(x,y+1,z)+aDiffusivity(x,y,z))*(aTensor(x,y+1,z)-aTensor(x,y,z));
      if (z <= 0) aFrontDiff = 0;
      else aFrontDiff = (aDiffusivity(x,y,z)+aDiffusivity(x,y,z-1))*(aTensor(x,y,z)-aTensor(x,y,z-1));
      if (z+1 >= aTensor.zSize()) aBackDiff = 0;
      else aBackDiff = (aDiffusivity(x,y,z+1)+aDiffusivity(x,y,z))*(aTensor(x,y,z+1)-aTensor(x,y,z));
      aDiff(x,y,z) = 0.5*mTau3D*(aLowerDiff-aLeftDiff-aUpperDiff+aBackDiff-aFrontDiff);
    }
  // Fronfloat and back rim
  for (int y = 1; y < aTensor.ySize()-1; y++)
    for (int x = 1; x < aTensor.xSize()-1; x++) {
      // fronfloat rim
      aLeftDiff = (aDiffusivity(x,y,0)+aDiffusivity(x-1,y,0))*(aTensor(x,y,0)-aTensor(x-1,y,0));
      aRightDiff = (aDiffusivity(x+1,y,0)+aDiffusivity(x,y,0))*(aTensor(x+1,y,0)-aTensor(x,y,0));
      aUpperDiff = (aDiffusivity(x,y,0)+aDiffusivity(x,y-1,0))*(aTensor(x,y,0)-aTensor(x,y-1,0));
      aLowerDiff = (aDiffusivity(x,y+1,0)+aDiffusivity(x,y,0))*(aTensor(x,y+1,0)-aTensor(x,y,0));
      aBackDiff = (aDiffusivity(x,y,1)+aDiffusivity(x,y,0))*(aTensor(x,y,1)-aTensor(x,y,0));
      aDiff(x,y,0) = 0.5*mTau3D*(aRightDiff-aLeftDiff+aLowerDiff-aUpperDiff+aBackDiff);
      // back rim
      int z = aTensor.zSize()-1;
      aLeftDiff = (aDiffusivity(x,y,z)+aDiffusivity(x-1,y,z))*(aTensor(x,y,z)-aTensor(x-1,y,z));
      aRightDiff = (aDiffusivity(x+1,y,z)+aDiffusivity(x,y,z))*(aTensor(x+1,y,z)-aTensor(x,y,z));
      aUpperDiff = (aDiffusivity(x,y,z)+aDiffusivity(x,y-1,z))*(aTensor(x,y,z)-aTensor(x,y-1,z));
      aLowerDiff = (aDiffusivity(x,y+1,z)+aDiffusivity(x,y,z))*(aTensor(x,y+1,z)-aTensor(x,y,z));
      aFrontDiff = (aDiffusivity(x,y,z)+aDiffusivity(x,y,z-1))*(aTensor(x,y,z)-aTensor(x,y,z-1));
      aDiff(x,y,z) = 0.5*mTau3D*(aRightDiff-aLeftDiff+aLowerDiff-aUpperDiff-aFrontDiff);
    }
  // Center
  for (int z = 1; z < aTensor.zSize()-1; z++)
    for (int y = 1; y < aTensor.ySize()-1; y++)
      for (int x = 1; x < aTensor.xSize()-1; x++) {
        aLeftDiff = (aDiffusivity(x,y,z)+aDiffusivity(x-1,y,z))*(aTensor(x,y,z)-aTensor(x-1,y,z));
        aRightDiff = (aDiffusivity(x+1,y,z)+aDiffusivity(x,y,z))*(aTensor(x+1,y,z)-aTensor(x,y,z));
        aUpperDiff = (aDiffusivity(x,y,z)+aDiffusivity(x,y-1,z))*(aTensor(x,y,z)-aTensor(x,y-1,z));
        aLowerDiff = (aDiffusivity(x,y+1,z)+aDiffusivity(x,y,z))*(aTensor(x,y+1,z)-aTensor(x,y,z));
        aFrontDiff = (aDiffusivity(x,y,z)+aDiffusivity(x,y,z-1))*(aTensor(x,y,z)-aTensor(x,y,z-1));
        aBackDiff = (aDiffusivity(x,y,z+1)+aDiffusivity(x,y,z))*(aTensor(x,y,z+1)-aTensor(x,y,z));
        aDiff(x,y,z) = 0.5*mTau3D*(aRightDiff-aLeftDiff+aLowerDiff-aUpperDiff+aBackDiff-aFrontDiff);
      }
  // Update image
  int wholeSize = aTensor.xSize()*aTensor.ySize()*aTensor.zSize();
  for (int i = 0; i < wholeSize; i++)
    aTensor.data()[i] += aDiff.data()[i];
}


// Anisotropic -----------------------------------------------------------------

// anisotropicDiffuse
void CDiffusion::anisotropicDiffuse(CMatrix<float>& aMatrix, CMatrix<float>& A, CMatrix<float>& B, CMatrix<float>& C) {
  CMatrix<float> aDiff(aMatrix.xSize(),aMatrix.ySize());
  // Upper and lower rim
  for (int x = 0; x < aMatrix.xSize(); x++) {
    // upper rim
    float diff = 0;
    if (x > 0) diff += (A(x-1,0) + A(x,0)) * (aMatrix(x-1,0) - aMatrix(x,0));
    if (x+1 < aMatrix.xSize()) diff += (A(x+1,0) + A(x,0)) * (aMatrix(x+1,0) - aMatrix(x,0));
    diff += (C(x,1) + C(x,0)) * (aMatrix(x,1) - aMatrix(x,0));
    if (x+1 < aMatrix.xSize()) diff += 0.5*(B(x+1,1) + B(x,0)) * (aMatrix(x+1,1) - aMatrix(x,0));
    if (x > 0) diff -= 0.5*(B(x-1,1) + B(x,0)) * (aMatrix(x-1,1) - aMatrix(x,0));
    aDiff(x,0) = 0.5*mTau2D*diff;
    // lower rim
    int y = aMatrix.ySize()-1;
    diff = 0;
    if (x > 0) diff += (A(x-1,y) + A(x,y)) * (aMatrix(x-1,y) - aMatrix(x,y));
    if (x+1 < aMatrix.xSize()) diff += (A(x+1,y) + A(x,y)) * (aMatrix(x+1,y) - aMatrix(x,y));
    diff += (C(x,y-1) + C(x,y)) * (aMatrix(x,y-1) - aMatrix(x,y));
    if (x > 0) diff += 0.5*(B(x-1,y-1) + B(x,y)) * (aMatrix(x-1,y-1) - aMatrix(x,y));
    if (x+1 < aMatrix.xSize()) diff -= 0.5*(B(x+1,y-1) + B(x,y)) * (aMatrix(x+1,y-1) - aMatrix(x,y));
    aDiff(x,y) = 0.5*mTau2D*diff;
  }
  // Left and righfloat rim
  for (int y = 1; y < aMatrix.ySize()-1; y++) {
    // leffloat rim
    float diff;
    diff  = (A(1,y) + A(0,y)) * (aMatrix(1,y) - aMatrix(0,y));
    diff += (C(0,y-1) + C(0,y)) * (aMatrix(0,y-1) - aMatrix(0,y));
    diff += (C(0,y+1) + C(0,y)) * (aMatrix(0,y+1) - aMatrix(0,y));
    diff += 0.5*(B(1,y+1) + B(0,y)) * (aMatrix(1,y+1) - aMatrix(0,y));
    diff -= 0.5*(B(1,y-1) + B(0,y)) * (aMatrix(1,y-1) - aMatrix(0,y));
    aDiff(0,y) = 0.5*mTau2D*diff;
    // righfloat rim
    int x = aMatrix.xSize()-1;
    diff  = (A(x-1,y) + A(x,y)) * (aMatrix(x-1,y) - aMatrix(x,y));
    diff += (C(x,y-1) + C(x,y)) * (aMatrix(x,y-1) - aMatrix(x,y));
    diff += (C(x,y+1) + C(x,y)) * (aMatrix(x,y+1) - aMatrix(x,y));
    diff += 0.5*(B(x-1,y-1) + B(x,y)) * (aMatrix(x-1,y-1) - aMatrix(x,y));
    diff -= 0.5*(B(x-1,y+1) + B(x,y)) * (aMatrix(x-1,y+1) - aMatrix(x,y));
    aDiff(x,y) = 0.5*mTau2D*diff;
  }
  // Center
  for (int y = 1; y < aMatrix.ySize()-1; y++)
    for (int x = 1; x < aMatrix.xSize()-1; x++) {
      float diff;
      diff  = (A(x-1,y) + A(x,y)) * (aMatrix(x-1,y) - aMatrix(x,y));
      diff += (A(x+1,y) + A(x,y)) * (aMatrix(x+1,y) - aMatrix(x,y));
      diff += (C(x,y-1) + C(x,y)) * (aMatrix(x,y-1) - aMatrix(x,y));
      diff += (C(x,y+1) + C(x,y)) * (aMatrix(x,y+1) - aMatrix(x,y));
      diff += 0.5*(B(x+1,y+1) + B(x,y)) * (aMatrix(x+1,y+1));
      diff += 0.5*(B(x-1,y-1) + B(x,y)) * (aMatrix(x-1,y-1));
      diff -= 0.5*(B(x+1,y-1) + B(x,y)) * (aMatrix(x+1,y-1));
      diff -= 0.5*(B(x-1,y+1) + B(x,y)) * (aMatrix(x-1,y+1));
      aDiff(x,y) = 0.5*mTau2D*diff;
    }
  // Update image
  for (int i = 0; i < aMatrix.xSize()*aMatrix.ySize(); i++)
    aMatrix.data()[i] += aDiff.data()[i];
}

// anisotropicEnhancedDiffuse
void CDiffusion::anisotropicEnhancedDiffuse(CMatrix<float>& aMatrix, CMatrix<float>& A, CMatrix<float>& B, CMatrix<float>& C) {
  CMatrix<float> aDiff(aMatrix.xSize(),aMatrix.ySize());
  // Upper and lower rim
  for (int x = 0; x < aMatrix.xSize(); x++) {
    // upper rim
    float diff = 0;
    if (x > 0) diff += (A(x-1,0)+A(x,0)-fabs(B(x-1,0))-fabs(B(x,0))) * (aMatrix(x-1,0) - aMatrix(x,0));
    if (x+1 < aMatrix.xSize()) diff += (A(x+1,0)+A(x,0)-fabs(B(x+1,0))-fabs(B(x,0))) * (aMatrix(x+1,0) - aMatrix(x,0));
    diff += (C(x,1)+C(x,0)-fabs(B(x,1))-fabs(B(x,0))) * (aMatrix(x,1) - aMatrix(x,0));
    if (x+1 < aMatrix.xSize()) diff += 0.5*(fabs(B(x+1,1))+B(x+1,1)+fabs(B(x,0))+B(x,0)) * (aMatrix(x+1,1) - aMatrix(x,0));
    if (x > 0) diff += 0.5*(fabs(B(x-1,1))-B(x-1,1)+fabs(B(x,0))-B(x,0)) * (aMatrix(x-1,1) - aMatrix(x,0));
    aDiff(x,0) = 0.5*mTau2D*diff;
    // lower rim
    int y = aMatrix.ySize()-1;
    diff = 0;
    if (x > 0) diff += (A(x-1,y)+A(x,y)-fabs(B(x-1,y))-fabs(B(x,y))) * (aMatrix(x-1,y) - aMatrix(x,y));
    if (x+1 < aMatrix.xSize()) diff += (A(x+1,y)+A(x,y)-fabs(B(x+1,y))-fabs(B(x,y))) * (aMatrix(x+1,y) - aMatrix(x,y));
    diff += (C(x,y-1)+C(x,y)-fabs(B(x,y-1))-fabs(B(x,y))) * (aMatrix(x,y-1) - aMatrix(x,y));
    if (x > 0) diff += 0.5*(fabs(B(x-1,y-1))+B(x-1,y-1)+fabs(B(x,y))+B(x,y)) * (aMatrix(x-1,y-1) - aMatrix(x,y));
    if (x+1 < aMatrix.xSize()) diff += 0.5*(fabs(B(x+1,y-1))-B(x+1,y-1)+fabs(B(x,y))-B(x,y)) * (aMatrix(x+1,y-1) - aMatrix(x,y));
    aDiff(x,y) = 0.5*mTau2D*diff;
  }
  // Left and righfloat rim
  for (int y = 1; y < aMatrix.ySize()-1; y++) {
    // leffloat rim
    float diff;
    diff  = (A(1,y)+A(0,y)-fabs(B(1,y))-fabs(B(0,y))) * (aMatrix(1,y) - aMatrix(0,y));
    diff += (C(0,y-1)+C(0,y)-fabs(B(0,y-1))-fabs(B(0,y))) * (aMatrix(0,y-1) - aMatrix(0,y));
    diff += (C(0,y+1)+C(0,y)-fabs(B(0,y+1))-fabs(B(0,y))) * (aMatrix(0,y+1) - aMatrix(0,y));
    diff += 0.5*(fabs(B(1,y+1))+B(1,y+1)+fabs(B(0,y))+B(0,y)) * (aMatrix(1,y+1) - aMatrix(0,y));
    diff += 0.5*(fabs(B(1,y-1))-B(1,y-1)+fabs(B(0,y))-B(0,y)) * (aMatrix(1,y-1) - aMatrix(0,y));
    aDiff(0,y) = 0.5*mTau2D*diff;
    // righfloat rim
    int x = aMatrix.xSize()-1;
    diff  = (A(x-1,y)+A(x,y)-fabs(B(x-1,y))-fabs(B(x,y))) * (aMatrix(x-1,y) - aMatrix(x,y));
    diff += (C(x,y-1)+C(x,y)-fabs(B(x,y-1))-fabs(B(x,y))) * (aMatrix(x,y-1) - aMatrix(x,y));
    diff += (C(x,y+1)+C(x,y)-fabs(B(x,y+1))-fabs(B(x,y))) * (aMatrix(x,y+1) - aMatrix(x,y));
    diff += 0.5*(fabs(B(x-1,y-1))+B(x-1,y-1)+fabs(B(x,y))+B(x,y)) * (aMatrix(x-1,y-1) - aMatrix(x,y));
    diff += 0.5*(fabs(B(x-1,y+1))-B(x-1,y+1)+fabs(B(x,y))-B(x,y)) * (aMatrix(x-1,y+1) - aMatrix(x,y));
    aDiff(x,y) = 0.5*mTau2D*diff;
  }
  // Center
  for (int y = 1; y < aMatrix.ySize()-1; y++)
    for (int x = 1; x < aMatrix.xSize()-1; x++) {
      float diff;
      float fabsB = fabs(B(x,y));
      diff  = (A(x-1,y)+A(x,y)-fabs(B(x-1,y))-fabsB) * (aMatrix(x-1,y) - aMatrix(x,y));
      diff += (A(x+1,y)+A(x,y)-fabs(B(x+1,y))-fabsB) * (aMatrix(x+1,y) - aMatrix(x,y));
      diff += (C(x,y-1)+C(x,y)-fabs(B(x,y-1))-fabsB) * (aMatrix(x,y-1) - aMatrix(x,y));
      diff += (C(x,y+1)+C(x,y)-fabs(B(x,y+1))-fabsB) * (aMatrix(x,y+1) - aMatrix(x,y));
      diff += 0.5*(fabs(B(x+1,y+1))+B(x+1,y+1)+fabsB+B(x,y)) * (aMatrix(x+1,y+1) - aMatrix(x,y));
      diff += 0.5*(fabs(B(x-1,y-1))+B(x-1,y-1)+fabsB+B(x,y)) * (aMatrix(x-1,y-1) - aMatrix(x,y));
      diff += 0.5*(fabs(B(x+1,y-1))-B(x+1,y-1)+fabsB-B(x,y)) * (aMatrix(x+1,y-1) - aMatrix(x,y));
      diff += 0.5*(fabs(B(x-1,y+1))-B(x-1,y+1)+fabsB-B(x,y)) * (aMatrix(x-1,y+1) - aMatrix(x,y));
      aDiff(x,y) = 0.5*mTau2D*diff;
    }
  // Update image
  for (int i = 0; i < aMatrix.xSize()*aMatrix.ySize(); i++)
    aMatrix.data()[i] += aDiff.data()[i];
}

// anisotropicDiffuse (3-D)
void CDiffusion::anisotropicDiffuse(CTensor<float>& aTensor, CTensor<float>& A, CTensor<float>& B, CTensor<float>& C, CTensor<float>& D, CTensor<float>& E, CTensor<float>& F) {
  CTensor<float> aDiff(aTensor.xSize(),aTensor.ySize(),aTensor.zSize());
  for (int z = 0; z < aTensor.zSize(); z++)
    for (int y = 0; y < aTensor.ySize(); y++)
      for (int x = 0; x < aTensor.xSize(); x++) {
        float diff = 0;
        if (x+1 < aTensor.xSize())                          diff += (A(x+1,y,z) + A(x,y,z)) * (aTensor(x+1,y,z) - aTensor(x,y,z));
        if (x > 0)                                          diff += (A(x-1,y,z) + A(x,y,z)) * (aTensor(x-1,y,z) - aTensor(x,y,z));
        if (y+1 < aTensor.ySize())                          diff += (B(x,y+1,z) + B(x,y,z)) * (aTensor(x,y+1,z) - aTensor(x,y,z));
        if (y > 0)                                          diff += (B(x,y-1,z) + B(x,y,z)) * (aTensor(x,y-1,z) - aTensor(x,y,z));
        if (z+1 < aTensor.zSize())                          diff += (C(x,y,z+1) + C(x,y,z)) * (aTensor(x,y,z+1) - aTensor(x,y,z));
        if (z > 0)                                          diff += (C(x,y,z-1) + C(x,y,z)) * (aTensor(x,y,z-1) - aTensor(x,y,z));
        if (x+1 < aTensor.xSize() && y+1 < aTensor.ySize()) diff += (D(x+1,y+1,z) + D(x,y,z)) * (aTensor(x+1,y+1,z) - aTensor(x,y,z));
        if (x > 0 && y > 0)                                 diff += (D(x-1,y-1,z) + D(x,y,z)) * (aTensor(x-1,y-1,z) - aTensor(x,y,z));
        if (x+1 < aTensor.xSize() && y > 0)                 diff -= (D(x+1,y-1,z) + D(x,y,z)) * (aTensor(x+1,y-1,z) - aTensor(x,y,z));
        if (x > 0 && y+1 < aTensor.ySize())                 diff -= (D(x-1,y+1,z) + D(x,y,z)) * (aTensor(x-1,y+1,z) - aTensor(x,y,z));
        if (x+1 < aTensor.xSize() && z+1 < aTensor.zSize()) diff += (E(x+1,y,z+1) + E(x,y,z)) * (aTensor(x+1,y,z+1) - aTensor(x,y,z));
        if (x > 0 && z > 0)                                 diff += (E(x-1,y,z-1) + E(x,y,z)) * (aTensor(x-1,y,z-1) - aTensor(x,y,z));
        if (x+1 < aTensor.xSize() && z > 0)                 diff -= (E(x+1,y,z-1) + E(x,y,z)) * (aTensor(x+1,y,z-1) - aTensor(x,y,z));
        if (x > 0 && z+1 < aTensor.zSize())                 diff -= (E(x-1,y,z+1) + E(x,y,z)) * (aTensor(x-1,y,z+1) - aTensor(x,y,z));
        if (y+1 < aTensor.ySize() && z+1 < aTensor.zSize()) diff += (F(x,y+1,z+1) + F(x,y,z)) * (aTensor(x,y+1,z+1) - aTensor(x,y,z));
        if (y > 0 && z > 0)                                 diff += (F(x,y-1,z-1) + F(x,y,z)) * (aTensor(x,y-1,z-1) - aTensor(x,y,z));
        if (y+1 < aTensor.ySize() && z > 0)                 diff -= (F(x,y+1,z-1) + F(x,y,z)) * (aTensor(x,y+1,z-1) - aTensor(x,y,z));
        if (y > 0 && z+1 < aTensor.zSize())                 diff -= (F(x,y-1,z+1) + F(x,y,z)) * (aTensor(x,y-1,z+1) - aTensor(x,y,z));
        aDiff(x,y,z) = diff;
      }
  // Update image
  int wholeSize = aTensor.xSize()*aTensor.ySize()*aTensor.zSize();
  for (int i = 0; i < wholeSize; i++)
    aTensor.data()[i] += aDiff.data()[i];
}

// anisotropicDiffuse (3-D)
void CDiffusion::anisotropicEnhancedDiffuse(CTensor<float>& aTensor, CTensor<float>& A, CTensor<float>& B, CTensor<float>& C, CTensor<float>& D, CTensor<float>& E, CTensor<float>& F) {
  anisotropicDiffuse(aTensor,A,B,C,D,E,F);
}

// Semi-implicifloat AOS -----------------------------------------------------------

// isotropicDiffuseSemiImplicifloat (1-D)
void CDiffusion::isotropicDiffuseSemiImplicit(CVector<float>& aVector, const CVector<float>& G) {
  CVector<float> M(aVector.size());
  CVector<float> Y(aVector.size());
  CVector<float> I(aVector.size());
  CVector<float> J(aVector.size());
  CVector<float> K(aVector.size());
  int aSize = aVector.size();
  // Compute diffusion
  // (a) One-sided differences
  if (mSmoothing == 0) {
    // alpha -> I   beta -> J    gamma -> K
    float aTemp1 = mTau1D*G(0);
    float aTemp2;
    I(0) = 1.0+aTemp1;
    J(0) = -aTemp1;
    for (int x = 1; x+1 < aSize; x++) {
      aTemp1 = mTau1D*G(x);
      aTemp2 = mTau1D*G(x-1);
      I(x) = 1+aTemp1+aTemp2;
      J(x) = -aTemp1;
      K(x-1) = -aTemp2;
    }
    aTemp2 = mTau1D*G(aSize-2);
    I(aSize-1) = 1.0+aTemp2;
    J(aSize-1) = 0;
    K(aSize-2) = -aTemp2;
  }
  // (b) Central differences
  else {
    float aTau = 0.5*mTau1D;
     // alpha -> I   beta -> J    gamma -> K
    float aTemp1 = aTau*(G(0)+G(1));
    float aTemp2;
    I(0) = 1.0+aTemp1;
    J(0) = -aTemp1;
    for (int x = 1; x+1 < aSize; x++) {
      aTemp1 = aTau*(G(x)+G(x+1));
      aTemp2 = aTau*(G(x)+G(x-1));
      I(x) = 1+aTemp1+aTemp2;
      J(x) = -aTemp1;
      K(x-1) = -aTemp2;
    }
    aTemp2 = aTau*(G(aSize-1)+G(aSize-2));
    I(aSize-1) = 1.0+aTemp2;
    J(aSize-1) = 0;
    K(aSize-2) = -aTemp2;
  }
  // Thomas algorithm
  // Forward substitution
  M(0) = 1.0/I(0);
  Y(0) = aVector(0);
  int i = 0;
  do {
    float l = K(i)*M(i);
    i++;
    M(i) = 1.0/(I(i)-l*J(i-1));
    Y(i) = aVector(i)-l*Y(i-1);
  }
  while (i < aSize-1);
  // Backward substitution
  aVector(aSize-1) = Y(aSize-1)*M(aSize-1);
  for (int j = aSize-2; j >= 0; j--)
    aVector(j) = (Y(j)-J(j)*aVector(j+1))*M(j);
}

// isotropicDiffuseSemiImplicifloat (2-D)
void CDiffusion::isotropicDiffuseSemiImplicit(CMatrix<float>& aMatrix, const CMatrix<float>& GX, const CMatrix<float>& GY) {
  CMatrix<float> aResult(aMatrix.xSize(),aMatrix.ySize());
  CMatrix<float> M(aMatrix.xSize(),aMatrix.ySize());
  CMatrix<float> Y(aMatrix.xSize(),aMatrix.ySize());
  CMatrix<float> I(aMatrix.xSize(),aMatrix.ySize());
  CMatrix<float> J(aMatrix.xSize(),aMatrix.ySize());
  CMatrix<float> K(aMatrix.xSize(),aMatrix.ySize());
  int aSize = aMatrix.xSize()*aMatrix.ySize();
  float aTau = mTau2D;
  if (mSmoothing == 0) aTau *= 2.0;
  // Compute diffusion in x direction
  // alpha -> I   beta -> J    gamma -> K
  int i = 0;
  if (mSmoothing == 0) {
    for (int y = 0; y < aMatrix.ySize(); y++) {
      float aTemp1 = aTau*GX.data()[i];
      float aTemp2;
      I.data()[i] = 1+aTemp1;
      J.data()[i] = -aTemp1;
      if (i > 0) K.data()[i-1] = 0;
      i++;
      for (int x = 1; x+1 < aMatrix.xSize(); x++) {
        aTemp1 = aTau*GX.data()[i];
        aTemp2 = aTau*GX.data()[i-1];
        I.data()[i] = 1+aTemp1+aTemp2;
        J.data()[i] = -aTemp1;
        K.data()[i-1] = -aTemp2;
        i++;
      }
      aTemp2 = aTau*GX.data()[i-1];
      I.data()[i] = 1+aTemp2;
      J.data()[i] = 0;
      K.data()[i-1] = -aTemp2;
      i++;
    }
  }
  else {
    for (int y = 0; y < aMatrix.ySize(); y++) {
      float aTemp1 = aTau*(GX.data()[i]+GX.data()[i+1]);
      float aTemp2;
      I.data()[i] = 1+aTemp1;
      J.data()[i] = -aTemp1;
      if (i > 0) K.data()[i-1] = 0;
      i++;
      for (int x = 1; x+1 < aMatrix.xSize(); x++) {
        aTemp1 = aTau*(GX.data()[i]+GX.data()[i+1]);
        aTemp2 = aTau*(GX.data()[i]+GX.data()[i-1]);
        I.data()[i] = 1+aTemp1+aTemp2;
        J.data()[i] = -aTemp1;
        K.data()[i-1] = -aTemp2;
        i++;
      }
      aTemp2 = aTau*(GX.data()[i]+GX.data()[i-1]);
      I.data()[i] = 1+aTemp2;
      J.data()[i] = 0;
      K.data()[i-1] = -aTemp2;
      i++;
    }
  }
  M.data()[0] = 1.0/I.data()[0];
  Y.data()[0] = aMatrix.data()[0];
  i = 0;
  do {
    float l = K.data()[i]*M.data()[i];
    i++;
    M.data()[i] = 1.0/(I.data()[i]-l*J.data()[i-1]);
    Y.data()[i] = aMatrix.data()[i]-l*Y.data()[i-1];
  }
  while (i < aSize-1);
  float v = Y.data()[aSize-1]*M.data()[aSize-1];
  aResult.data()[aSize-1] = v;
  for (int j = aSize-2; j >= 0; j--) {
    v = (Y.data()[j]-J.data()[j]*v)*M.data()[j];
    aResult.data()[j] = v;
  }
  // Compute diffusion in y direction
  // alpha -> I   beta -> J    gamma -> K
  i = 0;
  if (mSmoothing == 0) {
    for (int x = 0; x < aMatrix.xSize(); x++) {
      float aTemp1 = aTau*GY(x,0);
      float aTemp2;
      I.data()[i] = 1+aTemp1;
      J.data()[i] = -aTemp1;
      if (i > 0) K.data()[i-1] = 0;
      i++;
      for (int y = 1; y+1 < aMatrix.ySize(); y++) {
        aTemp1 = aTau*GY(x,y);
        aTemp2 = aTau*GY(x,y-1);
        I.data()[i] = 1+aTemp1+aTemp2;
        J.data()[i] = -aTemp1;
        K.data()[i-1] = -aTemp2;
        i++;
      }
      aTemp2 = aTau*GY(x,aMatrix.ySize()-2);
      I.data()[i] = 1+aTemp2;
      J.data()[i] = 0;
      K.data()[i-1] = -aTemp2;
      i++;
    }
  }
  else {
    for (int x = 0; x < aMatrix.xSize(); x++) {
      float aTemp1 = aTau*(GY(x,0)+GY(x,1));
      float aTemp2;
      I.data()[i] = 1+aTemp1;
      J.data()[i] = -aTemp1;
      if (i > 0) K.data()[i-1] = 0;
      i++;
      for (int y = 1; y+1 < aMatrix.ySize(); y++) {
        aTemp1 = aTau*(GY(x,y)+GY(x,y+1));
        aTemp2 = aTau*(GY(x,y)+GY(x,y-1));
        I.data()[i] = 1+aTemp1+aTemp2;
        J.data()[i] = -aTemp1;
        K.data()[i-1] = -aTemp2;
       i++;
      }
      aTemp2 = aTau*(GY(x,aMatrix.ySize()-1)+GY(x,aMatrix.ySize()-2));
      I.data()[i] = 1+aTemp2;
      J.data()[i] = 0;
      K.data()[i-1] = -aTemp2;
      i++;
    }
  }
  M.data()[0] = 1.0/I.data()[0];
  Y.data()[0] = aMatrix.data()[0];
  i = 0;
  int x = 0;
  int y = 0;
  do {
    float l = K.data()[i]*M.data()[i];
    i++;
    y++;
    if (y >= aMatrix.ySize()) {
      y = 0;
      x++;
    }
    M.data()[i] = 1.0/(I.data()[i]-l*J.data()[i-1]);
    Y.data()[i] = aMatrix(x,y)-l*Y.data()[i-1];
  }
  while (i < aSize-1);
  v = Y.data()[aSize-1]*M.data()[aSize-1];
  aResult.data()[aSize-1] += v;
  x = aMatrix.xSize()-1; y = aMatrix.ySize()-1;
  for (int j = aSize-2; j >= 0; j--) {
    y--;
    if (y < 0) {
      y = aMatrix.ySize()-1;
      x--;
    }
    v = (Y.data()[j]-J.data()[j]*v)*M.data()[j];
    aResult(x,y) += v;
  }
  // Average between x- and y-direction
  for (int j = 0; j < aSize; j++)
    aMatrix.data()[j] = 0.5*aResult.data()[j];
}

// isotropicDiffuseSemiImplicifloat (3-D)
void CDiffusion::isotropicDiffuseSemiImplicit(CTensor<float>& aTensor, const CTensor<float>& GX, const CTensor<float>& GY, const CTensor<float>& GZ) {
  CTensor<float> aResult(aTensor.xSize(),aTensor.ySize(),aTensor.zSize());
  CTensor<float> M(aTensor.xSize(),aTensor.ySize(),aTensor.zSize());
  CTensor<float> Y(aTensor.xSize(),aTensor.ySize(),aTensor.zSize());
  CTensor<float> I(aTensor.xSize(),aTensor.ySize(),aTensor.zSize());
  CTensor<float> J(aTensor.xSize(),aTensor.ySize(),aTensor.zSize());
  CTensor<float> K(aTensor.xSize(),aTensor.ySize(),aTensor.zSize());
  int aSize = aTensor.xSize()*aTensor.ySize()*aTensor.zSize();
  float aTau;
  if (mSmoothing == 0) aTau = 3.0*mTau3D;
  else aTau = 1.5*mTau3D;
  // Compute diffusion in x direction
  // alpha -> I   beta -> J    gamma -> K
  int i = 0;
  if (mSmoothing == 0) {
    for (int z = 0; z < aTensor.zSize(); z++)
      for (int y = 0; y < aTensor.ySize(); y++) {
        float aTemp1 = aTau*GX.data()[i];
        float aTemp2;
        I.data()[i] = 1+aTemp1;
        J.data()[i] = -aTemp1;
        if (i > 0) K.data()[i-1] = 0;
        i++;
        for (int x = 1; x+1 < aTensor.xSize(); x++) {
          aTemp1 = aTau*GX.data()[i];
          aTemp2 = aTau*GX.data()[i-1];
          I.data()[i] = 1+aTemp1+aTemp2;
          J.data()[i] = -aTemp1;
          K.data()[i-1] = -aTemp2;
          i++;
        }
        aTemp2 = aTau*GX.data()[i-1];
        I.data()[i] = 1+aTemp2;
        J.data()[i] = 0;
        K.data()[i-1] = -aTemp2;
        i++;
      }
  }
  else {
    for (int z = 0; z < aTensor.zSize(); z++)
      for (int y = 0; y < aTensor.ySize(); y++) {
        float aTemp1 = aTau*(GX.data()[i]+GX.data()[i+1]);
        float aTemp2;
        I.data()[i] = 1+aTemp1;
        J.data()[i] = -aTemp1;
        if (i > 0) K.data()[i-1] = 0;
        i++;
        for (int x = 1; x+1 < aTensor.xSize(); x++) {
          aTemp1 = aTau*(GX.data()[i]+GX.data()[i+1]);
          aTemp2 = aTau*(GX.data()[i]+GX.data()[i-1]);
          I.data()[i] = 1+aTemp1+aTemp2;
          J.data()[i] = -aTemp1;
          K.data()[i-1] = -aTemp2;
          i++;
        }
        aTemp2 = aTau*(GX.data()[i]+GX.data()[i-1]);
        I.data()[i] = 1+aTemp2;
        J.data()[i] = 0;
        K.data()[i-1] = -aTemp2;
        i++;
      }
  }
  M.data()[0] = 1.0/I.data()[0];
  Y.data()[0] = aTensor.data()[0];
  i = 0;
  do {
    float l = K.data()[i]*M.data()[i];
    i++;
    M.data()[i] = 1.0/(I.data()[i]-l*J.data()[i-1]);
    Y.data()[i] = aTensor.data()[i]-l*Y.data()[i-1];
  }
  while (i < aSize-1);
  float v = Y.data()[aSize-1]*M.data()[aSize-1];
  aResult.data()[aSize-1] = v;
  for (int j = aSize-2; j >= 0; j--) {
    v = (Y.data()[j]-J.data()[j]*v)*M.data()[j];
    aResult.data()[j] = v;
  }
  // Compute diffusion in y direction
  // alpha -> I   beta -> J    gamma -> K
  i = 0;
  if (mSmoothing == 0) {
    for (int z = 0; z < aTensor.zSize(); z++)
      for (int x = 0; x < aTensor.xSize(); x++) {
        float aTemp1 = aTau*GY(x,0,z);
        float aTemp2;
        I.data()[i] = 1+aTemp1;
        J.data()[i] = -aTemp1;
        if (i > 0) K.data()[i-1] = 0;
        i++;
        for (int y = 1; y+1 < aTensor.ySize(); y++) {
          aTemp1 = aTau*GY(x,y,z);
          aTemp2 = aTau*GY(x,y-1,z);
          I.data()[i] = 1+aTemp1+aTemp2;
          J.data()[i] = -aTemp1;
          K.data()[i-1] = -aTemp2;
          i++;
        }
        aTemp2 = aTau*GY(x,aTensor.ySize()-2,z);
        I.data()[i] = 1+aTemp2;
        J.data()[i] = 0;
        K.data()[i-1] = -aTemp2;
        i++;
      }
  }
  else {
    for (int z = 0; z < aTensor.zSize(); z++)
      for (int x = 0; x < aTensor.xSize(); x++) {
        float aTemp1 = aTau*(GY(x,0,z)+GY(x,1,z));
        float aTemp2;
        I.data()[i] = 1+aTemp1;
        J.data()[i] = -aTemp1;
        if (i > 0) K.data()[i-1] = 0;
        i++;
        for (int y = 1; y+1 < aTensor.ySize(); y++) {
          aTemp1 = aTau*(GY(x,y,z)+GY(x,y+1,z));
          aTemp2 = aTau*(GY(x,y,z)+GY(x,y-1,z));
          I.data()[i] = 1+aTemp1+aTemp2;
          J.data()[i] = -aTemp1;
          K.data()[i-1] = -aTemp2;
          i++;
        }
        aTemp2 = aTau*(GY(x,aTensor.ySize()-1,z)+GY(x,aTensor.ySize()-2,z));
        I.data()[i] = 1+aTemp2;
        J.data()[i] = 0;
        K.data()[i-1] = -aTemp2;
        i++;
      }
  }
  M.data()[0] = 1.0/I.data()[0];
  Y.data()[0] = aTensor.data()[0];
  i = 0;
  int x = 0; int y = 0; int z = 0;
  do {
    float l = K.data()[i]*M.data()[i];
    i++;
    y++;
    if (y >= aTensor.ySize()) {
      y = 0;
      x++;
      if (x >= aTensor.xSize()) {
        x = 0;
        z++;
      }
    }
    M.data()[i] = 1.0/(I.data()[i]-l*J.data()[i-1]);
    Y.data()[i] = aTensor(x,y,z)-l*Y.data()[i-1];
  }
  while (i < aSize-1);
  v = Y.data()[aSize-1]*M.data()[aSize-1];
  aResult.data()[aSize-1] += v;
  x = aTensor.xSize()-1; y = aTensor.ySize()-1; z = aTensor.zSize()-1;
  for (int j = aSize-2; j >= 0; j--) {
    y--;
    if (y < 0) {
      y = aTensor.ySize()-1;
      x--;
      if (x < 0) {
        x = aTensor.xSize()-1;
        z--;
      }
    }
    v = (Y.data()[j]-J.data()[j]*v)*M.data()[j];
    aResult(x,y,z) += v;
  }
  // Compute diffusion in z direction
  // alpha -> I   beta -> J    gamma -> K
  i = 0;
  if (mSmoothing == 0) {
    for (int y = 0; y < aTensor.ySize(); y++)
      for (int x = 0; x < aTensor.xSize(); x++) {
        float aTemp1 = aTau*GZ(x,y,0);
        float aTemp2;
        I.data()[i] = 1+aTemp1;
        J.data()[i] = -aTemp1;
        if (i > 0) K.data()[i-1] = 0;
        i++;
        for (int z = 1; z+1 < aTensor.zSize(); z++) {
          aTemp1 = aTau*GZ(x,y,z);
          aTemp2 = aTau*GZ(x,y,z-1);
          I.data()[i] = 1+aTemp1+aTemp2;
          J.data()[i] = -aTemp1;
          K.data()[i-1] = -aTemp2;
          i++;
        }
        aTemp2 = aTau*GZ(x,y,aTensor.zSize()-2);
        I.data()[i] = 1+aTemp2;
        J.data()[i] = 0;
        K.data()[i-1] = -aTemp2;
        i++;
      }
  }
  else {
    for (int y = 0; y < aTensor.ySize(); y++)
      for (int x = 0; x < aTensor.xSize(); x++) {
        float aTemp1 = aTau*(GZ(x,y,0)+GZ(x,y,1));
        float aTemp2;
        I.data()[i] = 1+aTemp1;
        J.data()[i] = -aTemp1;
        if (i > 0) K.data()[i-1] = 0;
        i++;
        for (int z = 1; z+1 < aTensor.zSize(); z++) {
          aTemp1 = aTau*(GZ(x,y,z)+GZ(x,y,z+1));
          aTemp2 = aTau*(GZ(x,y,z)+GZ(x,y,z-1));
          I.data()[i] = 1+aTemp1+aTemp2;
          J.data()[i] = -aTemp1;
          K.data()[i-1] = -aTemp2;
          i++;
        }
        aTemp2 = aTau*(GZ(x,y,aTensor.zSize()-1)+GZ(x,y,aTensor.zSize()-2));
        I.data()[i] = 1+aTemp2;
        J.data()[i] = 0;
        K.data()[i-1] = -aTemp2;
        i++;
      }
  }
  M.data()[0] = 1.0/I.data()[0];
  Y.data()[0] = aTensor.data()[0];
  i = 0;
  x = 0; y = 0; z = 0;
  do {
    float l = K.data()[i]*M.data()[i];
    i++;
    z++;
    if (z >= aTensor.zSize()) {
      z = 0;
      x++;
      if (x >= aTensor.xSize()) {
        x = 0;
        y++;
      }
    }
    M.data()[i] = 1.0/(I.data()[i]-l*J.data()[i-1]);
    Y.data()[i] = aTensor(x,y,z)-l*Y.data()[i-1];
  }
  while (i < aSize-1);
  v = Y.data()[aSize-1]*M.data()[aSize-1];
  aResult.data()[aSize-1] += v;
  x = aTensor.xSize()-1; y = aTensor.ySize()-1; z = aTensor.zSize()-1;
  for (int j = aSize-2; j >= 0; j--) {
    z--;
    if (z < 0) {
      z = aTensor.zSize()-1;
      x--;
      if (x < 0) {
        x = aTensor.xSize()-1;
        y--;
      }
    }
    v = (Y.data()[j]-J.data()[j]*v)*M.data()[j];
    aResult(x,y,z) += v;
  }
  // Average between x-, y-, and z-direction
  for (int j = 0; j < aSize; j++)
    aTensor.data()[j] = 0.33333333333333*aResult.data()[j];
}

// anisotropicDiffuseSemiImplicifloat (2-D)
void CDiffusion::anisotropicDiffuseSemiImplicit(CMatrix<float>& aMatrix, const CMatrix<float>& A, const CMatrix<float>& B, const CMatrix<float>& C) {
  CMatrix<float> aResult(aMatrix.xSize(),aMatrix.ySize());
  CMatrix<float> M(aMatrix.xSize(),aMatrix.ySize());
  CMatrix<float> Y(aMatrix.xSize(),aMatrix.ySize());
  CMatrix<float> I(aMatrix.xSize(),aMatrix.ySize());
  CMatrix<float> J(aMatrix.xSize(),aMatrix.ySize());
  CMatrix<float> K(aMatrix.xSize(),aMatrix.ySize());
  CMatrix<float> P(aMatrix.xSize(),aMatrix.ySize());
  int aSize = aMatrix.xSize()*aMatrix.ySize();
  float aTau = mTau2D;
  // Determine P for anisotropicy
  for (int j = 0; j < aSize; j++)
    P.data()[j] = 0.5*(fabs(B.data()[j])+NMath::min(A.data()[j],C.data()[j]));
  // Compute diffusion in x direction
  // diffusivity -> M   alpha -> I   beta -> J    gamma -> K
  for (int j = 0; j < aSize; j++)
    M.data()[j] = A.data()[j]-P.data()[j];
  int i = 0;
  for (int y = 0; y < aMatrix.ySize(); y++) {
    float aTemp1 = aTau*(M.data()[i]+M.data()[i+1]);
    float aTemp2;
    I.data()[i] = 1+aTemp1;
    J.data()[i] = -aTemp1;
    if (i > 0) K.data()[i-1] = 0;
    i++;
    for (int x = 1; x+1 < aMatrix.xSize(); x++) {
      aTemp1 = aTau*(M.data()[i]+M.data()[i+1]);
      aTemp2 = aTau*(M.data()[i]+M.data()[i-1]);
      I.data()[i] = 1+aTemp1+aTemp2;
      J.data()[i] = -aTemp1;
      K.data()[i-1] = -aTemp2;
      i++;
    }
    aTemp2 = aTau*(M.data()[i]+M.data()[i-1]);
    I.data()[i] = 1+aTemp2;
    J.data()[i] = 0;
    K.data()[i-1] = -aTemp2;
    i++;
  }
  M.data()[0] = 1.0/I.data()[0];
  Y.data()[0] = aMatrix.data()[0];
  i = 0;
  do {
    float l = K.data()[i]*M.data()[i];
    i++;
    M.data()[i] = 1.0/(I.data()[i]-l*J.data()[i-1]);
    Y.data()[i] = aMatrix.data()[i]-l*Y.data()[i-1];
  }
  while (i < aSize-1);
  float v = Y.data()[aSize-1]*M.data()[aSize-1];
  aResult.data()[aSize-1] = v;
  for (int j = aSize-2; j >= 0; j--) {
    v = (Y.data()[j]-J.data()[j]*v)*M.data()[j];
    aResult.data()[j] = v;
  }
  // Compute diffusion in y direction
  // diffusivity -> M   alpha -> I   beta -> J    gamma -> K
  for (int j = 0; j < aSize; j++)
    M.data()[j] = C.data()[j]-P.data()[j];
  i = 0;
  for (int x = 0; x < aMatrix.xSize(); x++) {
    float aTemp1 = aTau*(M(x,0)+M(x,1));
    float aTemp2;
    I.data()[i] = 1+aTemp1;
    J.data()[i] = -aTemp1;
    if (i > 0) K.data()[i-1] = 0;
    i++;
    for (int y = 1; y+1 < aMatrix.ySize(); y++) {
      aTemp1 = aTau*(M(x,y)+M(x,y+1));
      aTemp2 = aTau*(M(x,y)+M(x,y-1));
      I.data()[i] = 1+aTemp1+aTemp2;
      J.data()[i] = -aTemp1;
      K.data()[i-1] = -aTemp2;
     i++;
    }
    aTemp2 = aTau*(M(x,aMatrix.ySize()-1)+M(x,aMatrix.ySize()-2));
    I.data()[i] = 1+aTemp2;
    J.data()[i] = 0;
    K.data()[i-1] = -aTemp2;
    i++;
  }
  M.data()[0] = 1.0/I.data()[0];
  Y.data()[0] = aMatrix.data()[0];
  i = 0;
  int x = 0;
  int y = 0;
  do {
    float l = K.data()[i]*M.data()[i];
    i++;
    y++;
    if (y >= aMatrix.ySize()) {
      y = 0;
      x++;
    }
    M.data()[i] = 1.0/(I.data()[i]-l*J.data()[i-1]);
    Y.data()[i] = aMatrix(x,y)-l*Y.data()[i-1];
  }
  while (i < aSize-1);
  v = Y.data()[aSize-1]*M.data()[aSize-1];
  aResult.data()[aSize-1] += v;
  x = aMatrix.xSize()-1; y = aMatrix.ySize()-1;
  for (int j = aSize-2; j >= 0; j--) {
    y--;
    if (y < 0) {
      y = aMatrix.ySize()-1;
      x--;
    }
    v = (Y.data()[j]-J.data()[j]*v)*M.data()[j];
    aResult(x,y) += v;
  }
  aTau *= 0.5;
  // Compute diffusion in xy direction
  // diffusivity -> M   alpha -> I   beta -> J    gamma -> K
  for (int j = 0; j < aSize; j++)
    M.data()[j] = P.data()[j]+B.data()[j];
  for (int j = -(aMatrix.ySize()-2); j < aMatrix.xSize()-1; j++) {
    i = 0;
    // Starting point of line
    int x,y;
    if (j < 0) {
      x = 0; y = -j;
    }
    else {
      x = j; y = 0;
    }
    // Length of line
    int aLength = (int)NMath::min(aMatrix.ySize()-y,aMatrix.xSize()-x);
    // Build up diffusion matrix
    float aTemp1 = aTau*(M(x,y)+M(x+1,y+1));
    float aTemp2;
    I.data()[i] = 1+aTemp1;
    J.data()[i] = -aTemp1;
    if (i > 0) K.data()[i-1] = 0;
    i++;
    for (int k = 1; k < aLength-1; k++) {
      aTemp1 = aTau*(M(x+k,y+k)+M(x+k+1,y+k+1));
      aTemp2 = aTau*(M(x+k,y+k)+M(x+k-1,y+k-1));
      I.data()[i] = 1+aTemp1+aTemp2;
      J.data()[i] = -aTemp1;
      K.data()[i-1] = -aTemp2;
      i++;
    }
    aTemp2 = aTau*(M(x+i,y+i)+M(x+i-1,y+i-1));
    I.data()[i] = 1+aTemp2;
    J.data()[i] = 0;
    K.data()[i-1] = -aTemp2;
    // Solve linear system
    M.data()[0] = 1.0/I.data()[0];
    Y.data()[0] = aMatrix(x,y);
    for (int k = 0; k < aLength-1;) {
      float l = K.data()[k]*M.data()[k];
      k++;
      M.data()[k] = 1.0/(I.data()[k]-l*J.data()[k-1]);
      Y.data()[k] = aMatrix(x+k,y+k)-l*Y.data()[k-1];
    }
    v = Y.data()[aLength-1]*M.data()[aLength-1];
    aResult(x+aLength-1,y+aLength-1) += v;
    for (int k = aLength-2; k >= 0; k--) {
      v = (Y.data()[k]-J.data()[k]*v)*M.data()[k];
      aResult(x+k,y+k) += v;
    }
  }
  aResult(aMatrix.xSize()-1,0) += aMatrix(aMatrix.xSize()-1,0);
  aResult(0,aMatrix.ySize()-1) += aMatrix(0,aMatrix.ySize()-1);
  // Compute diffusion in -xy direction
  // diffusivity -> M   alpha -> I   beta -> J    gamma -> K
  for (int j = 0; j < aSize; j++)
    M.data()[j] = P.data()[j]-B.data()[j];
  for (int j = -(aMatrix.ySize()-2); j < aMatrix.xSize()-1; j++) {
    int i = 0;
    // Starting point of line
    int x,y;
    if (j < 0) {
      x = 0; y = j+aMatrix.ySize()-1;
    }
    else {
      x = j; y = aMatrix.ySize()-1;
    }
    // Length of line
    int aLength = (int)NMath::min(y+1,aMatrix.xSize()-x);
    float aTemp1 = aTau*(M(x,y)+M(x+1,y-1));
    float aTemp2;
    I.data()[i] = 1+aTemp1;
    J.data()[i] = -aTemp1;
    if (i > 0) K.data()[i-1] = 0;
    i++;
    for (int k = 1; k < aLength-1; k++) {
      aTemp1 = aTau*(M(x+k,y-k)+M(x+k+1,y-k-1));
      aTemp2 = aTau*(M(x+k,y-k)+M(x+k-1,y-k+1));
      I.data()[i] = 1+aTemp1+aTemp2;
      J.data()[i] = -aTemp1;
      K.data()[i-1] = -aTemp2;
      i++;
    }
    aTemp2 = aTau*(M(x+i,y-i)+M(x+i-1,y-i+1));
    I.data()[i] = 1+aTemp2;
    J.data()[i] = 0;
    K.data()[i-1] = -aTemp2;
    M.data()[0] = 1.0/I.data()[0];
    Y.data()[0] = aMatrix(x,y);
    for (int k = 0; k < aLength-1;) {
      float l = K.data()[k]*M.data()[k];
      k++;
      M.data()[k] = 1.0/(I.data()[k]-l*J.data()[k-1]);
      Y.data()[k] = aMatrix(x+k,y-k)-l*Y.data()[k-1];
    }
    v = Y.data()[aLength-1]*M.data()[aLength-1];
    aResult(x+aLength-1,y-aLength+1) += v;
    for (int k = aLength-2; k >= 0; k--) {
      v = (Y.data()[k]-J.data()[k]*v)*M.data()[k];
      aResult(x+k,y-k) += v;
    }
  }
  aResult(0,0) += aMatrix(0,0);
  aResult(aMatrix.xSize()-1,aMatrix.ySize()-1) += aMatrix(aMatrix.xSize()-1,aMatrix.ySize()-1);
  // Average between x- and y-direction
  for (int j = 0; j < aSize; j++)
    aMatrix.data()[j] = 0.25*aResult.data()[j];
}

// Two-pixel numeric -----------------------------------------------------------

// Gradienfloat computation --------------------------------------------------------

// twoPixelGrad (1-D)
void CDiffusion::twoPixelGrad(const CVector<float>& aData, CVector<float>& aGrad, int aOffset) {
  int aBoundary = aData.size()-1;
  for (int x = aOffset; x < aBoundary; x+=2) {
    float aTemp = aData(x+1)-aData(x);
    aGrad(x) = aTemp*aTemp;
  }
}

// twoPixelGradX (2-D)
void CDiffusion::twoPixelGradX(const CMatrix<float>& aData, CMatrix<float>& aGrad, int aOffset) {
  int aBoundary = aData.xSize()-1;
  CDerivative<float> aDerivative(3);
  NFilter::filter(aData,aGrad,1,aDerivative);
  for (int y = 0; y < aData.ySize(); y++)
    for (int x = aOffset; x < aBoundary; x+=2) {
      float aTemp = 0.5*(aGrad(x,y)+aGrad(x+1,y));
      float aTemp2 = aData(x+1,y)-aData(x,y);
      aGrad(x,y) = aTemp*aTemp+aTemp2*aTemp2;
    }
}

// twoPixelGradY (2-D)
void CDiffusion::twoPixelGradY(const CMatrix<float>& aData, CMatrix<float>& aGrad, int aOffset) {
  int aBoundary = aData.ySize()-1;
  CDerivative<float> aDerivative(3);
  NFilter::filter(aData,aGrad,aDerivative,1);
  for (int y = aOffset; y < aBoundary; y+=2)
    for (int x = 0; x < aData.xSize(); x++) {
      float aTemp = 0.5*(aGrad(x,y)+aGrad(x,y+1));
      float aTemp2 = aData(x,y+1)-aData(x,y);
      aGrad(x,y) = aTemp*aTemp+aTemp2*aTemp2;
    }
}

// twoPixelGradXMultiChannel (2-D)
void CDiffusion::twoPixelGradXMultiChannel(const CTensor<float>& aData, CMatrix<float>& aGrad, int aOffset) {
  int aBoundary = aData.xSize()-1;
  CDerivative<float> aDerivative(3);
  CMatrix<float> aTemp(aData.xSize(),aData.ySize());
  aGrad = 0;
  for (int k = aData.zSize()-1; k >= 0; k--) {
    aData.getMatrix(aTemp,k);
    for (int y = 0; y < aData.ySize(); y++)
      for (int x = aOffset; x < aBoundary; x+=2) {
        float aHelp = aTemp(x+1,y)-aTemp(x,y);
        aGrad(x,y) += aHelp*aHelp;
      }
    NFilter::filter(aTemp,1,aDerivative);
    for (int y = 0; y < aData.ySize(); y++)
      for (int x = aOffset; x < aBoundary; x+=2) {
        float aHelp = 0.5*(aTemp(x,y)+aTemp(x+1,y));
        aGrad(x,y) += aHelp*aHelp;
      }
  }
}

// twoPixelGradYMultiChannel (2-D)
void CDiffusion::twoPixelGradYMultiChannel(const CTensor<float>& aData, CMatrix<float>& aGrad, int aOffset) {
  int aBoundary = aData.ySize()-1;
  CDerivative<float> aDerivative(3);
  CMatrix<float> aTemp(aData.xSize(),aData.ySize());
  aGrad = 0;
  for (int k = aData.zSize()-1; k >= 0; k--) {
    aData.getMatrix(aTemp,k);
    for (int y = aOffset; y < aBoundary; y+=2)
      for (int x = 0; x < aData.xSize(); x++) {
        float aHelp = aTemp(x,y+1)-aTemp(x,y);
        aGrad(x,y) += aHelp*aHelp;
      }
    NFilter::filter(aTemp,aDerivative,1);
    for (int y = aOffset; y < aBoundary; y+=2)
      for (int x = 0; x < aData.xSize(); x++) {
        float aHelp = 0.5*(aTemp(x,y)+aTemp(x,y+1));
        aGrad(x,y) += aHelp*aHelp;
      }
  }
}

// twoPixelGradX (3-D)
void CDiffusion::twoPixelGradX(const CTensor<float>& aData, CTensor<float>& aGrad, int aOffset) {
  int aBoundary = aData.xSize()-1;
  CDerivative<float> aDerivative(3);
  CTensor<float> aTempGrad(aGrad.xSize(),aGrad.ySize(),aGrad.zSize());
  NFilter::filter(aData,aGrad,1,aDerivative,1);
  NFilter::filter(aData,aTempGrad,1,1,aDerivative);
  for (int z = 0; z < aData.zSize(); z++)
    for (int y = 0; y < aData.ySize(); y++)
      for (int x = aOffset; x < aBoundary; x+=2) {
        float aTemp = aGrad(x,y,z)+aGrad(x+1,y,z);
        float aTemp2 = aTempGrad(x,y,z)+aTempGrad(x+1,y,z);
        float aTemp3 = aData(x+1,y,z)-aData(x,y,z);
        aGrad(x,y,z) = 0.25*(aTemp*aTemp+aTemp2*aTemp2+aTemp3*aTemp3);
      }
}

// twoPixelGradY (3-D)
void CDiffusion::twoPixelGradY(const CTensor<float>& aData, CTensor<float>& aGrad, int aOffset) {
  int aBoundary = aData.ySize()-1;
  CDerivative<float> aDerivative(3);
  CTensor<float> aTempGrad(aGrad.xSize(),aGrad.ySize(),aGrad.zSize());
  NFilter::filter(aData,aGrad,aDerivative,1,1);
  NFilter::filter(aData,aTempGrad,1,1,aDerivative);
  for (int z = 0; z < aData.zSize(); z++)
    for (int y = aOffset; y < aBoundary; y+=2)
      for (int x = 0; x < aData.xSize(); x++) {
        float aTemp = aGrad(x,y,z)+aGrad(x,y+1,z);
        float aTemp2 = aTempGrad(x,y,z)+aTempGrad(x,y+1,z);
        float aTemp3 = aData(x,y+1,z)-aData(x,y,z);
        aGrad(x,y,z) = 0.25*(aTemp*aTemp+aTemp2*aTemp2+aTemp3*aTemp3);
      }
}

// twoPixelGradZ (3-D)
void CDiffusion::twoPixelGradZ(const CTensor<float>& aData, CTensor<float>& aGrad, int aOffset) {
  int aBoundary = aData.zSize()-1;
  CDerivative<float> aDerivative(3);
  CTensor<float> aTempGrad(aGrad.xSize(),aGrad.ySize(),aGrad.zSize());
  NFilter::filter(aData,aGrad,aDerivative,1,1);
  NFilter::filter(aData,aTempGrad,1,aDerivative,1);
  for (int z = aOffset; z < aBoundary; z+=2)
    for (int y = 0; y < aData.ySize(); y++)
      for (int x = 0; x < aData.xSize(); x++) {
        float aTemp = aGrad(x,y,z)+aGrad(x,y,z+1);
        float aTemp2 = aTempGrad(x,y,z)+aTempGrad(x,y,z+1);
        float aTemp3 = aData(x,y,z+1)-aData(x,y,z);
        aGrad(x,y,z) = 0.25*(aTemp*aTemp+aTemp2*aTemp2+aTemp3*aTemp3);
      }
}

// twoPixelGradXMultiChannel (3-D)
void CDiffusion::twoPixelGradXMultiChannel(const CTensor4D<float>& aData, CTensor<float>& aGrad, int aOffset) {
  int aBoundary = aData.xSize()-1;
  CDerivative<float> aDerivative(3);
  CTensor<float> aTemp(aData.xSize(),aData.ySize(),aData.zSize());
  CTensor<float> aTemp2(aData.xSize(),aData.ySize(),aData.zSize());
  aGrad = 0;
  for (int k = aData.aSize()-1; k >= 0; k--) {
    aData.getTensor3D(aTemp,k);
    for (int z = 0; z < aData.zSize(); z++)
      for (int y = 0; y < aData.ySize(); y++)
        for (int x = aOffset; x < aBoundary; x+=2) {
          float aHelp = aTemp(x+1,y,z)-aTemp(x,y,z);
          aGrad(x,y,z) += aHelp*aHelp;
        }
    NFilter::filter(aTemp,aTemp2,1,aDerivative,1);
    NFilter::filter(aTemp,1,1,aDerivative);
    for (int z = 0; z < aData.zSize(); z++)
      for (int y = 0; y < aData.ySize(); y++)
        for (int x = aOffset; x < aBoundary; x+=2) {
          float aHelp = aTemp(x,y,z)+aTemp(x+1,y,z);
          float aHelp2 = aTemp2(x,y,z)+aTemp2(x+1,y,z);
          aGrad(x,y,z) += 0.25*(aHelp*aHelp+aHelp2*aHelp2);
        }
  }
}

// twoPixelGradYMultiChannel (3-D)
void CDiffusion::twoPixelGradYMultiChannel(const CTensor4D<float>& aData, CTensor<float>& aGrad, int aOffset) {
  int aBoundary = aData.ySize()-1;
  CDerivative<float> aDerivative(3);
  CTensor<float> aTemp(aData.xSize(),aData.ySize(),aData.zSize());
  CTensor<float> aTemp2(aData.xSize(),aData.ySize(),aData.zSize());
  aGrad = 0;
  for (int k = aData.aSize()-1; k >= 0; k--) {
    aData.getTensor3D(aTemp,k);
    for (int z = 0; z < aData.zSize(); z++)
      for (int y = aOffset; y < aBoundary; y+=2)
        for (int x = 0; x < aData.xSize(); x++) {
          float aHelp = aTemp(x,y+1,z)-aTemp(x,y,z);
          aGrad(x,y,z) += aHelp*aHelp;
        }
    NFilter::filter(aTemp,aTemp2,aDerivative,1,1);
    NFilter::filter(aTemp,1,1,aDerivative);
    for (int z = 0; z < aData.zSize(); z++)
      for (int y = aOffset; y < aBoundary; y+=2)
        for (int x = 0; x < aData.xSize(); x++) {
          float aHelp = aTemp(x,y,z)+aTemp(x,y+1,z);
          float aHelp2 = aTemp2(x,y,z)+aTemp2(x,y+1,z);
          aGrad(x,y,z) += 0.25*(aHelp*aHelp+aHelp2*aHelp2);
        }
  }
}

// twoPixelGradZMultiChannel (3-D)
void CDiffusion::twoPixelGradZMultiChannel(const CTensor4D<float>& aData, CTensor<float>& aGrad, int aOffset) {
  int aBoundary = aData.zSize()-1;
  CDerivative<float> aDerivative(3);
  CTensor<float> aTemp(aData.xSize(),aData.ySize(),aData.zSize());
  CTensor<float> aTemp2(aData.xSize(),aData.ySize(),aData.zSize());
  aGrad = 0;
  for (int k = aData.aSize()-1; k >= 0; k--) {
    aData.getTensor3D(aTemp,k);
    for (int z = aOffset; z < aBoundary; z+=2)
      for (int y = 0; y < aData.ySize(); y++)
        for (int x = 0; x < aData.xSize(); x++) {
          float aHelp = aTemp(x,y,z+1)-aTemp(x,y,z);
          aGrad(x,y,z) += aHelp*aHelp;
        }
    NFilter::filter(aTemp,aTemp2,aDerivative,1,1);
    NFilter::filter(aTemp,1,aDerivative,1);
    for (int z = aOffset; z < aBoundary; z+=2)
      for (int y = 0; y < aData.ySize(); y++)
        for (int x = 0; x < aData.xSize(); x++) {
          float aHelp = 0.5*(aTemp(x,y,z)+aTemp(x,y,z+1));
          aGrad(x,y,z) += aHelp*aHelp;
        }
    for (int z = aOffset; z < aBoundary; z+=2)
      for (int y = 0; y < aData.ySize(); y++)
        for (int x = 0; x < aData.xSize(); x++) {
          float aHelp = 0.5*(aTemp2(x,y,z)+aTemp2(x,y,z+1));
          aGrad(x,y,z) += aHelp*aHelp;
        }
  }
}

// Diffusion -------------------------------------------------------------------

// twoPixelDiffuseLeft (1-D)
void CDiffusion::twoPixelDiffuseLeft(CVector<float>& aData, CVector<float>& aGrad) {
  float invLambdaSqr = 1.0/(mLambda*mLambda);
  for (int x = 1; x < aData.size(); x+=2) {
    float aTemp = aData(x)-aData(x-1);
    float aNominator;
    if (aTemp == 0) aNominator = 0;
    else {
      aNominator = mTau1D*diffusivity(mDiffFunc,aGrad(x-1),invLambdaSqr);
      if (aNominator > 0.5) aNominator = 0.5;
      aNominator *= aTemp;
    }
    aData(x-1) += aNominator;
    aData(x) -= aNominator;
  }
}

// twoPixelDiffuseRight (1-D)
void CDiffusion::twoPixelDiffuseRight(CVector<float>& aData, CVector<float>& aGrad) {
  float invLambdaSqr = 1.0/(mLambda*mLambda);
  int aRightBoundary = aData.size()-1;
  for (int x = 1; x < aRightBoundary; x+=2) {
    float aTemp = aData(x)-aData(x+1);
    float aNominator;
    if (aTemp == 0) aNominator = 0;
    else {
      aNominator = mTau1D*diffusivity(mDiffFunc,aGrad(x),invLambdaSqr);
      if (aNominator > 0.5) aNominator = 0.5;
      aNominator *= aTemp;
    }
    aData(x+1) += aNominator;
    aData(x) -= aNominator;
  }
}

// twoPixelDiffuseLeft (2-D)
void CDiffusion::twoPixelDiffuseLeft(CMatrix<float>& aData, CMatrix<float>& aGrad) {
  float invLambdaSqr = 1.0/(mLambda*mLambda);
  for (int y = 0; y < aData.ySize(); y++)
    for (int x = 1; x < aData.xSize(); x+=2) {
      float aTemp = aData(x,y)-aData(x-1,y);
      float aNominator;
      if (aTemp == 0) aNominator = 0;
      else {
        aNominator = mTau2D*diffusivity(mDiffFunc,aGrad(x-1,y),invLambdaSqr);
        if (aNominator > 0.5) aNominator = 0.5;
        aNominator *= aTemp;
      }
      aData(x-1,y) += aNominator;
      aData(x,y) -= aNominator;
    }
}

// twoPixelDiffuseRight (2-D)
void CDiffusion::twoPixelDiffuseRight(CMatrix<float>& aData, CMatrix<float>& aGrad) {
  float invLambdaSqr = 1.0/(mLambda*mLambda);
  int aRightBoundary = aData.xSize()-1;
  for (int y = 0; y < aData.ySize(); y++)
    for (int x = 1; x < aRightBoundary; x+=2) {
      float aTemp = aData(x,y)-aData(x+1,y);
      float aNominator;
      if (aTemp == 0) aNominator = 0;
      else {
        aNominator = mTau2D*diffusivity(mDiffFunc,aGrad(x,y),invLambdaSqr);
        if (aNominator > 0.5) aNominator = 0.5;
        aNominator *= aTemp;
      }
      aData(x+1,y) += aNominator;
      aData(x,y) -= aNominator;
    }
}

// twoPixelDiffuseUp (2-D)
void CDiffusion::twoPixelDiffuseUp(CMatrix<float>& aData, CMatrix<float>& aGrad) {
  float invLambdaSqr = 1.0/(mLambda*mLambda);
  for (int y = 1; y < aData.ySize(); y+=2)
    for (int x = 0; x < aData.xSize(); x++) {
      float aTemp = aData(x,y)-aData(x,y-1);
      float aNominator;
      if (aTemp == 0) aNominator = 0;
      else {
        aNominator = mTau2D*diffusivity(mDiffFunc,aGrad(x,y-1),invLambdaSqr);
        if (aNominator > 0.5) aNominator = 0.5;
        aNominator *= aTemp;
      }
      aData(x,y-1) += aNominator;
      aData(x,y) -= aNominator;
    }
}

// twoPixelDiffuseDown (2-D)
void CDiffusion::twoPixelDiffuseDown(CMatrix<float>& aData, CMatrix<float>& aGrad) {
  int aBottomBoundary = aData.ySize()-1;
  for (int y = 1; y < aBottomBoundary; y+=2)
    for (int x = 0; x < aData.xSize(); x++) {
      float aTemp = aData(x,y)-aData(x,y+1);
      float aNominator;
      if (aTemp == 0) aNominator = 0;
      else {
        aNominator = mTau2D*diffusivity(mDiffFunc,aGrad(x,y),0);
        if (aNominator > 0.5) aNominator = 0.5;
        aNominator *= aTemp;
      }
      aData(x,y+1) += aNominator;
      aData(x,y) -= aNominator;
    }
}

// twoPixelDiffuseLeft (3-D)
void CDiffusion::twoPixelDiffuseLeft(CTensor<float>& aData, CTensor<float>& aGrad) {
  for (int z = 0; z < aData.zSize(); z++)
    for (int y = 0; y < aData.ySize(); y++)
      for (int x = 1; x < aData.xSize(); x+=2) {
        float aTemp = aData(x,y,z)-aData(x-1,y,z);
        float aNominator;
        if (aTemp == 0) aNominator = 0;
        else {
          aNominator = mTau3D*diffusivity(mDiffFunc,aGrad(x-1,y,z),0);
          if (aNominator > 0.5) aNominator = 0.5;
          aNominator *= aTemp;
        }
        aData(x-1,y,z) += aNominator;
        aData(x,y,z) -= aNominator;
      }
}

// twoPixelDiffuseRight (3-D)
void CDiffusion::twoPixelDiffuseRight(CTensor<float>& aData, CTensor<float>& aGrad) {
  int aRightBoundary = aData.xSize()-1;
  for (int z = 0; z < aData.zSize(); z++)
    for (int y = 0; y < aData.ySize(); y++)
      for (int x = 1; x < aRightBoundary; x+=2) {
        float aTemp = aData(x,y,z)-aData(x+1,y,z);
        float aNominator;
        if (aTemp == 0) aNominator = 0;
        else {
          aNominator = mTau3D*diffusivity(mDiffFunc,aGrad(x,y,z),0);
          if (aNominator > 0.5) aNominator = 0.5;
          aNominator *= aTemp;
        }
        aData(x+1,y,z) += aNominator;
        aData(x,y,z) -= aNominator;
      }
}

// twoPixelDiffuseUp (3-D)
void CDiffusion::twoPixelDiffuseUp(CTensor<float>& aData, CTensor<float>& aGrad) {
  for (int z = 0; z < aData.zSize(); z++)
    for (int y = 1; y < aData.ySize(); y+=2)
      for (int x = 0; x < aData.xSize(); x++) {
        float aTemp = aData(x,y,z)-aData(x,y-1,z);
        float aNominator;
        if (aTemp == 0) aNominator = 0;
        else {
          aNominator = mTau3D*diffusivity(mDiffFunc,aGrad(x,y-1,z),0);
          if (aNominator > 0.5) aNominator = 0.5;
          aNominator *= aTemp;
        }
        aData(x,y-1,z) += aNominator;
        aData(x,y,z) -= aNominator;
      }
}

// twoPixelDiffuseDown (3-D)
void CDiffusion::twoPixelDiffuseDown(CTensor<float>& aData, CTensor<float>& aGrad) {
  int aBottomBoundary = aData.ySize()-1;
  for (int z = 0; z < aData.zSize(); z++)
    for (int y = 1; y < aBottomBoundary; y+=2)
      for (int x = 0; x < aData.xSize(); x++) {
        float aTemp = aData(x,y,z)-aData(x,y+1,z);
        float aNominator;
        if (aTemp == 0) aNominator = 0;
        else {
           aNominator = mTau3D*diffusivity(mDiffFunc,aGrad(x,y,z),0);
          if (aNominator > 0.5) aNominator = 0.5;
          aNominator *= aTemp;
        }
        aData(x,y+1,z) += aNominator;
        aData(x,y,z) -= aNominator;
      }
}

// twoPixelDiffuseInner (3-D)
void CDiffusion::twoPixelDiffuseInner(CTensor<float>& aData, CTensor<float>& aGrad) {
  for (int z = 1; z < aData.zSize(); z+=2)
    for (int y = 0; y < aData.ySize(); y++)
      for (int x = 0; x < aData.xSize(); x++) {
        float aTemp = aData(x,y,z)-aData(x,y,z-1);
        float aNominator;
        if (aTemp == 0) aNominator = 0;
        else {
          aNominator = mTau3D*diffusivity(mDiffFunc,aGrad(x,y,z-1),0);
          if (aNominator > 0.5) aNominator = 0.5;
          aNominator *= aTemp;
        }
        aData(x,y,z-1) += aNominator;
        aData(x,y,z) -= aNominator;
      }
}

// twoPixelDiffuseOuter (3-D)
void CDiffusion::twoPixelDiffuseOuter(CTensor<float>& aData, CTensor<float>& aGrad) {
  int aOuterBoundary = aData.zSize()-1;
  for (int z = 1; z < aOuterBoundary; z+=2)
    for (int y = 0; y < aData.ySize(); y++)
      for (int x = 0; x < aData.xSize(); x++) {
        float aTemp = aData(x,y,z)-aData(x,y,z+1);
        float aNominator;
        if (aTemp == 0) aNominator = 0;
        else {
          aNominator = mTau3D*diffusivity(mDiffFunc,aGrad(x,y,z),0);
          if (aNominator > 0.5) aNominator = 0.5;
          aNominator *= aTemp;
        }
        aData(x,y,z+1) += aNominator;
        aData(x,y,z) -= aNominator;
      }
}

