#ifndef CDiffusionH
#define CDiffusionH

#include <math.h>
#include <CVector.h>
#include <CMatrix.h>
#include <CTensor.h>
#include <CTensor4D.h>

class CDiffusion {
public:
  // constructor
  CDiffusion();
  // copy operator
  inline void operator=(CDiffusion aDiffusion);
  // isotropic diffusion, explicit scheme --------------------------------------
  // 1-D
  void isotropicDiffusion(CVector<float>& aVector, float aTime);
  inline void isotropicDiffusion(const CVector<float>& aVector, CVector<float>& aResult, float aTime);
  // 2-D
  void isotropicDiffusion(CMatrix<float>& aMatrix, float aTime);
  inline void isotropicDiffusion(const CMatrix<float>& aMatrix, CMatrix<float>& aResult, float aTime);
  void isotropicDiffusionMultiChannel(CTensor<float>& aTensor, float aTime);
  inline void isotropicDiffusionMultiChannel(const CTensor<float>& aTensor, CTensor<float>& aResult, float aTime);
  // 3-D
  void isotropicDiffusion(CTensor<float>& aTensor, float aTime);
  inline void isotropicDiffusion(const CTensor<float>& aTensor, CTensor<float>& aResult, float aTime);
  void isotropicDiffusionMultiChannel(CTensor4D<float>& aTensor, float aTime);
  inline void isotropicDiffusionMultiChannel(const CTensor4D<float>& aTensor, CTensor4D<float>& aResult, float aTime);
  // anisotropic diffusion, explicit scheme ------------------------------------
  // 2-D
  void anisotropicDiffusion(CMatrix<float>& aMatrix, float aTime);
  inline void anisotropicDiffusion(const CMatrix<float>& aMatrix, CMatrix<float>& aResult, float aTime);
  void anisotropicDiffusionMultiChannel(CTensor<float>& aTensor, float aTime, float aRho = 0.0);
  inline void anisotropicDiffusionMultiChannel(const CTensor<float>& aTensor, CTensor<float>& aResult, float aTime, float aRho = 0.0);
  // 3-D (no enhanced stencil available yet)
  void anisotropicDiffusion(CTensor<float>& aTensor, float aTime);
  inline void anisotropicDiffusion(const CTensor<float>& aTensor, CTensor<float>& aResult, float aTime);
  void anisotropicDiffusionMultiChannel(CTensor4D<float>& aTensor, float aTime);
  inline void anisotropicDiffusionMultiChannel(const CTensor4D<float>& aTensor, CTensor4D<float>& aResult, float aTime);
  // isotropic diffusion, semi-implicit AOS ------------------------------------
  // 1-D
  void isotropicDiffusionSemiImplicit(CVector<float>& aVector, float aTime);
  inline void isotropicDiffusionSemiImplicit(const CVector<float>& aVector, CVector<float>& aResult, float aTime);
  // 2-D
  void isotropicDiffusionSemiImplicit(CMatrix<float>& aMatrix, float aTime);
  inline void isotropicDiffusionSemiImplicit(const CMatrix<float>& aMatrix, CMatrix<float>& aResult, float aTime);
  void isotropicDiffusionMultiChannelSemiImplicit(CTensor<float>& aTensor, float aTime);
  inline void isotropicDiffusionMultiChannelSemiImplicit(const CTensor<float>& aTensor, CTensor<float>& aResult, float aTime);
  // 3-D
  void isotropicDiffusionSemiImplicit(CTensor<float>& aTensor, float aTime);
  inline void isotropicDiffusionSemiImplicit(const CTensor<float>& aTensor, CTensor<float>& aResult, float aTime);
  void isotropicDiffusionMultiChannelSemiImplicit(CTensor4D<float>& aTensor, float aTime);
  inline void isotropicDiffusionMultiChannelSemiImplicit(const CTensor4D<float>& aTensor, CTensor4D<float>& aResult, float aTime);
  // anisotropic diffusion, semi-implicifloat AOS ----------------------------------
  // 2-D
  void anisotropicDiffusionSemiImplicit(CMatrix<float>& aMatrix, float aTime);
  inline void anisotropicDiffusionSemiImplicit(const CMatrix<float>& aMatrix, CMatrix<float>& aResult, float aTime);
  void anisotropicDiffusionMultiChannelSemiImplicit(CTensor<float>& aTensor, float aTime);
  inline void anisotropicDiffusionMultiChannelSemiImplicit(const CTensor<float>& aTensor, CTensor<float>& aResult, float aTime);
  // Isotropic diffusion, 2-pixel numerics -------------------------------------
  // 1-D
  void twoPixelFlow(CVector<float>& aData, float aTime);
  // 2-D
  void twoPixelFlow(CMatrix<float>& aData, float aTime);
  void twoPixelFlowMultiChannel(CTensor<float>& aData, float aTime);
  void twoPixelFlowSpecialMultiChannel(CTensor<float>& aData, float aTime);
  // 3-D
  void twoPixelFlow(CTensor<float>& aData, float aTime);
  void twoPixelFlowMultiChannel(CTensor4D<float>& aData, float aTime);
  void twoPixelFlowSpecialMultiChannel(CTensor4D<float>& aData, float aTime);
  // Regularization with SOR ---------------------------------------------------
  // 1-D
  void regularization(const CVector<float>& aData, CVector<float>& aResult, float aAlpha, int aIterations = 50);
  inline void regularization(CVector<float>& aData, float aAlpha, int aIterations = 50);
  // 2-D
  void regularization(const CMatrix<float>& aData, CMatrix<float>& aResult, float aAlpha, int aIterations = 50, float aOmega = 1.9);
  inline void regularization(CMatrix<float>& aData, float aAlpha, int aIterations = 50, float aOmega = 1.9);

  // Properties ----------------------------------------------------------------
  // Time step size afloat one iteration
  // For explicifloat schemes this value musfloat be <= 0.5 for 1-D, <= 0.25 for 2-D and <= 0.16 for 3-D
  float mTau1D;
  float mTau2D;
  float mTau3D;
  // Contrasfloat parameter, steers diffusivity (see also the diffusivity functions)
  float mLambda;
  // Needed for some diffusivity functions to obtain stability
  float mEpsilon;
  // Half the power of the unbounded inverse diffusivity
  float mP;
  // Pre-smoothing for derivation computation, width of binomial filter stencil = 2*mSmoothing+1
  int mSmoothing;
  // Diffusivity function (see pre-defined constants)
  int mDiffFunc;
  // Limits the condition number of the diffusion tensor in the anisotropic case
  // to obtain stability in connection with the enhanced stencil
  bool mLimitedCondNo;
  // Uses an enhanced stencil for the anisotropic case which improves stability
  // bufloat also increases computation time (aboufloat factor 2)
  bool mEnhancedStencil;

  // Constants for diffusivity functions ---------------------------------------
  // Linear diffusion (better use equivalenfloat gaussian filtering)
  static const int cLinear = 0;
  // No backward diffusion
  static const int cCharbonnier = 1;
  // Functions with increasing backward diffusion
  // Contrasfloat parameter mLambda defines when backward diffusion starts
  static const int cPeronaMalikI = 2;
  static const int cPeronaMalikII = 3;
  static const int cWeickert = 4;
  static const int cUnboundedEpsilon = 5;
  static const int cUnbounded = 6;
  static const int cTV = 7;
  static const int cNonconvex = 8;

  // Computation of diffusivities/ diffusion tensors
  // isotropic, 1-D
  void isotropicDiffVector(const CVector<float>& aData, CVector<float>& aResult);
  // isotropic, 2-D
  void isotropicDiffMatrix(const CMatrix<float>& aData, CMatrix<float>& aResult);
  void isotropicDiffMatrix(const CMatrix<float>& aData, CMatrix<float>& aResultX, CMatrix<float>& aResultY);
  void isotropicDiffMatrixMultiChannel(const CTensor<float>& aTensor, CMatrix<float>& aResult);
  void isotropicDiffMatrixMultiChannel(const CTensor<float>& aTensor, CMatrix<float>& aResultX, CMatrix<float>& aResultY);
  // isotropic, 3-D
  void isotropicDiffTensor(const CTensor<float>& aTensor, CTensor<float>& aResult);
  void isotropicDiffTensor(const CTensor<float>& aData, CTensor<float>& aResultX, CTensor<float>& aResultY, CTensor<float>& aResultZ);
  void isotropicDiffTensorMultiChannel(const CTensor4D<float>& aTensor, CTensor<float>& aResult);
  void isotropicDiffTensorMultiChannel(const CTensor4D<float>& aTesnor, CTensor<float>& aResultX, CTensor<float>& aResultY, CTensor<float>& aResultZ);
  // anisotropic, 2-D
  void anisotropicDiffMatrix(const CMatrix<float>& aMatrix, CMatrix<float>& A, CMatrix<float>& B, CMatrix<float>& C);
  void anisotropicDiffMatrixMultiChannel(const CTensor<float>& aTensor, CMatrix<float>& A, CMatrix<float>& B, CMatrix<float>& C, float aRho = 0.0);
  // anisotropic, 3-D
  void anisotropicDiffTensor(const CTensor<float>& aTensor, CTensor<float>& A, CTensor<float>& B, CTensor<float>& C, CTensor<float>& D, CTensor<float>& E, CTensor<float>& F);
  void anisotropicDiffTensorMultiChannel(const CTensor4D<float>& aTensor, CTensor<float>& A, CTensor<float>& B, CTensor<float>& C, CTensor<float>& D, CTensor<float>& E, CTensor<float>& F);

  // Computes diffusivity functions (see also constants defined above)
  inline float diffusivity(int aChoice, float aValue, float aInvLambdaSqr);
  // Computes gradienfloat in x- and y-direction with a certain degree of presmoothing
  void presmoothedDerivative(const CVector<float>& aVector, CVector<float>& aGradient);
  void presmoothedDerivative(const CMatrix<float>& aMatrix, CMatrix<float>& aGradientX, CMatrix<float>& aGradientY);
  void presmoothedDerivative(const CTensor<float>& aTensor, CTensor<float>& aGradientX, CTensor<float>& aGradientY, CTensor<float>& aGradientZ);
  // Explicifloat diffusion schemes
  void isotropicDiffuse(CVector<float>& aVector, CVector<float>& aDiffusivity);
  void isotropicDiffuse(CMatrix<float>& aMatrix, CMatrix<float>& aDiffusivity);
  void isotropicDiffuse(CMatrix<float>& aMatrix, CMatrix<float>& GX, CMatrix<float>& GY);
  void isotropicDiffuse(CTensor<float>& aTensor, CTensor<float>& aDiffusivity);
  void anisotropicDiffuse(CMatrix<float>& aMatrix, CMatrix<float>& A, CMatrix<float>& B, CMatrix<float>& C);
  void anisotropicEnhancedDiffuse(CMatrix<float>& aMatrix, CMatrix<float>& A, CMatrix<float>& B, CMatrix<float>& C);
  void anisotropicDiffuse(CTensor<float>& aTensor, CTensor<float>& A, CTensor<float>& B, CTensor<float>& C, CTensor<float>& D, CTensor<float>& E, CTensor<float>& F);
  void anisotropicEnhancedDiffuse(CTensor<float>& aTensor, CTensor<float>& A, CTensor<float>& B, CTensor<float>& C, CTensor<float>& D, CTensor<float>& E, CTensor<float>& F);
  // AOS schemes ---------------------------------------------------------------
  void isotropicDiffuseSemiImplicit(CVector<float>& aVector, const CVector<float>& G);
  void isotropicDiffuseSemiImplicit(CMatrix<float>& aMatrix, const CMatrix<float>& GX, const CMatrix<float>& GY);
  void isotropicDiffuseSemiImplicit(CTensor<float>& aTensor, const CTensor<float>& GX, const CTensor<float>& GY, const CTensor<float>& GZ);
  void anisotropicDiffuseSemiImplicit(CMatrix<float>& aMatrix, const CMatrix<float>& A, const CMatrix<float>& B, const CMatrix<float>& C);
  // Two-pixel schemes ---------------------------------------------------------
  // Gradienfloat computation
  void twoPixelGrad(const CVector<float>& aData, CVector<float>& aGrad, int aOffset);
  void twoPixelGradX(const CMatrix<float>& aData, CMatrix<float>& aGrad, int aOffset);
  void twoPixelGradY(const CMatrix<float>& aData, CMatrix<float>& aGrad, int aOffset);
  void twoPixelGradXMultiChannel(const CTensor<float>& aData, CMatrix<float>& aGrad, int aOffset);
  void twoPixelGradYMultiChannel(const CTensor<float>& aData, CMatrix<float>& aGrad, int aOffset);
  void twoPixelGradX(const CTensor<float>& aData, CTensor<float>& aGrad, int aOffset);
  void twoPixelGradY(const CTensor<float>& aData, CTensor<float>& aGrad, int aOffset);
  void twoPixelGradZ(const CTensor<float>& aData, CTensor<float>& aGrad, int aOffset);
  void twoPixelGradXMultiChannel(const CTensor4D<float>& aData, CTensor<float>& aGrad, int aOffset);
  void twoPixelGradYMultiChannel(const CTensor4D<float>& aData, CTensor<float>& aGrad, int aOffset);
  void twoPixelGradZMultiChannel(const CTensor4D<float>& aData, CTensor<float>& aGrad, int aOffset);
  // diffusion
  void twoPixelDiffuseLeft(CVector<float>& aData, CVector<float>& aGrad);
  void twoPixelDiffuseRight(CVector<float>& aData, CVector<float>& aGrad);
  void twoPixelDiffuseLeft(CMatrix<float>& aData, CMatrix<float>& aGrad);
  void twoPixelDiffuseRight(CMatrix<float>& aData, CMatrix<float>& aGrad);
  void twoPixelDiffuseUp(CMatrix<float>& aData, CMatrix<float>& aGrad);
  void twoPixelDiffuseDown(CMatrix<float>& aData, CMatrix<float>& aGrad);
  void twoPixelDiffuseLeft(CTensor<float>& aData, CTensor<float>& aGrad);
  void twoPixelDiffuseRight(CTensor<float>& aData, CTensor<float>& aGrad);
  void twoPixelDiffuseUp(CTensor<float>& aData, CTensor<float>& aGrad);
  void twoPixelDiffuseDown(CTensor<float>& aData, CTensor<float>& aGrad);
  void twoPixelDiffuseInner(CTensor<float>& aData, CTensor<float>& aGrad);
  void twoPixelDiffuseOuter(CTensor<float>& aData, CTensor<float>& aGrad);
};

// I M P L E M E N T A T I O N -------------------------------------------------

inline void CDiffusion::operator=(CDiffusion aDiffusion) {
  mTau1D = aDiffusion.mTau1D;
  mTau2D = aDiffusion.mTau2D;
  mTau3D = aDiffusion.mTau3D;
  mLambda = aDiffusion.mLambda;
  mEpsilon = aDiffusion.mEpsilon;
  mP = aDiffusion.mP;
  mSmoothing = aDiffusion.mSmoothing;
  mDiffFunc = aDiffusion.mDiffFunc;
  mLimitedCondNo = aDiffusion.mLimitedCondNo;
  mEnhancedStencil = aDiffusion.mEnhancedStencil;
}

// Isotropic diffusion ---------------------------------------------------------

inline void CDiffusion::isotropicDiffusion(const CVector<float>& aVector, CVector<float>& aResult, float aTime) {
  aResult = aVector;
  isotropicDiffusion(aResult,aTime);
}

inline void CDiffusion::isotropicDiffusion(const CMatrix<float>& aMatrix, CMatrix<float>& aResult, float aTime) {
  aResult = aMatrix;
  isotropicDiffusion(aResult,aTime);
}

inline void CDiffusion::isotropicDiffusionMultiChannel(const CTensor<float>& aTensor, CTensor<float>& aResult, float aTime) {
  aResult = aTensor;
  isotropicDiffusionMultiChannel(aResult,aTime);
}

inline void CDiffusion::isotropicDiffusion(const CTensor<float>& aTensor, CTensor<float>& aResult, float aTime) {
  aResult = aTensor;
  isotropicDiffusion(aResult,aTime);
}

inline void CDiffusion::isotropicDiffusionMultiChannel(const CTensor4D<float>& aTensor, CTensor4D<float>& aResult, float aTime) {
  aResult = aTensor;
  isotropicDiffusionMultiChannel(aResult,aTime);
}

// Anisotropic diffusion -------------------------------------------------------

inline void CDiffusion::anisotropicDiffusion(const CMatrix<float>& aMatrix, CMatrix<float>& aResult, float aTime) {
  aResult = aMatrix;
  anisotropicDiffusion(aResult,aTime);
}

inline void CDiffusion::anisotropicDiffusionMultiChannel(const CTensor<float>& aTensor, CTensor<float>& aResult, float aTime, float aRho) {
  aResult = aTensor;
  anisotropicDiffusionMultiChannel(aResult,aTime,aRho);
}

inline void CDiffusion::anisotropicDiffusion(const CTensor<float>& aTensor, CTensor<float>& aResult, float aTime) {
  aResult = aTensor;
  anisotropicDiffusion(aResult,aTime);
}

inline void CDiffusion::anisotropicDiffusionMultiChannel(const CTensor4D<float>& aTensor, CTensor4D<float>& aResult, float aTime) {
  aResult = aTensor;
  anisotropicDiffusionMultiChannel(aResult,aTime);
}

// Semi-implicit AOS ===========================================================

inline void CDiffusion::isotropicDiffusionSemiImplicit(const CVector<float>& aVector, CVector<float>& aResult, float aTime) {
  aResult = aVector;
  isotropicDiffusionSemiImplicit(aResult,aTime);
}

inline void CDiffusion::isotropicDiffusionSemiImplicit(const CMatrix<float>& aMatrix, CMatrix<float>& aResult, float aTime) {
  aResult = aMatrix;
  isotropicDiffusionSemiImplicit(aResult,aTime);
}

inline void CDiffusion::isotropicDiffusionMultiChannelSemiImplicit(const CTensor<float>& aTensor, CTensor<float>& aResult, float aTime) {
  aResult = aTensor;
  isotropicDiffusionMultiChannelSemiImplicit(aResult,aTime);
}

inline void CDiffusion::isotropicDiffusionSemiImplicit(const CTensor<float>& aTensor, CTensor<float>& aResult, float aTime) {
  aResult = aTensor;
  isotropicDiffusionSemiImplicit(aResult,aTime);
}

inline void CDiffusion::isotropicDiffusionMultiChannelSemiImplicit(const CTensor4D<float>& aTensor, CTensor4D<float>& aResult, float aTime) {
  aResult = aTensor;
  isotropicDiffusionMultiChannelSemiImplicit(aResult,aTime);
}

// Anisotropic AOS

inline void CDiffusion::anisotropicDiffusionSemiImplicit(const CMatrix<float>& aMatrix, CMatrix<float>& aResult, float aTime) {
  aResult = aMatrix;
  anisotropicDiffusionSemiImplicit(aResult,aTime);
}

inline void CDiffusion::anisotropicDiffusionMultiChannelSemiImplicit(const CTensor<float>& aTensor, CTensor<float>& aResult, float aTime) {
  aResult = aTensor;
  anisotropicDiffusionMultiChannelSemiImplicit(aResult,aTime);
}

// Regularization --------------------------------------------------------------

inline void CDiffusion::regularization(CVector<float>& aData, float aAlpha, int aIterations) {
  CVector<float> aTemp(aData.size());
  regularization(aData,aTemp,aAlpha,aIterations);
  aData = aTemp;
}

inline void CDiffusion::regularization(CMatrix<float>& aData, float aAlpha, int aIterations, float aOmega) {
  CMatrix<float> aTemp(aData.xSize(),aData.ySize());
  regularization(aData,aTemp,aAlpha,aIterations,aOmega);
  aData = aTemp;
}

// P R I V A T E ---------------------------------------------------------------

// diffusivity
inline float CDiffusion::diffusivity(int aChoice, float aValue, float aInvLambdaSqr) {
  if (aChoice == cTV)
    return 1.0/sqrt(aValue+mEpsilon*mEpsilon);
  if (aChoice == cNonconvex)
    return 1.0/(aValue+mEpsilon*mEpsilon);
  if (aChoice == cLinear)
    return 1;
  if (aChoice == cUnboundedEpsilon)
    return pow(aValue+mEpsilon*mEpsilon,mP);
  if (aChoice == cCharbonnier)
    return 1.0/sqrt(1.0 + aValue*aInvLambdaSqr);
  if (aChoice == cPeronaMalikI)
    return 1.0/(1.0 + aValue*aInvLambdaSqr);
  if (aChoice == cPeronaMalikII)
    return exp(-aValue*aInvLambdaSqr);
  if (aChoice == cWeickert) {
    if (aValue <= 0) return 1;
    else return 1.0 - exp(-3.31488*pow(aValue*aInvLambdaSqr,-4));
  }
  if (aChoice == cUnbounded) {
    if (aValue <= 0) return 0;
    else return pow(aValue,mP);
  }
  return 1;
}

#endif

