#include "CKMeans.h"
#include <queue>
#include <NMath.h>

// kmeansRandomInit
void kmeansRandomInit(CVector<float>* aData, CVector<float>* aSum, float* aNum, CVector<int>& aLabels, CVector<int>& aLength, int K) {
  int N = aLength.size();
  bool aNullFound;
  do {
    for (int n = 0; n < N; n++) {
      aLabels(n) = (int)(K*NMath::random());
      if (aLabels(n) >= K) aLabels(n) = K-1;
    }
    kmeansRecomputeStatistics(aData,aSum,aNum,aLabels,aLength,K);
    aNullFound = false;
    for (int k = 0; k < K; k++)
      if (aNum[k] == 0.0f) { aNullFound = true; break; }
  }
  while (aNullFound);
}

// kmeansOptimize
void kmeansOptimize(CVector<float>* aData, CVector<float>* aMean, CVector<float>* aSum, float* aNum, CVector<int>& aLabels, CVector<int>& aLength, int K, int aMaxIterations = 10) {
  int N = aLength.size();
  int aDim = aData[0].size();
  int aChanged;
  int aIterations = 0;
  for (int k = 0; k < K; k++) {
    aMean[k] = aSum[k];
    if (aNum[k] > 0) aMean[k] *= 1.0f/aNum[k];
  }  
  do {
    aChanged = 0;
    for (int n = 0; n < N; n++) {
      CVector<float>& aVec1 = aData[n];
      float bestDist = 1e20f;
      int bestLabel;
      for (int k = 0; k < K; k++) {
        float dist = 0;
        CVector<float>& aVec2 = aMean[k];
        for (int i = 0; i < aDim; i++) {
          float temp = aVec1(i)-aVec2(i);
          dist += temp*temp;
        }
        if (dist < bestDist) {
          bestDist = dist;
          bestLabel = k;
        }
      }
      if (bestLabel != aLabels(n)) {
        int oldLabel = aLabels(n);
        aSum[oldLabel] -= ((float)aLength(n)*aVec1);
        aNum[oldLabel] -= aLength(n);
        aMean[oldLabel] = aSum[oldLabel];
        if (aNum[oldLabel] > 0.0f) aMean[oldLabel] *= 1.0f/aNum[oldLabel];
        aSum[bestLabel] += ((float)aLength(n)*aVec1);
        aNum[bestLabel] += aLength(n);
        aMean[bestLabel] = aSum[bestLabel];
        aMean[bestLabel] *= 1.0f/aNum[bestLabel];
        aLabels(n) = bestLabel;
        aChanged++;
      }
    }
    aIterations++;
  }
  while (aChanged > 0 && aIterations <= aMaxIterations);
  if (aChanged > 0)
    for (int k = 0; k < K; k++) {
      aMean[k] = aSum[k];
      if (aNum[k] > 0) aMean[k] *= 1.0f/aNum[k];
    }
}

// twoMeans
void twoMeans(CVector<float>* aData, CVector<float>* aMean, float* aNum, CVector<int>& aLabels, CVector<int>& aLength) {
  int aDim = aData[0].size();
  CVector<float> aSum[2]; aSum[0].setSize(aDim); aSum[1].setSize(aDim);
  kmeansRandomInit(aData,aSum,aNum,aLabels,aLength,2);
  kmeansOptimize(aData,aMean,aSum,aNum,aLabels,aLength,2,10);
}

class CState {
public:
  CState() {}
  CState(CVector<int>& aLabels) : mLabels(aLabels) {}
  int mK;
  int mSplit;
  CVector<int> mLabels;
  std::vector<float> mVariances;
  float mSplitVariance;
};

bool operator<(const CState& a, const CState& b) {
  if (a.mSplitVariance == b.mSplitVariance) return a.mK < b.mK;
  else return a.mSplitVariance < b.mSplitVariance;
}

#if 0
void showLabels2(CVector<int>& aLabels, int K, CVector<int>& aNodeToIndex, std::vector<CTrack>& aTracks, int aXSize, int aYSize, int aSequenceLength, int aCase) {
  for (int t = 0; t < aSequenceLength; t++) {
    CMatrix<float>* aShow = new CMatrix<float>[K];
    for (int i = 0; i < K; i++) {
      aShow[i].setSize(aXSize,aYSize);
      aShow[i] = 0;
    }
    for (int ai = 0; ai < aLabels.size(); ai++) {
      int i = aNodeToIndex(ai);
      int aTime = t-aTracks[i].mSetupTime;
      if (aTime < 0) continue;
      if ((int)aTracks[i].mx.size() <= aTime) continue;
      int x,y;
      if (aTime > 0) {
        x = (int)aTracks[i].mx[aTime-1];
        y = (int)aTracks[i].my[aTime-1];
      }
      else {
        x = aTracks[i].mox;
        y = aTracks[i].moy;
      }
      int r = 2;
      int x1 = x-r; if (x1 < 0) x1 = 0;
      int y1 = y-r; if (y1 < 0) y1 = 0;
      int x2 = x+r; if (x2 >= aShow[0].xSize()) x2 = aShow[0].xSize()-1;
      int y2 = y+r; if (y2 >= aShow[0].ySize()) y2 = aShow[0].ySize()-1;
      for (int k = 0; k < K; k++)
        for (int ay = y1; ay <= y2; ay++)
          for (int ax = x1; ax <= x2; ax++)
            if (aLabels(ai) == k) aShow[k](ax,ay) = 255;
            else aShow[k](ax,ay) = 0;
    }
    int row = (int)ceil(sqrt(K));
    CMatrix<float> aAll(row*aXSize,row*aYSize,0);
    char buffer[40];
    for (int k = 0; k < K; k++)
      aAll.paste(aShow[k],(k/row)*aXSize,(k%row)*aYSize);
    sprintf(buffer,"kmeans%02d_%02d.pgm",aCase,t);
    aAll.downsample(aAll.xSize()/2,aAll.ySize()/2);
    aAll.writeToPGM(buffer);
    delete[] aShow;
  }
}

void showDifference(CVector<int>& aLabelsA, CVector<int>& aLabelsB, int K, CVector<int>& aNodeToIndex, std::vector<CTrack>& aTracks, int aXSize, int aYSize, int aSequenceLength) {
  for (int t = 0; t < aSequenceLength; t++) {
    CMatrix<float> aShow(aXSize,aYSize,0);
    for (int ai = 0; ai < aLabels.size(); ai++) {
      int i = aNodeToIndex(ai);
      int aTime = t-aTracks[i].mSetupTime;
      if (aTime < 0) continue;
      if ((int)aTracks[i].mx.size() <= aTime) continue;
      int x,y;
      if (aTime > 0) {
        x = (int)aTracks[i].mx[aTime-1];
        y = (int)aTracks[i].my[aTime-1];
      }
      else {
        x = aTracks[i].mox;
        y = aTracks[i].moy;
      }
      int r = 2;
      int x1 = x-r; if (x1 < 0) x1 = 0;
      int y1 = y-r; if (y1 < 0) y1 = 0;
      int x2 = x+r; if (x2 >= aShow[0].xSize()) x2 = aShow[0].xSize()-1;
      int y2 = y+r; if (y2 >= aShow[0].ySize()) y2 = aShow[0].ySize()-1;
      for (int k = 0; k < K; k++)
        for (int ay = y1; ay <= y2; ay++)
          for (int ax = x1; ax <= x2; ax++)
            if (aLabelsA(ai) != aLabelsB(ai)) aShow(ax,ay) = 255;
    }
    char buffer[40];
    sprintf(buffer,"kmeansdiff%02d.pgm",t);
    aShow.downsample(aXSize/2,aYSize/2);
    aShow.writeToPGM(buffer);
  }
}
#endif

// kmeansHierarchicalInit
void kmeansHierarchicalInit(CVector<float>* aData, CVector<int>& aLength, std::vector<CState>& aInits, int K, unsigned int aMaxInits, CVector<int>& aNodeToIndex, std::vector<CTrack>& aTracks, int aXSize, int aYSize, int aSequenceLength) {
  int N = aLength.size();
  int aDim = aData[0].size();
  CVector<int> aLabels(N);
  CVector<float> aMean[2]; aMean[0].setSize(aDim); aMean[1].setSize(aDim);
  float aNum[2];
  std::priority_queue<CState> aQueue;
  twoMeans(aData,aMean,aNum,aLabels,aLength);
  if (aNum[0] <= 0 || aNum[1] <= 0) return;
  float aVar[2] = {0,0};
  for (int n = 0; n < N; n++) {
    CVector<float>& aVec1 = aData[n];
    CVector<float>& aVec2 = aMean[aLabels(n)];
    float dist = 0.0f;
    for (int i = 0; i < aDim; i++) {
      float temp = aVec1(i)-aVec2(i);
      dist += temp*temp;
    }
    aVar[aLabels(n)] += (float)aLength(n)*dist;
  }
  aVar[0] /= aNum[0];
  aVar[1] /= aNum[1];
  CState aState;
  aState.mK = 2;
  aState.mLabels = aLabels;
  aState.mVariances.push_back(aVar[0]);
  aState.mVariances.push_back(aVar[1]);
  aState.mSplit = 0; aState.mSplitVariance = aVar[0];
  aQueue.push(aState);
  aState.mSplit = 1; aState.mSplitVariance = aVar[1];
  aQueue.push(aState);
  while (!aQueue.empty() && aInits.size() < aMaxInits) {
    CState aState = aQueue.top(); aQueue.pop();
    // Split designated cluster; ignore result when an empty cluster appears
    int num = 0;
    for (int n = 0; n < N; n++)
      if (aState.mLabels(n) == aState.mSplit) num++;
    if (num <= 1) continue;
    CVector<float>* aLocalData = new CVector<float>[num];
    CVector<int> aLocalLength(num);
    CVector<int> aNumToN(num);
    CVector<int> aLocalLabels(num);
    num = 0;
    for (int n = 0; n < N; n++)
      if (aState.mLabels(n) == aState.mSplit) {
        aLocalData[num] = aData[n];
        aLocalLength[num] = aLength[n];
        aNumToN(num) = n;
        num++;
      }
    twoMeans(aLocalData,aMean,aNum,aLocalLabels,aLocalLength);
    if (aNum[0] <= 0 || aNum[1] <= 0) continue;
    // Compute variances of new parts
    float aVar[2] = {0,0};
    for (int n = 0; n < num; n++) {
      CVector<float>& aVec1 = aLocalData[n];
      CVector<float>& aVec2 = aMean[aLocalLabels(n)];
      float dist = 0.0f;
      for (int i = 0; i < aDim; i++) {
        float temp = aVec1(i)-aVec2(i);
        dist += temp*temp;
      }
      aVar[aLocalLabels(n)] += (float)aLocalLength(n)*dist;
    }
    aVar[0] /= aNum[0];
    aVar[1] /= aNum[1];
    // Transfer split to labels
    int aK = aState.mK+1;
    int aLabel0 = aState.mSplit;
    int aLabel1 = aState.mK;
    for (int n = 0; n < num; n++)
      if (aLocalLabels(n) > 0) aState.mLabels(aNumToN(n)) = aLabel1;
    delete[] aLocalData;
    // Final region number reached -> save this labeling as one initialization
    if (aK == K) aInits.push_back(CState(aState.mLabels));
    // Otherwise create new states and push to queue
    else {
      aState.mK = aK;
      aState.mVariances[aLabel0] = aVar[0];
      aState.mVariances.push_back(aVar[1]);
      for (int k = 0; k < aK; k++) {
        aState.mSplit = k;
        aState.mSplitVariance = aState.mVariances[k];
        if (aQueue.size() < 40 || aState.mSplitVariance > aQueue.top().mSplitVariance) aQueue.push(aState);
      }
    }

  }
}

// P U B L I C -----------------------------------------------------------------

// kmeans
void kmeans(CVector<float>* aData, CVector<float>* aMean, float* aNum, CVector<int>& aLabels, CVector<int>& aLength, CVector<std::vector<float> >& g, CVector<std::vector<int> >& idx, int K, float aNu, CVector<int>& aNodeToIndex, std::vector<CTrack>& aTracks, int aXSize, int aYSize, int aSequenceLength) {
  int N = aLength.size();
  float bestEnergy = 1e20f;
  CVector<int> bestLabels;
  CVector<float>* aSum = new CVector<float>[K];
  for (int k = 0; k < K; k++)
    aSum[k].setSize(aData[0].size());
  // Random intialization
  for (int t2 = 0; t2 < 10; t2++) {
    kmeansRandomInit(aData,aSum,aNum,aLabels,aLength,K);
    kmeansOptimize(aData,aMean,aSum,aNum,aLabels,aLength,K);
    float aEnergy = computeEnergy(aData,aMean,aLabels,aLength);
    if (aEnergy < bestEnergy) {
      bestLabels = aLabels;
      bestEnergy = aEnergy;
    }
  }
  // Hierarchical initialization
  if (K > 2) {
    std::vector<CState> aInits;
    kmeansHierarchicalInit(aData,aLength,aInits,K,20,aNodeToIndex,aTracks,aXSize,aYSize,aSequenceLength);
    for (unsigned int i = 0; i < aInits.size(); i++) {
      aLabels = aInits[i].mLabels;
      kmeansRecomputeStatistics(aData,aSum,aNum,aLabels,aLength,K);
      kmeansOptimize(aData,aMean,aSum,aNum,aLabels,aLength,K);
      float aEnergy = computeEnergy(aData,aMean,aLabels,aLength);
      if (aEnergy < bestEnergy) {
        bestLabels = aLabels;
        bestEnergy = aEnergy;
      }
    }
  }
  aLabels = bestLabels;
  // Recompute means and nums for best labels
  for (int k = 0; k < K; k++) {
    aSum[k] = 0.0f;
    aNum[k] = 0.0f;
  }
  for (int n = 0; n < N; n++) {
    aSum[aLabels(n)] += ((float)aLength(n)*aData[n]);
    aNum[aLabels(n)] += aLength(n);
  }
  for (int k = 0; k < K; k++) {
    aMean[k] = aSum[k];
    if (aNum[k] > 0) aMean[k] *= 1.0f/aNum[k];
  }
  delete[] aSum;
}

// computeEnergy
float computeEnergy(CVector<float>* aData, CVector<float>* aMean, CVector<int>& aLabels, CVector<int>& aLength) {
  float aEnergy = 0.0f;
  int aDim = aData[0].size();
  int N = aLength.size();
  for (int n = 0; n < N; n++) {
    // Unitary terms
    CVector<float>& aVec1 = aData[n];
    CVector<float>& aVec2 = aMean[aLabels(n)];
    float dist = 0.0f;
    for (int i = 0; i < aDim; i++) {
      float temp = aVec1(i)-aVec2(i);
      dist += temp*temp;
    }
    aEnergy += (float)aLength(n)*dist;
  }
  return aEnergy;
}

float computeEnergy(CVector<float>* aData, CVector<float>* aMean, CVector<int>& aLabels, CVector<int>& aLength, CVector<std::vector<float> >& g, CVector<std::vector<int> >& idx, float aNu) {
  // Unitary terms
  float aEnergy = computeEnergy(aData,aMean,aLabels,aLength);
  // Pairwise terms
  int N = aLength.size();
  for (int n = 0; n < N; n++) {
    float aPairwise = 0.0f;
    for (unsigned int r = 0; r < idx(n).size(); r++)
      if (aLabels(idx(n)[r]) != aLabels(n)) aPairwise += g(n)[r];
    aEnergy += aLength(n)*aNu*aPairwise;
  }
  return aEnergy;
}

// kmeansRecomputeStatistics
void kmeansRecomputeStatistics(CVector<float>* aData, CVector<float>* aSum, float* aNum, CVector<int>& aLabels, CVector<int>& aLength, int K) {
  int N = aLength.size();
  for (int k = 0; k < K; k++) {
    aSum[k] = 0.0f;
    aNum[k] = 0.0f;
  }
  for (int n = 0; n < N; n++) {
    aSum[aLabels(n)] += ((float)aLength(n)*aData[n]);
    aNum[aLabels(n)] += aLength(n);
  }
}


