#include "CMatching.h"
#include <CFilter.h>
#include <COpticFlow.h>
#include "CGraph.h"

// CTrack file I/O -------------------------------------------------------------
void CTrack::readFromFile(std::ifstream& aIn) {
  aIn >> mox >> moy;
  int aCount;
  aIn >> aCount;
  float dummy;
  for (int i = 0; i < aCount; i++) {
    aIn >> dummy; mx.push_back(dummy);
    aIn >> dummy; my.push_back(dummy);
    aIn >> dummy; mFlowVar.push_back(dummy);
  }
  aIn >> mLabel >> mStopped >> mScale >> mSetupTime >> aCount;
  mProjection.setSize(aCount);
  for (int i = 0; i < mProjection.size(); i++)
    aIn >> mProjection(i);
  aIn >> aCount;
//  mNearestNeighbors.setSize(aCount);
//  mNearestNeighborsDX.setSize(aCount);
//  mNearestNeighborsDY.setSize(aCount);
  for (int i = 0; i < aCount; i++)
    aIn >> mNearestNeighbors(i) >> mNearestNeighborsDX(i) >> mNearestNeighborsDY(i);
}

void CTrack::writeToFile(std::ofstream& aOut) {
  aOut << mox << " " << moy << std::endl;
  aOut << mx.size() << std::endl;
  for (unsigned int i = 0; i < mx.size(); i++)
    aOut << mx[i] << " " << my[i] << " " << mFlowVar[i] << std::endl;
  aOut << mLabel << " " << mStopped << " " << mScale << " " << mSetupTime << std::endl;
  aOut << mProjection.size() << std::endl;
  for (int i = 0; i < mProjection.size(); i++)
    aOut << mProjection(i) << " ";
  aOut << std::endl;
  aOut << mNearestNeighbors.size() << std::endl;
  for (int i = 0; i < mNearestNeighbors.size(); i++)
    aOut << mNearestNeighbors(i) << " " << mNearestNeighborsDX(i) << " " << mNearestNeighborsDY(i) << " ";
}

// computeHOG ------------------------------------------------------------------
void CMatching::computeHOG(CTensor<float>& aBox, CTensor<float>& aHOG) {
  float inv15pi = 15.0/(2.0*NMath::Pi);
  int aXSize = aBox.xSize();
  int aYSize = aBox.ySize();
  int aSize = aXSize*aYSize;
  aHOG.setSize(aXSize,aYSize,15); aHOG = 0;
  CDerivative<float> aDerivative(3);
  CTensor<float> dx(aXSize,aYSize,aBox.zSize());
  CTensor<float> dy(aXSize,aYSize,aBox.zSize());
  NFilter::filter(aBox,dx,aDerivative,1,1);
  NFilter::filter(aBox,dy,1,aDerivative,1);
  int i2 = 0;
  for (int z = 0; z < aBox.zSize(); z++)
    for (int i = 0; i < aSize; i++,i2++) {
      float gradx = dx.data()[i2];
      float grady = dy.data()[i2];
      if (gradx == 0 && grady == 0) continue;
      float mag = sqrt(gradx*gradx+grady*grady);
      float ori = atan2(grady,gradx)+NMath::Pi;  // now between 0 and 2pi
      int dori = (int)(ori*inv15pi);
      if (dori == 15) dori = 0;
      aHOG.data()[i+dori*aSize] += mag;
    }
  CSmooth<float> aGauss(0.8,1.0);
  NFilter::filter(aHOG,1,1,aGauss);
  for (int i = 0; i < aSize; i++) {
    float temp1 = aHOG.data()[i];
    float temp2 = aHOG.data()[i+14*aSize];
    aHOG.data()[i] = aGauss(-1)*temp2+(1.0-aGauss(-1))*temp1;
    aHOG.data()[i+14*aSize] = aGauss(1)*temp1+(1.0-aGauss(-1))*temp2;
  }
  NFilter::boxFilterX(aHOG,7);
  NFilter::boxFilterY(aHOG,7);
}

// computeCorners --------------------------------------------------------------
void CMatching::computeCorners(CTensor<float>& aImage, CMatrix<float>& aCorners, float aRho) {
  aCorners.setSize(aImage.xSize(),aImage.ySize());
  int aXSize = aImage.xSize();
  int aYSize = aImage.ySize();
  int aSize = aXSize*aYSize;
  // Compute gradient
  CTensor<float> dx(aXSize,aYSize,aImage.zSize());
  CTensor<float> dy(aXSize,aYSize,aImage.zSize());
  CDerivative<float> aDerivative(3);
  NFilter::filter(aImage,dx,aDerivative,1,1);
  NFilter::filter(aImage,dy,1,aDerivative,1);
  // Compute second moment matrix
  CMatrix<float> dxx(aXSize,aYSize,0);
  CMatrix<float> dyy(aXSize,aYSize,0);
  CMatrix<float> dxy(aXSize,aYSize,0);
  int i2 = 0;
  for (int k = 0; k < aImage.zSize(); k++)
    for (int i = 0; i < aSize; i++,i2++) {
      dxx.data()[i] += dx.data()[i2]*dx.data()[i2];
      dyy.data()[i] += dy.data()[i2]*dy.data()[i2];
      dxy.data()[i] += dx.data()[i2]*dy.data()[i2];
    }
  // Smooth second moment matrix
  NFilter::recursiveSmoothX(dxx,aRho);
  NFilter::recursiveSmoothY(dxx,aRho);
  NFilter::recursiveSmoothX(dyy,aRho);
  NFilter::recursiveSmoothY(dyy,aRho);
  NFilter::recursiveSmoothX(dxy,aRho);
  NFilter::recursiveSmoothY(dxy,aRho);
  // Compute smallest eigenvalue
  for (int i = 0; i < aSize; i++) {
    float a = dxx.data()[i];
    float b = dxy.data()[i];
    float c = dyy.data()[i];
    float temp = 0.5*(a+c);
    float temp2 = temp*temp+b*b-a*c;
    if (temp2 < 0.0f) aCorners.data()[i] = 0.0f;
    else aCorners.data()[i] = temp-sqrt(temp2);
  }
}

// hogMatching -----------------------------------------------------------------
void CMatching::hogMatching(std::vector<CMatch>& aPoints, CTensor<float>& aImage1, CTensor<float>& aImage2, CTensor<float>& aHOG1, CTensor<float>& aHOG2, CMatrix<float>& aEnergy1, CMatrix<float>& aEnergy2, int rx1, int ry1, int rx2, int ry2, CMatrix<bool>& aAllowed) {
  int aXSize = aHOG1.xSize();
  int aYSize = aHOG1.ySize();
  CVector<float> aDesc1(9*aHOG1.zSize());
  CVector<float> aDesc2(9*aHOG1.zSize());
  for (unsigned int match = 0; match < aPoints.size(); match++) {
    if (aPoints[match].mStopped) continue;
    int ax = aPoints[match].mox;
    int ay = aPoints[match].moy;
    // Descriptor in first image
    int i = 0;
    for (int y = -4; y <= 4; y+=4)
      for (int x = -4; x <= 4; x+=4)
        for (int k = 0; k < aHOG1.zSize(); k++,i++)
          aDesc1(i) = aHOG1(ax+x,ay+y,k);
    // Compute search radius in second image
    int x1 = ax+rx1; if (x1 < 4) x1 = 4;
    int y1 = ay+ry1; if (y1 < 4) y1 = 4;
    int x2 = ax+rx2; if (x2 >= aXSize-4) x2 = aXSize-5;
    int y2 = ay+ry2; if (y2 >= aYSize-4) y2 = aYSize-5;
    // Find best match in second image
    int bestX,bestY;
    float bestDist = 1e8f;
    float best2Dist = 1e8f;
    for (int by = y1; by <= y2; by++)
      for (int bx = x1; bx <= x2; bx++) {
        if (aAllowed(bx-ax-rx1,by-ay-ry1) == false) continue;
        if (aEnergy2(bx,by) > 1.25f*aEnergy1(ax,ay) || aEnergy1(ax,ay) > 1.25f*aEnergy2(bx,by)) continue;
        float temp = aImage1(ax,ay,0)-aImage2(bx,by,0);
        float dist = temp*temp;
        temp = aImage1(ax,ay,1)-aImage2(bx,by,1);
        dist += temp*temp;
        temp = aImage1(ax,ay,2)-aImage2(bx,by,2);
        dist += temp*temp;
        if (dist > 300) continue;
        dist = 0.0f;
        int i = 0;
        for (int y = -4; y <= 4; y+=4)
          for (int x = -4; x <= 4; x+=4)
            for (int k = 0; k < aHOG1.zSize(); k++,i++)
              aDesc2(i) = aHOG2(bx+x,by+y,k);
        for (int i = 0; i < aDesc1.size(); i++) {
          temp = aDesc1(i)-aDesc2(i);
          dist += temp*temp;
        }
        if (dist < bestDist) {
          best2Dist = bestDist;
          bestDist = dist; bestX = bx; bestY = by;
        }
        else if (dist < best2Dist) best2Dist = dist;
      }
    if (bestDist >= 1e8f) {aPoints[match].mStopped = true; continue; }
    // Consistency check: descriptor in first image best match of descriptor in second image?
    i = 0;
    for (int y = -4; y <= 4; y+=4)
      for (int x = -4; x <= 4; x+=4)
        for (int k = 0; k < aHOG1.zSize(); k++,i++)
          aDesc2(i) = aHOG2(bestX+x,bestY+y,k);
    x1 = bestX-rx2; if (x1 < 4) x1 = 4;
    y1 = bestY-ry2; if (y1 < 4) y1 = 4;
    x2 = bestX-rx1; if (x2 >= aXSize-4) x2 = aXSize-5;
    y2 = bestY-ry1; if (y2 >= aYSize-4) y2 = aYSize-5;
    bestDist = 1e8f;
    int newBestX,newBestY;
    for (int by = y1; by <= y2; by++)
      for (int bx = x1; bx <= x2; bx++) {
        if (aAllowed(bestX-bx-rx1,bestY-by-ry1) == false) continue;
        if (aEnergy2(bestX,bestY) > 1.25f*aEnergy1(bx,by) || aEnergy1(bx,by) > 1.25f*aEnergy2(bestX,bestY)) continue;
        float temp = aImage1(bx,by,0)-aImage2(bestX,bestY,0);
        float dist = temp*temp;
        temp = aImage1(bx,by,1)-aImage2(bestX,bestY,1);
        dist += temp*temp;
        temp = aImage1(bx,by,2)-aImage2(bestX,bestY,2);
        dist += temp*temp;
        if (dist > 300) continue;
        dist = 0.0f;
        int i = 0;
        for (int y = -4; y <= 4; y+=4)
          for (int x = -4; x <= 4; x+=4)
            for (int k = 0; k < aHOG1.zSize(); k++,i++)
              aDesc1(i) = aHOG1(bx+x,by+y,k);
        for (int i = 0; i < aDesc1.size(); i++) {
          temp = aDesc1(i)-aDesc2(i);
          dist += temp*temp;
        }
        if (dist < bestDist) {
          bestDist = dist;
          newBestX = bx;
          newBestY = by;
        }
      }
    // Consistency check passed?
    if (abs(newBestX-ax) <= 1 && abs(newBestY-ay) <= 1) {
      aPoints[match].mx = bestX;
      aPoints[match].my = bestY;
      if (best2Dist >= 1e8f) best2Dist = 10.0f*bestDist;
      aPoints[match].mScore = (best2Dist-bestDist)/(bestDist+1e-6);
    }
    else aPoints[match].mStopped = true;
  }
}

// findMotions -----------------------------------------------------------------
void CMatching::findMotions(CTensor<float>& aImage1, CTensor<float>& aImage2, CTensor<float>& aHOG1, CTensor<float>& aHOG2, CMatrix<float>& aEnergy1, CMatrix<float>& aEnergy2, CMatrix<float>& aCorners, float aAverageCornerness, int& rx1, int& ry1, int& rx2, int& ry2, CMatrix<bool>& aAllowed) {
  int aXSize = aImage1.xSize();
  int aYSize = aImage2.ySize();
  // Initialize search radius
  int aFirstStep = 1;
  int dummy = 20;
  while (dummy < aXSize) {
    dummy <<= 1; aFirstStep <<= 1;
  }
  int r = (int)(0.3*aXSize);
  rx1 = -r;
  ry1 = -r;
  rx2 = r;
  ry2 = r;
  aAllowed.setSize(rx2-rx1+1,ry2-ry1+1);
  aAllowed = true;
  // Coarse-to-fine strategy: first find dominant motions on subsampled features, then refine area of possible motions
  for (int aStep = aFirstStep; aStep >= 8; aStep >>= 1) {
    // Set up initial points for matching (simple spatial subsampling)
    std::vector<CMatch> aPoints;
    for (int ay = 4; ay < aYSize-4; ay+=aStep)
      for (int ax = 4; ax < aXSize-4; ax+=aStep) {
        if (aCorners(ax,ay) < 0.5*aAverageCornerness) continue;
        aPoints.push_back(CMatch());
        CMatch& newMatch = aPoints.back();
        newMatch.mox = ax;
        newMatch.moy = ay;
      }
    // Compute correspondences
    hogMatching(aPoints,aImage1,aImage2,aHOG1,aHOG2,aEnergy1,aEnergy2,rx1,ry1,rx2,ry2,aAllowed);
    // Compute motion clusters
    for (unsigned int i = 0; i < aPoints.size(); i++)
      aPoints[i].mLabel = -1;
    fastOneFrameClustering(aPoints);
    // Bound on possible motion vectors
    int minu = 10000; int maxu = -10000;
    int minv = 10000; int maxv = -10000;
    for (unsigned int i = 0; i < aPoints.size(); i++) {
      if (aPoints[i].mStopped) continue;
      int u = (int)aPoints[i].mx-aPoints[i].mox;
      int v = (int)aPoints[i].my-aPoints[i].moy;
      if (u < minu) minu = u;
      if (u > maxu) maxu = u;
      if (v < minv) minv = v;
      if (v > maxv) maxv = v;
    }
    int var = aStep >> 1;
    rx1 = minu-var;
    ry1 = minv-var;
    rx2 = maxu+var;
    ry2 = maxv+var;
    aAllowed.setSize(rx2-rx1+1,ry2-ry1+1);
    aAllowed = false;
    for (unsigned int i = 0; i < aPoints.size(); i++) {
      if (aPoints[i].mStopped) continue;
      int u = (int)aPoints[i].mx-aPoints[i].mox;
      int v = (int)aPoints[i].my-aPoints[i].moy;
      aAllowed(u-rx1,v-ry1) = true;
    }
    // Dilation of allowed field
    int aXSize = aAllowed.xSize();
    int aYSize = aAllowed.ySize();
    for (int y = 0; y < aYSize; y++) {
      CVector<int> aDist(aXSize);
      // Forward sweep
      if (aAllowed(0,y)) aDist(0) = 0;
      else aDist(0) = var+1;
      for (int x = 1; x < aXSize; x++)
        if (aAllowed(x,y)) aDist(x) = 0;
        else aDist(x) = aDist(x-1)+1;
      for (int x = 1; x < aXSize; x++)
        if (aDist(x) <= var) aAllowed(x,y) = true;
      // Backward sweep
      if (aAllowed(aXSize-1,y)) aDist(aXSize-1) = 0;
      else aDist(aXSize-1) = var+1;
      for (int x = aXSize-2; x >= 0; x--)
        if (aAllowed(x,y)) aDist(x) = 0;
        else aDist(x) = aDist(x+1)+1;
      for (int x = 0; x < aXSize-1; x++)
        if (aDist(x) <= var) aAllowed(x,y) = true;
    }
    for (int x = 0; x < aXSize; x++) {
      CVector<int> aDist(aYSize);
      // Forward sweep
      if (aAllowed(x,0)) aDist(0) = 0;
      else aDist(0) = var+1;
      for (int y = 1; y < aYSize; y++)
        if (aAllowed(x,y)) aDist(y) = 0;
        else aDist(y) = aDist(y-1)+1;
      for (int y = 1; y < aYSize; y++)
        if (aDist(y) <= var) aAllowed(x,y) = true;
      // Backward sweep
      if (aAllowed(x,aYSize-1)) aDist(aYSize-1) = 0;
      else aDist(aYSize-1) = var+1;
      for (int y = aYSize-2; y >= 0; y--)
        if (aAllowed(x,y)) aDist(y) = 0;
        else aDist(y) = aDist(y+1)+1;
      for (int y = 0; y < aYSize-1; y++)
        if (aDist(y) <= var) aAllowed(x,y) = true;
    }
  }
}

// hogFlow ---------------------------------------------------------------------
void CMatching::hogFlow(CTensor<float>& aImage1, CTensor<float>& aImage2, CTensor<float>& aHOG1, CTensor<float>& aHOG2, CMatrix<float>& aEnergy1, CMatrix<float>& aEnergy2, CTensor<float>& aForward, CTensor<float>& aBackward) {
  int aStep = 4;
  int aXSize = aImage1.xSize();
  int aYSize = aImage1.ySize();
  int aSize = aXSize*aYSize;
  // Set up points for matching (Harris detector)
  CMatrix<float> aCorners;
  computeCorners(aImage1,aCorners,2.5f);
  float aAverage = 0.0f;
  for (int i = 0; i < aSize; i++)
    aAverage += aCorners.data()[i];
  aAverage /= aSize;
  // Find restricted search area by sparse, coarse-to-fine blockmatching
  int rx1,ry1,rx2,ry2;
  CMatrix<bool> aAllowed;
  findMotions(aImage1,aImage2,aHOG1,aHOG2,aEnergy1,aEnergy2,aCorners,aAverage,rx1,ry1,rx2,ry2,aAllowed);
  // Forward flow ---------------------------------------
  std::vector<CMatch> aPoints;
  for (int ay = 8; ay < aYSize-8; ay+=aStep)
    for (int ax = 8; ax < aXSize-8; ax+=aStep) {
      if (aCorners(ax,ay) < 0.125*aAverage) continue;
      aPoints.push_back(CMatch());
      CMatch& newMatch = aPoints.back();
      newMatch.mox = ax;
      newMatch.moy = ay;
    }
  // Compute correspondences
  hogMatching(aPoints,aImage1,aImage2,aHOG1,aHOG2,aEnergy1,aEnergy2,rx1,ry1,rx2,ry2,aAllowed);
  // Run variational optical flow
  CTensor<float> aGeoU(aXSize,aYSize,1);
  CTensor<float> aGeoV(aXSize,aYSize,1);
  CTensor<float> aGeoConfidence(aXSize,aYSize,1,0);
  for (unsigned int i = 0; i < aPoints.size(); i++) {
    if (aPoints[i].mStopped) continue;
    int x = aPoints[i].mox;
    int y = aPoints[i].moy;
    aGeoU(x,y,0) = aPoints[i].mx-aPoints[i].mox;
    aGeoV(x,y,0) = aPoints[i].my-aPoints[i].moy;
    aGeoConfidence(x,y,0) = aPoints[i].mScore;
  }
  COpticFlow::warpingGeometric(aImage1,aImage2,aForward,aGeoU,aGeoV,aGeoConfidence,0,0.0f,70.0f,400.0f,5.0f,0.0f,0.95f,3,5,1.85f);
  // Backward flow --------------------------------------
  // Set up points for matching (Harris detector)
  computeCorners(aImage2,aCorners,2.5f);
  aAverage = 0.0f;
  for (int i = 0; i < aSize; i++)
    aAverage += aCorners.data()[i];
  aAverage /= aSize;
  aPoints.clear();
  for (int ay = 8; ay < aYSize-8; ay+=aStep)
    for (int ax = 8; ax < aXSize-8; ax+=aStep) {
      if (aCorners(ax,ay) < 0.125*aAverage) continue;
      aPoints.push_back(CMatch());
      CMatch& newMatch = aPoints.back();
      newMatch.mox = ax;
      newMatch.moy = ay;
    }
  // Compute correspondences
  CMatrix<bool> aAllowedBack(aAllowed.xSize(),aAllowed.ySize());
  for (int y = 0; y < aAllowed.ySize(); y++)
    for (int x = 0; x < aAllowed.xSize(); x++)
      aAllowedBack(aAllowed.xSize()-x-1,aAllowed.ySize()-y-1) = aAllowed(x,y);
  hogMatching(aPoints,aImage2,aImage1,aHOG2,aHOG1,aEnergy2,aEnergy1,-rx2,-ry2,-rx1,-ry1,aAllowedBack);
  // Run variational optical flow
  aGeoConfidence = 0;
  for (unsigned int i = 0; i < aPoints.size(); i++) {
    if (aPoints[i].mStopped) continue;
    int x = aPoints[i].mox;
    int y = aPoints[i].moy;
    aGeoU(x,y,0) = aPoints[i].mx-aPoints[i].mox;
    aGeoV(x,y,0) = aPoints[i].my-aPoints[i].moy;
    aGeoConfidence(x,y,0) = aPoints[i].mScore;
  }
  COpticFlow::warpingGeometric(aImage2,aImage1,aBackward,aGeoU,aGeoV,aGeoConfidence,0,0.0f,70.0f,400.0f,5.0f,0.0f,0.95f,3,5,1.85f);
}

// P R I V A T E ---------------------------------------------------------------

// fastOneFrameClustering ------------------------------------------------------
// consists of several subroutines that build an affinity matrix and run the Lanzcos method

// computeDistance
inline float computeDistance(float u1, float v1, float u2, float v2) {
  float temp = u1-u2;
  float dist = temp*temp;
  temp = v1-v2;
  dist += temp*temp;
  float mag1 = u1*u1+v1*v1+4;
  float mag2 = u2*u2+v2*v2+4;
  dist /= NMath::min(mag1,mag2);
  return dist;
}

// Efficient graph structure for a purely translational motion model between two frames
// Idea: Exactly the same motion appears multiple times, so nodes can be merged
bool buildGraph(std::vector<CMatch>& aPoints, std::vector<int>& aValidMatchs, CMatrix<std::vector<unsigned int> >& aMatchIndex, std::vector<int>& aList, CGraph& aEdges) {
  // Determine area of existing motions
  int minu = 10000; int maxu = -10000;
  int minv = 10000; int maxv = -10000;
  for (unsigned int ai = 0; ai < aValidMatchs.size(); ai++) {
    int i = aValidMatchs[ai];
    int u = (int)aPoints[i].mx-aPoints[i].mox;
    int v = (int)aPoints[i].my-aPoints[i].moy;
    if (u < minu) minu = u;
    if (u > maxu) maxu = u;
    if (v < minv) minv = v;
    if (v > maxv) maxv = v;
  }
  if (minu > maxu || minv > maxv) return false;
  // For each motion, save references to the tracks with this motion
  aMatchIndex.setSize(maxu-minu+1,maxv-minv+1);
  for (unsigned int ai = 0; ai < aValidMatchs.size(); ai++) {
    int i = aValidMatchs[ai];
    int u = (int)aPoints[i].mx-aPoints[i].mox;
    int v = (int)aPoints[i].my-aPoints[i].moy;
    aMatchIndex(u-minu,v-minv).push_back(i);
  }
  // Build a list with all existing motions -> nodes in the reduced graph
  aList.clear();
  for (int i = 0; i < aMatchIndex.size(); i++)
    if (aMatchIndex.data()[i].size() > 0) {
      aList.push_back(i);
      aEdges.push_back(std::vector<CEdge>());
    }
  // Build a graph from this list
  for (unsigned int i = 0; i < aList.size(); i++) {
    int u1 = aList[i] % aMatchIndex.xSize();
    int v1 = aList[i] / aMatchIndex.xSize();
    for (unsigned int j = i+1; j < aList.size(); j++) {
      int u2 = aList[j] % aMatchIndex.xSize();
      int v2 = aList[j] / aMatchIndex.xSize();
      float w = exp(-computeDistance(u1,v1,u2,v2));
      if (w > 0.001) {
        w *= aMatchIndex.data()[aList[i]].size()*aMatchIndex.data()[aList[j]].size();
        aEdges[i].push_back(CEdge(j,w));
        aEdges[j].push_back(CEdge(i,w));
      }
    }
  }
  return true;
}

void CMatching::fastOneFrameClustering(std::vector<CMatch>& aPoints) {
  // Successively extract dominant cluster from point cloud
  std::vector<int> aValidMatchs;
  for (unsigned int i = 0; i < aPoints.size(); i++)
    if (aPoints[i].mStopped == false) aValidMatchs.push_back(i);
  int aClusterSize;
  int r = 0;
  do {
    CGraph aGraph;
    CMatrix<std::vector<unsigned int> > aMatchIndex;
    std::vector<int> aList;
    if (!buildGraph(aPoints,aValidMatchs,aMatchIndex,aList,aGraph)) continue;
    CVector<float> X;
    lanczos(aGraph,X);
    for (unsigned int i = 0; i < aList.size(); i++) {
      // Extract dominant cluster
      if (X(i) > 0.01)
        for (unsigned int k = 0; k < aMatchIndex.data()[aList[i]].size(); k++) {
          int aIndex = aMatchIndex.data()[aList[i]][k];
          aPoints[aIndex].mLabel = r;
        }
    }
    // Remove extracted cluster from valid points
    aClusterSize = 0;
    for (unsigned int ai = 0; ai < aValidMatchs.size(); ai++) {
      int i = aValidMatchs[ai];
      if (aPoints[i].mLabel >= 0) {
        aValidMatchs[ai] = aValidMatchs.back();
        aValidMatchs.pop_back();
        ai--;
        aClusterSize++;
      }
    }
    r++;
  } while (aValidMatchs.size() > 0 && aClusterSize > 0.025*aPoints.size());
  // Remove cluster that was too small
  if (aClusterSize <= 0.025*aPoints.size()) {
    r--;
    for (unsigned int i = 0; i < aPoints.size(); i++)
      if (aPoints[i].mLabel == r) aPoints[i].mLabel = -1;
  }
}



