#ifndef CMatchingH
#define CMatchingH

#include <vector>
#include <CVector.h>

// Requires ARPACK to be linked to the project

class CMatch {
public:
  CMatch() {mStopped = false; mLabel = -1;}
  int mox,moy;                          // original point
  int mx,my;                            // matched point
  int mLabel;                           // assignment to a cluster
  bool mStopped;                        // matching failed due to occlusion, etc.
  float mScore;                         // matching score
};

class CTrack {
public:
  CTrack() {mStopped = false; mLabel = -1; mNearestNeighbors.setSize(12); mNearestNeighbors = -1; mNearestNeighborsDX.setSize(12); mNearestNeighborsDX = 1e9; mNearestNeighborsDY.setSize(12); mNearestNeighborsDY = 1e9;}
  std::vector<float> mx,my;             // current position of the track
  std::vector<float> mFlowVar;          // flow variation in the area around the track integrated over time
  int mox,moy;                          // original starting point of the track
  int mLabel;                           // assignment to a region
  bool mStopped;                        // tracking stopped due to occlusion, etc.
  int mScale;                           // existance only of a fine grid (mScale = 1) or also at coarser grids (mScale = 2, 4, 8, 16)
  int mSetupTime;                       // Time when track was created
  CVector<float> mProjection;           // Low-dimensional projection of track by Laplacian eigenmap
  CVector<float> mNearestNeighborsDX;   // k-nearest spatial neighbors distance (x component)
  CVector<float> mNearestNeighborsDY;   // k-nearest spatial neighbors distance (y component)
  CVector<int> mNearestNeighbors;       // k-nearest spatial neighbors

  void readFromFile(std::ifstream& aIn);
  void writeToFile(std::ofstream& aOut);
};

class CMatching {
public:
  CMatching();
  // Computes oriented gradient histograms for all pixels in the image
  static void computeHOG(CTensor<float>& aImage, CTensor<float>& aHOG);
  // Computes the smaller eigenvalue of the structure tensor as a measure of cornerness
  static void computeCorners(CTensor<float>& aImage, CMatrix<float>& aCorners, float aRho);
  // Computes correspondences between two images based on the HOG descriptor
  // Initial points as listed in aTracks, search area restricted by r and aAllowed
  static void hogMatching(std::vector<CMatch>& aPoints, CTensor<float>& aImage1, CTensor<float>& aImage2, CTensor<float>& aHOG1, CTensor<float>& aHOG2, CMatrix<float>& aEnergy1, CMatrix<float>& aEnergy2, int rx1, int ry1, int rx2, int ry2, CMatrix<bool>& aAllowed);
  // Computes a motion histogram including outlier removal
  static void findMotions(CTensor<float>& aImage1, CTensor<float>& aImage2, CTensor<float>& aHOG1, CTensor<float>& aHOG2, CMatrix<float>& aEnergy1, CMatrix<float>& aEnergy2, CMatrix<float>& aCorners, float aAverageCornerness, int& rx1, int& ry1, int& rx2, int& ry2, CMatrix<bool>& aAllowed);
  // Computes bidirectional flow fields supported by HOG matches, assumes images to be smoothed a little already
  static void hogFlow(CTensor<float>& aImage1, CTensor<float>& aImage2, CTensor<float>& aHOG1, CTensor<float>& aHOG2, CMatrix<float>& aEnergy1, CMatrix<float>& aEnergy2, CTensor<float>& aForward, CTensor<float>& aBackward);
private:
  static void fastOneFrameClustering(std::vector<CMatch>& aPoints);
};

#endif
